/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
  * © 2016 OX Software GmbH, Germany. info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/model/cellcollection', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/utils/iteratorutils',
    'io.ox/office/tk/utils/simplemap',
    'io.ox/office/tk/object/triggerobject',
    'io.ox/office/tk/object/timermixin',
    'io.ox/office/editframework/utils/border',
    'io.ox/office/editframework/utils/attributeutils',
    'io.ox/office/editframework/utils/hyperlinkutils',
    'io.ox/office/spreadsheet/utils/operations',
    'io.ox/office/spreadsheet/utils/sheetutils',
    'io.ox/office/spreadsheet/utils/subtotalresult',
    'io.ox/office/spreadsheet/model/cellmodel',
    'io.ox/office/spreadsheet/model/cellmodelmatrix',
    'io.ox/office/spreadsheet/model/celloperationsbuilder',
    'io.ox/office/spreadsheet/model/formula/formulautils',
    'io.ox/office/spreadsheet/model/formula/tokenarray'
], function (Utils, IteratorUtils, SimpleMap, TriggerObject, TimerMixin, Border, AttributeUtils, HyperlinkUtils, Operations, SheetUtils, SubtotalResult, CellModel, CellModelMatrix, CellOperationsBuilder, FormulaUtils, TokenArray) {

    'use strict';

    // convenience shortcuts
    var ErrorCode = SheetUtils.ErrorCode;
    var Interval = SheetUtils.Interval;
    var Address = SheetUtils.Address;
    var Range = SheetUtils.Range;
    var IntervalArray = SheetUtils.IntervalArray;
    var RangeArray = SheetUtils.RangeArray;
    var ChangeDescriptor = SheetUtils.ChangeDescriptor;
    var MoveDescriptor = SheetUtils.MoveDescriptor;

    // regular expression to match a leading integer in a string
    var LEADING_INT_RE = /^(\d+)(.*)$/i;

    // regular expression to match a trailing integer in a string
    var TRAILING_INT_RE = /^(.*?)(\d+)$/i;

    // private global functions ===============================================

    /**
     * Returns the visibility flags for columns and rows according to the
     * option 'visible' supported by various methods of a cell collection.
     *
     * @param {Object} options
     *  Optional parameters passed to a method of class CellCollection with a
     *  'visible' option that can be a boolean, or one of the strings 'columns'
     *  or 'rows'.
     *
     * @returns {Object}
     *  An object with the boolean flags 'cols' and 'rows' specifying whether
     *  to visit visible columns and/or visible rows only.
     */
    function getVisibleMode(options) {
        var mode = Utils.getOption(options, 'visible');
        return { cols: (mode === true) || (mode === 'columns'), rows: (mode === true) || (mode === 'rows') };
    }

    /**
     * Prepares the passed range addresses, intended to be used by various cell
     * iterators.
     *
     * @param {RangeArray|Range} ranges
     *  An array of range addresses, or a single cell range address.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.reverse=false]
     *      If set to true, the ranges will be prepared for a reverse iterator.
     *      If a start position inside the ranges has been specified, the parts
     *      before this start position will be included in the result. Without
     *      reverse mode, the parts after the start position will be returned.
     *  @param {Address} [options.startIndex=0]
     *      If specified, the array index of the cell range in the passed
     *      range array that will be the first range in the result. MUST BE a
     *      valid array index. MUST remain zero, if a single range has been
     *      passed to this method.
     *  @param {Address} [options.startAddress]
     *      If specified, the address of a cell inside the first cell range
     *      (see option 'startIndex'). The cell address MUST BE located inside
     *      the start range. The start range will be split and shortened to the
     *      trailing parts (or leading parts in reverse mode) according to this
     *      start address.
     *  @param {Boolean} [options.skipStart=false]
     *      If set to true, the specified start address will not be included in
     *      the result ranges. This option has no effect without the option
     *      'startAddress'.
     *  @param {Boolean} [options.wrap=false]
     *      If set to true, the leading parts of the ranges (in reverse mode:
     *      the trailing parts of the ranges), according to the start position,
     *      will be appended to the resulting cell ranges. This option has no
     *      effect, if no custom start range or start address has been set
     *      (options 'startIndex' and 'startAddress').
     *
     * @returns {RangeArray}
     *  The resulting array of cell range addresses. Each range object in this
     *  array contains the following additional properties:
     *  - {Range} orig
     *      The original cell range address. May be larger, if using the option
     *      'startAddress' which leads to split the start range into multiple
     *      parts.
     *  - {Number} index
     *      The array index of the original range in the range array passed to
     *      this method (or zero, if a single range has been passed).
     */
    function getIteratorRanges(ranges, options) {

        // whether to generate the ranges for a reverse iterator
        var reverse = Utils.getBooleanOption(options, 'reverse', false);
        // a specific start array index
        var startIndex = Utils.getIntegerOption(options, 'startIndex', 0);
        // a specific start address
        var startAddress = Utils.getOption(options, 'startAddress', null);
        // whether to skip the start address in the beginning
        var skipStart = Utils.getBooleanOption(options, 'skipStart', false);
        // whether to wrap at the end of the range array
        var wrap = Utils.getBooleanOption(options, 'wrap', false);

        // insert original range addresses, and array indexes, to the ranges
        var cloneRanges = RangeArray.map(ranges, function (range, index) {
            var cloneRange = range.clone();
            cloneRange.orig = range;
            cloneRange.index = index;
            return cloneRange;
        });

        // no further processing required without custom start position
        if (!startAddress && (startIndex === 0)) {
            return cloneRanges;
        }

        // the start range according to the passed options
        var startRange = cloneRanges[startIndex];
        // the effective ranges, in the correct order according to the passed options
        var resultRanges = new RangeArray();

        // appends a new range (part of the start range) to the array of effective iterator ranges
        function appendRange(col1, row1, col2, row2) {
            var range = Range.create(col1, row1, col2, row2);
            range.orig = startRange.orig;
            range.index = startIndex;
            resultRanges.push(range);
        }

        // add the parts of the range array following the start address (in forward mode, or in wrapping mode)
        if (!reverse || wrap) {
            // add the trailing parts of the start range
            if (startAddress) {
                var startCol = startAddress[0] + ((reverse !== skipStart) ? 1 : 0);
                if (startCol === startRange.start[0]) {
                    appendRange(startRange.start[0], startAddress[1], startRange.end[0], startRange.end[1]);
                } else {
                    if (startCol <= startRange.end[0]) {
                        appendRange(startCol, startAddress[1], startRange.end[0], startAddress[1]);
                    }
                    if (startAddress[1] < startRange.end[1]) {
                        appendRange(startRange.start[0], startAddress[1] + 1, startRange.end[0], startRange.end[1]);
                    }
                }
            } else if (!reverse) {
                // no start address: add the start range in forward mode (in backward mode, it must be the last array element)
                resultRanges.push(startRange);
            }
            // add the ranges following the start range
            resultRanges.append(cloneRanges.slice(startIndex + 1));
        }

        // add the parts of the range array preceding the start address (in backward mode, or in wrapping mode)
        if (reverse || wrap) {
            // add the ranges preceding the start range
            resultRanges.append(cloneRanges.slice(0, startIndex));
            // add the leading parts of the start range
            if (startAddress) {
                var endCol = startAddress[0] - ((reverse === skipStart) ? 1 : 0);
                if (endCol === startRange.end[0]) {
                    appendRange(startRange.start[0], startRange.start[1], startRange.end[0], startAddress[1]);
                } else {
                    if (startRange.start[1] < startAddress[1]) {
                        appendRange(startRange.start[0], startRange.start[1], startRange.end[0], startAddress[1] - 1);
                    }
                    if (startRange.start[0] <= endCol) {
                        appendRange(startRange.start[0], startAddress[1], endCol, startAddress[1]);
                    }
                }
            } else if (reverse) {
                // no start address: add the start range in backward mode (in forward mode, it must be the first array element)
                resultRanges.push(startRange);
            }
        }

        return resultRanges;
    }

    // class SharedFormulaModel ===============================================

    function SharedFormulaModel(id, formula, range, anchor) {

        this.id = id;
        this.formula = formula;
        this.range = range.clone();
        this.anchor = anchor.clone();
    } // class SharedFormulaModel

    // public methods ---------------------------------------------------------

    SharedFormulaModel.prototype.clone = function () {
        return new SharedFormulaModel(this.id, this.formula, this.range, this.anchor);
    };

    SharedFormulaModel.prototype.setRange = function (range) {
        this.range = range.clone();
    };

    SharedFormulaModel.prototype.setAnchor = function (anchor) {
        this.anchor = anchor.clone();
    };

    // class CellCollection ===================================================

    /**
     * Collects cell contents and formatting attributes for a single sheet.
     * To save memory and to improve performance, instances of this class store
     * specific parts of the sheet only. More cell data will be fetched from
     * the server on demand.
     *
     * Triggers the following events:
     *  - 'change:cells'
     *      After the value, formula expression, or auto-style of one or more
     *      cells in this collection have been changed due to a 'changeCells'
     *      document operation. Event handler callbacks receive the following
     *      parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {ChangeDescriptor} changeDesc
     *          A descriptor object for the changed cells.
     *      (3) {Boolean} external
     *          Whether the changed cells originate from an external document
     *          operation received from the server.
     *      (4) {Boolean} results
     *          Whether this change event has been caused by the dependency
     *          manager while updating the results of formula cells.
     *  - 'move:cells'
     *      After some cells in this collection have been moved to a new
     *      positio ndue to a 'moveCells', 'insertColumns', 'deleteColumns',
     *      'insertRows', or 'deleteRows' document operation. Event handler
     *      callbacks receive the following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {MoveDescriptor} moveDesc
     *          A descriptor object for the moved cells.
     *      (3) {Boolean} external
     *          Whether the moved cells originate from an external document
     *          operation received from the server.
     *  - 'change:usedarea'
     *      After the size of the used area in the sheet has been changed.
     *      Event handlers receive the following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {Number} usedCols
     *          The number of used columns in the sheet.
     *      (3) {Number} usedRows
     *          The number of used rows in the sheet.
     *
     * @constructor
     *
     * @extends TriggerObject
     * @extends TimerMixin
     *
     * @param {SheetModel} sheetModel
     *  The sheet model instance containing this collection.
     */
    function CellCollection(sheetModel) {

        // self reference
        var self = this;

        // the document model, and model containers
        var docModel = sheetModel.getDocModel();
        var numberFormatter = docModel.getNumberFormatter();
        var formulaGrammar = docModel.getFormulaGrammar('op');
        var autoStyles = docModel.getCellAutoStyles();

        // the defined-lists collection
        var listCollection = docModel.getListCollection();

        // the column/row collections of the active sheet
        var colCollection = sheetModel.getColCollection();
        var rowCollection = sheetModel.getRowCollection();

        // all cell models mapped by address key
        var modelMap = new SimpleMap();

        // all cells, as sorted array of sorted columns
        var colMatrix = new CellModelMatrix(true);

        // all cells, as sorted array of sorted columns
        var rowMatrix = new CellModelMatrix(false);

        // all shared formulas mapped by ID
        var sharedFormulaMap = new SimpleMap();

        // the bounding range of existing cells in this collection
        var usedRange = null;

        // base constructors --------------------------------------------------

        TriggerObject.call(this);
        TimerMixin.call(this);

        // private methods ----------------------------------------------------

        /**
         * Recalculates the number of used columns and rows in the sheet, and
         * triggers a 'change:usedarea' event, if at least one of the counts
         * has been changed.
         */
        function updateUsedArea() {

            // get the used column and row intervals from the column matrix (null for empty matrix)
            var usedColInterval = colMatrix.getUsedInterval();
            var usedRowInterval = rowMatrix.getUsedInterval();
            // the new used range
            var newUsedRange = (usedColInterval && usedRowInterval) ? Range.createFromIntervals(usedColInterval, usedRowInterval) : null;

            // this collection remains empty
            if (!usedRange && !newUsedRange) { return; }

            // trigger change event if this collection is now empty
            if (!newUsedRange) {
                usedRange = null;
                self.trigger('change:usedarea', 0, 0);
                return;
            }

            // trigger change event if the end address of the used range has changed (the event contains the
            // number of columns/rows starting from cell A1, therefore changed start address does not matter),
            // or if this collection was empty before
            if (!usedRange || (usedRange.end[0] !== newUsedRange.end[0]) || (usedRange.end[1] !== newUsedRange.end[1])) {
                usedRange = newUsedRange;
                self.trigger('change:usedarea', usedRange.end[0] + 1, usedRange.end[1] + 1);
                return;
            }

            // store the new used range anyway (the start address may have changed)
            usedRange = newUsedRange;
        }

        /**
         * Inserts the passed new cell model into the internal containers.
         */
        function implInsertCellModel(cellModel, skipColMatrix, skipRowMatrix) {
            modelMap.insert(cellModel.key(), cellModel);
            if (!skipColMatrix) { colMatrix.insertModel(cellModel); }
            if (!skipRowMatrix) { rowMatrix.insertModel(cellModel); }
        }

        /**
         * Removes the passed new cell model from the internal containers.
         */
        function implRemoveCellModel(cellModel, skipColMatrix, skipRowMatrix) {
            modelMap.remove(cellModel.key());
            if (!skipColMatrix) { colMatrix.removeModel(cellModel); }
            if (!skipRowMatrix) { rowMatrix.removeModel(cellModel); }
        }

        /**
         * Creates a new blank cell model for this collection. The collection
         * MUST NOT contain an existing cell model at the specified address.
         *
         * @param {Address} address
         *  The address of the cell. The cell MUST be undefined.
         *
         * @returns {CellModel}
         *  A new blank cell model.
         */
        function createNewCellModel(address) {
            var cellModel = new CellModel(address);
            implInsertCellModel(cellModel);
            return cellModel;
        }

        /**
         * Removes a cell model from this collection.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Object}
         *  A flag set containing change flags for modified properties of the
         *  cell. The set will contain the following properties:
         *  - {Boolean} value
         *      Whether the deleted cell was not blank.
         *  - {Boolean} formula
         *      Whether the deleted cell contained a formula expression.
         *  - {Boolean} style
         *      Whether the auto-style of the deleted cell was different than
         *      the default column/row auto-style.
         *  - {Boolean} url
         *      Whether the deleted cell had contained a dynamic URL.
         */
        function deleteCellModel(address) {

            var cellModel = self.getCellModel(address);
            var changeFlags = { value: false, formula: false, style: false };

            if (cellModel) {
                changeFlags.value = cellModel.v !== null;
                changeFlags.formula = cellModel.isFormula() || cellModel.sr !== null || cellModel.si !== null;
                changeFlags.style = !autoStyles.areEqualStyleIds(cellModel.s, self.getDefaultStyleId(address));
                changeFlags.url = cellModel.url !== null;
                implRemoveCellModel(cellModel);
            }
            return changeFlags;
        }

        /**
         * Updates the value, formula, and auto-style of a cell model.
         *
         * @param {Address} address
         *  The address of the cell to be updated.
         *
         * @param {Object} contents
         *  The properties of the cell to be changed, with the following
         *  optional properties:
         *  - {Number|String|Boolean|ErrorCode|Null} [contents.v]
         *      The new value (or formula result) of the cell.
         *  - {String|Null} [contents.f]
         *      The new formula expression for the cell, or null to remove an
         *      existing formula.
         *  - {String} [contents.s]
         *      The identifier of an auto-style. The empty string can be used
         *      for the default auto-style of the document.
         *  - {String|Null} [contents.url]
         *      The URL of a dynamic hyperlink returned from a cell formula.
         *
         * @returns {Object}
         *  A flag set containing change flags for modified properties of the
         *  cell. The set will contain the following properties:
         *  - {Boolean} value
         *      Whether the cell value (or formula result) has changed.
         *  - {Boolean} formula
         *      Whether the formula expression has been changed.
         *  - {Boolean} style
         *      Whether the auto-style identifier has been changed.
         *  - {Boolean} url
         *      Whether the dynamic URL of the cell formula has changed.
         */
        function updateCellModel(address, contents) {

            // the cell model to be updated
            var cellModel = self.getCellModel(address);
            // the old auto-style (default to column/row auto style)
            var oldStyleId = cellModel ? cellModel.s : self.getDefaultStyleId(address);
            // the changed flags
            var changeFlags = { value: false, formula: false, style: false };

            var sharedFormula;

            // update cell value
            if ('v' in contents) {
                cellModel = cellModel || createNewCellModel(address);
                changeFlags.value = cellModel.setValue(contents.v);
            }

            // update cell formula
            if ('f' in contents) {
                cellModel = cellModel || createNewCellModel(address);
                changeFlags.formula = cellModel.setFormula(contents.f);
            }

            // update cell auto-style
            if ('s' in contents) {
                cellModel = cellModel || createNewCellModel(address);
                cellModel.s = contents.s;
                changeFlags.style = !autoStyles.areEqualStyleIds(oldStyleId, contents.s);
            }

            // update dynamic cell URL
            if ('url' in contents) {
                cellModel = cellModel || createNewCellModel(address);
                changeFlags.url = cellModel.setURL(contents.url);
            }

            // shared formula id
            if ('si' in contents) {
                cellModel = cellModel || createNewCellModel(address);
                changeFlags.formula = cellModel.setSharedFormulaID(contents.si) || changeFlags.formula;
            }

            // shared formula range
            if ('sr' in contents) {
                cellModel = cellModel || createNewCellModel(address);
                changeFlags.formula = cellModel.setSharedFormulaRange(contents.sr) || changeFlags.formula;
            }

            // update the shared formula if the shared formula cell has a new shared formula range or a new formula
            if (changeFlags.formula && _.isNumber(cellModel.si)) {
                sharedFormula = sharedFormulaMap.get(cellModel.si, null);
                if (sharedFormula) {
                    if (cellModel.sr) {
                        sharedFormula.setRange(cellModel.sr);
                        if ('sa' in contents) {
                            sharedFormula.setAnchor(contents.sa);
                        } else {
                            sharedFormula.setAnchor(address);
                        }

                        if (cellModel.f) {
                            sharedFormula.formula = cellModel.f;
                            if ('sa' in contents) {
                                sharedFormula.setAnchor(contents.sa);
                            } else {
                                sharedFormula.setAnchor(address);
                            }
                        }
                    }
                }
            }

            // update the display string, if value or attributes have changed
            // (performance: nothing to do if formula has changed, but the result has not)
            if (changeFlags.value || changeFlags.style) {
                var parsedFormat = autoStyles.getParsedFormat(cellModel.s);
                cellModel.d = numberFormatter.formatValue(parsedFormat, cellModel.v);
            }

            // delete the cached subtotal settings in the column/row matrixes, if the value has changed
            if (changeFlags.value) {
                cellModel.cv.clearSubtotals();
                cellModel.rv.clearSubtotals();
            }

            return changeFlags;
        }

        /**
         * Returns the token array of the given CellModel.
         *
         * @param {CellModel} cellModel to get the tokenArray.
         *
         * @returns {TokenArray|Null}
         *  The token array of this cell model; or null, if the cell model does not
         *  contain a formula expression.
         */
        function getCellModelTokenArray(cellModel) {
            if (cellModel) {
                if (cellModel.isSharedFormulaWithoutFormula()) {
                    var sharedFormula = sharedFormulaMap.get(cellModel.si);
                    if (sharedFormula && sharedFormula.formula) {
                        // Parse Formula
                        var tokenArray = new TokenArray(sheetModel, 'cell');
                        tokenArray.parseFormula('op', sharedFormula.formula, { refAddress: sharedFormula.anchor });
                        tokenArray.relocateRanges(sharedFormula.anchor, cellModel.a);
                        return tokenArray;
                    }
                }
                return cellModel.getTokenArray(sheetModel);
            }
            return null;
        }

        /**
         * Returns a predicate function that maches cell models for specific
         * type specifiers.
         *
         * @param {String} type
         *  The cell type specifier. See option 'type' of the private method
         *  createAddressIterator() for details.
         *
         * @returns {Function}
         *  A predicate function that receives either null or an instance of
         *  CellModel, and return whether that input value matches the type
         *  specifier.
         */
        var getCellTypePredicate = _.memoize(function (type) {
            switch (type) {
                case 'defined':
                    return _.identity;
                case 'value':
                    return CellModel.hasValue;
                case 'formula':
                    return function (type) {
                        return CellModel.hasFormula(type) && !CellModel.hasSharedFormula(type);
                    };
                case 'anyformula':
                    return CellModel.hasAnyFormula;
                case 'sharedformula':
                    return CellModel.hasSharedFormula;
            }
            // default: all cells
            return _.constant(true);
        });

        function createIntervalIterator(colIntervals, rowIntervals, options) {

            // which cell entries (by type) will be visited
            var type = Utils.getStringOption(options, 'type', 'all');
            // whether to iterate in reversed order
            var reverse = Utils.getBooleanOption(options, 'reverse', false);
            // the matcher predicate function for the cell type
            var matcher = getCellTypePredicate(type);

            // use the column or row matrix to visit the existing cell models faster
            // (test by checking whether the matcher does not accept missing cell models)
            if (!matcher(null)) {

                // the model iterator created by the cell matrix
                // TODO: use a parallel iterator in the column matrix for a small number of columns
                var modelIt = (colIntervals.size() === 1) ?
                    colMatrix.createCellIterator(colIntervals, rowIntervals, reverse) :
                    rowMatrix.createCellIterator(rowIntervals, colIntervals, reverse);

                // filter for specific cell types, and transform cell models to their addresses
                return IteratorUtils.createTransformIterator(modelIt, function (cellModel) {
                    return matcher(cellModel) ? { value: cellModel.a.clone(), model: cellModel } : null;
                });
            }

            // the outer iterator for all rows in the row intervals
            rowIntervals = IntervalArray.get(rowIntervals);
            var rowIt = rowIntervals.indexIterator({ reverse: reverse });
            // the generator callback for all columns in the column intervals
            colIntervals = IntervalArray.get(colIntervals);
            var colItGen = colIntervals.indexIterator.bind(colIntervals, { reverse: reverse });
            // create a nested iterator that visits all rows and columns, and returns a result with a cell address
            return IteratorUtils.createNestedIterator(rowIt, colItGen, function (rowResult, colResult) {
                var address = new Address(colResult.value, rowResult.value);
                return { value: address, model: self.getCellModel(address) };
            });
        }

        /**
         * Creates an iterator that generates the cell addresses for all, or
         * specific, cells contained in the passed cell ranges.
         *
         * @param {RangeArray|Range}
         *  An array of cell range addresses, or a single cell range address.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {String} [options.type='all']
         *      Specifies which type of cells will be covered by the iterator.
         *      Must be one of the following values:
         *      - 'all': Generates the addresses for all defined and undefined
         *          cells in the passed ranges.
         *      - 'defined': Generates the addresses for all defined cells that
         *          really exist in this collection, regardless of their
         *          content value (including blank but formatted cells).
         *      - 'value': Generates the addresses for all non-blank cells
         *          (with any content value), regardless if the value is given
         *          literally, or is the result of a formula.
         *      - 'formula': Generates the addresses of formula cells only.
         *  @param {Boolean|String} [options.visible=false]
         *      Specifies how to handle hidden columns and rows:
         *      - true (boolean): Only visible cells (cells in visible columns,
         *          AND visible rows) will be covered by the iterator.
         *      - 'columns': Cells in hidden columns will be skipped, but cells
         *          in visible columns and hidden rows will be visited.
         *      - 'rows': Cells in hidden rows will be skipped, but cells in
         *          visible rows and hidden columns will be visited.
         *      - false (boolean, default): By default, all visible and hidden
         *          cells in the passed cell ranges will be covered.
         *  @param {Boolean} [options.reverse=false]
         *      If set to true, the ranges AND the cell addresses in each range
         *      will be visited in reversed order.
         *
         * @returns {Object}
         *  An iterator object with the method next(). The result object of the
         *  iterator will contain the following value properties on success:
         *  - {Address} value
         *      The cell address currently visited.
         *  - {CellModel|Null} model
         *      The cell model at the current address if existing; otherwise
         *      null.
         *  - {Range} orig
         *      The address of the original cell range (from the passed range
         *      array) containing the current cell address.
         *  - {Number} index
         *      The array index of the original cell range contained in the
         *      result property 'orig'.
         */
        function createAddressIterator(ranges, options) {

            // whether to visit visible columns and/or rows only
            var visible = getVisibleMode(options);
            // whether to iterate in reversed order
            var reverse = Utils.getBooleanOption(options, 'reverse', false);

            // creates the appropriate iterator for the passed cell range
            function createRangeIterator(range) {

                // the column intervals of the range to be visited (all or visible columns)
                var colIntervals = visible.cols ? colCollection.getVisibleIntervals(range.colInterval()) : new IntervalArray(range.colInterval());
                // the row intervals of the range to be visited (all or visible rows)
                var rowIntervals = visible.rows ? rowCollection.getVisibleIntervals(range.rowInterval()) : new IntervalArray(range.rowInterval());

                // create an iterator that visits the cells row-by-row through the column intervals
                return createIntervalIterator(colIntervals, rowIntervals, options);
            }

            // combines the results of the outer and inner iterators
            function createIteratorResult(arrayResult, matrixResult) {
                return { value: matrixResult.value, model: matrixResult.model, orig: arrayResult.value, index: arrayResult.index };
            }

            // create the nested iterator, combining an outer array iterator with an inner address iterator for each range
            var iterator = RangeArray.get(ranges).iterator({ reverse: reverse });
            return IteratorUtils.createNestedIterator(iterator, createRangeIterator, createIteratorResult);
        }

        function createParallelAddressIterator(ranges, options) {

            var iterators = ranges.map(function (range) { return createAddressIterator(range, options); });

            var iterator = IteratorUtils.createParallelIterator(iterators, function (address, result) {
                return result.orig.indexAt(address);
            });

            return IteratorUtils.createTransformIterator(iterator, function (addresses, result) {
                result.value = ranges.map(function (range) { return range.addressAt(result.offset); });
                return result;
            });
        }

        /**
         * Returns a callback function that converts a cell address to a
         * numeric sorting index that is oriented horizontally, intended to be
         * used by synchronized and parallel address iterators.
         *
         * @returns {Function}
         *  A callback function that converts a cell address to a numeric
         *  sorting index.
         */
        function createAddressIndexer() {

            // the number of columns in the sheet
            var colCount = docModel.getMaxCol() + 1;

            // convert the passed cell address to a row-oriented sort index
            return function (address) { return address[1] * colCount + address[0]; };
        }

        /**
         * Returns the new format code for a cell to be applied while changing
         * the contents of that cell.
         *
         * @param {Address} address
         *  The address of the cell to be changed.
         *
         * @param {ParsedFormat} parsedFormat
         *  The parsed format intended to be applied for the cell, e.g. as
         *  received after parsing an edit string, or after calculating the
         *  result of a formula.
         *
         * @returns {String|Null}
         *  The format code contained in the parsed format, if it needs to be
         *  applied to the specified cell; or null, if the cell's number format
         *  will be retained.
         */
        function getNewCellFormatCode(address, parsedFormat) {

            // ignore the standard number formats
            if (parsedFormat.isStandard()) { return null; }

            // ignore changes inside any of the date/time categories (e.g. do not change a date cell to a time cell)
            var oldParsedFormat = self.getCellParsedFormat(address);
            if (parsedFormat.hasEqualCategory(oldParsedFormat, { anyDateTime: true })) { return null; }

            // return the format code of the parsed format on success
            return parsedFormat.formatCode;
        }

        /**
         * Returns an object with various flags specifying how a cell will be
         * changed according to the passed cell content data.
         *
         * @param {Object} contents
         *  The cell content data to be examined.
         *
         * @returns {Object}
         *  A result object with the following properties:
         *  - {Boolean} undefine
         *      Whether the cell will be undefined (physically deleted from
         *      this collection).
         *  - {Boolean} blank
         *      The cell will become blank (value and formula will be removed).
         *  - {Boolean} value
         *      The cell value will be changed, and the cell may not be blank
         *      afterwards (this flag will only be set, if 'blank' is false).
         *  - {Boolean} formula
         *      The cell formula will be changed, or the cell formula will be
         *      deleted without resulting in a blank cell (this flag will only
         *      be set, if 'blank' is false).
         *  - {Boolean} style
         *      The cell formatting attributes, or the auto-style will be
         *      changed.
         */
        function getChangeFlags(contents) {

            var undefine = contents.u === true;
            var blank = undefine || ((contents.v === null) && (contents.f === null));

            return {
                undefine: undefine,
                blank: blank,
                value: !blank && ('v' in contents),
                formula: !blank && (('f' in contents) || ('si' in contents) || ('sr' in contents)),
                style: !undefine && (('s' in contents) || ('a' in contents) || ('format' in contents))
            };
        }

        /**
         * Divides the passed cell range addresses into column ranges, row
         * ranges, and other cell ranges.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *
         * @returns {Object}
         *  A result descriptor with the following properties:
         *  - {RangeArray} [colRanges]
         *      The column ranges from the passed range array. The property
         *      will be missing, if the passed range array does not contain any
         *      column ranges.
         *  - {RangeArray} [rowRanges]
         *      The row ranges from the passed range array that are not ranges
         *      covering the entire sheet (considered to be column ranges). The
         *      property will be missing, if the passed range array does not
         *      contain any row ranges.
         *  - {RangeArray} [cellRanges]
         *      The remaining cell ranges from the passed range array that are
         *      neither column ranges nor row ranges. The property will be
         *      missing, if the passed range array does not contain any regular
         *      cell ranges.
         */
        function getRangeGroups(ranges) {
            return RangeArray.group(ranges, function (range) {
                return docModel.isColRange(range) ? 'colRanges' : docModel.isRowRange(range) ? 'rowRanges' : 'cellRanges';
            });
        }

        /**
         * Tags the passed cell range addresses with an additional 'rangeMode'
         * property specifying how to process the cell formatting of the range.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *
         * @returns {RangeArray}
         *  Copies of all passed ranges, with additional 'rangeMode' properties
         *  as expected e.g. by the method generateCellContentOperations(). The
         *  row ranges will be split into row ranges covering rows with and
         *  without active custom auto-styles.
         */
        function getTaggedRanges(ranges) {

            // group the ranges into column ranges, row ranges, and other cell ranges
            var rangeGroups = RangeArray.group(ranges, function (range) {
                return docModel.isColRange(range) ? 'colRanges' : docModel.isRowRange(range) ? 'rowRanges' : 'cellRanges';
            });

            // the resulting ranges to be filled, with additional 'rangeMode' property
            var taggedRanges = new RangeArray();

            // Split row ranges into rows with and without custom auto-style. In rows without custom
            // auto-style, the new formatting attributes must be merged with column default formatting
            // (range mode 'row:merge'). Otherwise, only existing cells need to be processed in rows
            // that already contain an active custom auto-style (range mode 'row:def').
            if (rangeGroups.rowRanges) {
                // convert the ranges to row intervals, forward all existing fill data objects
                var rowIntervals = IntervalArray.map(rangeGroups.rowRanges, function (range) {
                    var interval = range.rowInterval();
                    interval.rangeMode = range.rangeMode;
                    interval.fillData = range.fillData;
                    return interval;
                });
                // split the rows into formatted and unformatted ranges, forward all existing fill data objects
                IteratorUtils.forEach(rowCollection.createStyleIterator(rowIntervals), function (interval, result) {
                    var rowRange = docModel.makeRowRange(interval);
                    rowRange.rangeMode = result.orig.rangeMode || ((result.style === null) ? 'row:merge' : 'row:def');
                    rowRange.fillData = result.orig.fillData;
                    taggedRanges.push(rowRange);
                });
            }

            // Add the column ranges. New formatting attributes must be merged over active custom
            // auto-styles of the rows (range mode 'col:merge').
            if (rangeGroups.colRanges) {
                rangeGroups.colRanges.forEach(function (colRange) {
                    colRange.rangeMode = colRange.rangeMode || 'col:merge';
                });
                taggedRanges.append(rangeGroups.colRanges);
            }

            // Add the remaining ranges not covering entire columns or rows.
            if (rangeGroups.cellRanges) {
                taggedRanges.append(rangeGroups.cellRanges);
            }

            return taggedRanges;
        }

        /**
         * Returns the range addresses of the adjacent cells of all passed cell
         * ranges.
         */
        function getAdjacentRanges(ranges, columns, leading) {
            var maxIndex = leading ? 0 : docModel.getMaxIndex(columns);
            var checkRange = leading ? function (range) { return range.getStart(columns) > 0; } : function (range) { return range.getEnd(columns) < maxIndex; };
            var makeRange = leading ? function (range) { return range.lineRange(columns, -1); } : function (range) { return range.lineRange(columns, range.size(columns)); };
            return RangeArray.map(ranges, function (range) { return checkRange(range) ? makeRange(range) : null; });
        }

        function getHomogeneousOutlineBorders(range) {

            function detectHomogeneousBorder(range, columns, leading) {
                var outerKey        = SheetUtils.getOuterBorderKey(columns, leading),
                    detectedBorder  = null;

                IteratorUtils.forEach(self.createStyleIterator(range, { covered: true }), function (r, res) {
                    var attributes  = autoStyles.getBorderMap(res.style),
                        border      = attributes[outerKey];

                    if (border && border.style !== 'none' && (detectedBorder === null || _.isEqual(detectedBorder, border))) {
                        detectedBorder = border;
                    } else {
                        detectedBorder = null;
                        return Utils.BREAK;
                    }
                });

                return detectedBorder;
            }

            return {
                borderTop:      detectHomogeneousBorder(range.headerRow(),   false, true),
                borderRight:    detectHomogeneousBorder(range.trailingCol(), true,  false),
                borderBottom:   detectHomogeneousBorder(range.footerRow(),   false, false),
                borderLeft:     detectHomogeneousBorder(range.leadingCol(),  true,  true)
            };
        }

        /**
         * Generates the cell operations, and the undo operations, to change
         * the specified contents for multiple cells in the sheet, and
         * optionally the appropriate hyperlink operations.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {Address} address
         *  The address of the top-left cell to be modified.
         *
         * @param {Array<Object>} matrix
         *  An array of row descriptors (with properties 'c' and 'r') to change
         *  multiple cells at once. See description of the public method
         *  CellCollection.generateCellContentOperations() for details.
         *  Additionally, the following internal properties are supported:
         *  - {String} [contents.rangeMode='cell:merge']
         *      Specifies which cells will be processed in the ranges covered
         *      by the contents matrix. The following values
         *      are supported:
         *      - 'cell:merge' (default):
         *          Process all cells covered by a cell range. Used for regular
         *          cell formatting that will be merged over column and row
         *          default styles.
         *      - 'row:def':
         *          Process only existing cells (do not create new cells). Used
         *          when formatting entire rows with own active default style.
         *      - 'row:merge':
         *          Merge attributes for undefined cells with column default
         *          styles. Used when formatting entire rows without own active
         *          default style.
         *      - 'col:merge':
         *          Merge attributes for undefined cells with active row
         *          default styles. Used when formatting entire columns.
         *  - {String} [contents.cacheKey]
         *      A unique key for the formatting attributes and auto-style
         *      carried in this cell contents object. If set, an internal cache
         *      will be filled with the resulting auto-style identifier, and
         *      subsequent cell entries with the same cache key and old
         *      auto-style will resolve the new auto-style from that cache,
         *      instead of trying to receive the auto-style identifier from the
         *      auto-style collection (performance optimization).
         *
         * @param {Object} [options]
         *  Optional parameters that will be passed to the finalize method of
         *  the cell operations builders passed to the callback function. See
         *  CellOperationsBuilder.finalizeOperations() for details.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved with the addresses of the ranges
         *  that have really been changed, after all operations have been
         *  generated.
         */
        var generateCellContentOperations = SheetUtils.profileAsyncMethod('CellCollection.generateCellContentOperations()', function (generator, address, matrix, options) {
            SheetUtils.log('address=' + address + ' matrix=', matrix);

            // the cell operation builder
            var builder = new CellOperationsBuilder(sheetModel, generator, { startAddress: address });
            // cell range arrays per URL, will be merged together below
            var linkRangesMap = {};
            // the start address for the next row (or row band)
            var nextRowAddress = address.clone();
            // the number of cells already processed, for overflow detection
            var changedCount = 0;
            // the indexer for the synchronized address iterators
            var addressIndexer = createAddressIndexer();

            // creates an appropriate address iterator for the passed cell range and cell content properties
            function createIterators(range, contents) {

                // the change flags specifying how the cell will be potentially changed
                var changeFlags = getChangeFlags(contents);
                // special behavior when clearing values without new formatting
                var clearValues = changeFlags.blank && !changeFlags.style;

                function createIteratorForRange(iterRange, type) {
                    var repeatCols = type === 'all';
                    var colCount = repeatCols ? iterRange.cols() : 1;
                    var destRange = repeatCols ? iterRange.leadingCol() : iterRange;
                    var iterator = createAddressIterator(destRange, { type: type });
                    return IteratorUtils.createTransformIterator(iterator, function (address, result) {
                        result.contents = contents;
                        result.colCount = colCount;
                        result.countCells = !clearValues;
                        return result;
                    });
                }

                function createIteratorForIntervals(colIntervals, rowIntervals, type) {
                    var iterator = createIntervalIterator(colIntervals, rowIntervals, { type: type });
                    return IteratorUtils.createTransformIterator(iterator, function (address, result) {
                        result.contents = contents;
                        result.colCount = 1;
                        result.countCells = !clearValues;
                        return result;
                    });
                }

                // undefine cells: visit all existing cell models, regardless of range mode (cells/columns/rows)
                if (changeFlags.undefine) { return createIteratorForRange(range, 'defined'); }

                // removing values, but not the formulas, is an illegal operation
                if ((contents.v === null) && (contents.f !== null)) {
                    Utils.error('CellCollection.generateCellContentOperations(): illegal operation: clearing values without formulas');
                    return null;
                }

                // clear values without new formatting: visit value cells only
                if (clearValues) { return createIteratorForRange(range, 'value'); }

                // when setting values or formulas, all cells must be visited (regardless of passed range mode)
                var rangeMode = contents.rangeMode || 'cell:merge';
                if (changeFlags.value || changeFlags.formula || (changeFlags.style && (rangeMode === 'cell:merge'))) {
                    return createIteratorForRange(range, 'all');
                }

                // early exit, if nothing will be changed at all (remaining change flags are 'blank' and 'style')
                if (!changeFlags.blank && !changeFlags.style) { return null; }

                // formatting of blank ranges specific to range mode
                switch (rangeMode) {

                    // create an iterator that visits existing cells only
                    case 'row:def':
                        return createIteratorForRange(range, 'defined');

                    // When changing the formatting of unformatted rows, the column intervals with default auto-style do
                    // not need to be merged with the new formatting attributes (the new row default auto-style will suffice),
                    // but column intervals with other auto-styles need to be merged with the new formatting attributes.
                    case 'row:merge':
                        var colIntervals1 = new IntervalArray();
                        var colIntervals2 = new IntervalArray();
                        IteratorUtils.forEach(colCollection.createStyleIterator(range.colInterval()), function (colInterval, result) {
                            // undefined cells in columns with the default style do not need to be processed
                            var colIntervals = autoStyles.isDefaultStyleId(result.style) ? colIntervals1 : colIntervals2;
                            colIntervals.push(colInterval);
                        });
                        var colIterator1 = createIteratorForIntervals(colIntervals1.merge(), range.rowInterval(), 'defined');
                        var colIterator2 = createIteratorForIntervals(colIntervals2.merge(), range.rowInterval(), 'all');
                        return [colIterator1, colIterator2];

                    // When changing the formatting of columns, the row intervals without custom auto-style (row attribute
                    // 'customFormat' set to false) do not need to be merged with the new formatting attributes (the new column
                    // default auto-style will suffice), but row intervals with custom formatting need to be merged with the
                    // new formatting attributes.
                    case 'col:merge':
                        var rowIterator = rowCollection.createStyleIterator(range.rowInterval());
                        return IteratorUtils.createNestedIterator(rowIterator, function (rowInterval, result) {
                            // if the current row interval is not formatted, create a cell iterator that visits existing cells only
                            var rowRange = Range.createFromIntervals(range.colInterval(), rowInterval);
                            return createIteratorForRange(rowRange, (result.style === null) ? 'defined' : 'all');
                        });
                }

                Utils.error('CellCollection.generateCellContentOperations(): invalid range mode "' + rangeMode + '"');
                return null;
            }

            // use a time-sliced loop on the row array to prevent freezing browser
            var promise = self.iterateArraySliced(matrix, function (rowData) {

                // row repetition count
                var rowCount = rowData.r || 1;
                // the start address of the next cell contents range
                var rowAddress = nextRowAddress.clone();
                // update the address of the next row band
                nextRowAddress[1] += rowCount;

                // skip entire row (or multiple rows), if no 'c' property exists
                if (!_.isArray(rowData.c)) { return; }

                // the cell iterators for all cell data objects in this row band
                var nextCellAddress = rowAddress.clone();
                var rangeIterators = rowData.c.reduce(function (iterators, contents) {

                    // repetition count for the cell data
                    var colCount = contents.r || 1;
                    // address of the cell range covered by the cell contents object
                    var range = Range.createFromAddressAndSize(nextCellAddress.clone(), colCount, rowCount);

                    // update address of the next cell data object
                    nextCellAddress[0] += colCount;

                    // collect the URLs for hyperlinks
                    if (contents.url) {
                        var linkRanges = linkRangesMap[contents.url] || (linkRangesMap[contents.url] = new RangeArray());
                        linkRanges.push(range);
                    }

                    // create cell iterators (will be null, if the cells will not change at all, e.g. URL only)
                    var newIterators = createIterators(range, contents);
                    return newIterators ? iterators.concat(newIterators) : iterators;
                }, []);

                // create a synchronized iterator that visits the cells row-by-row
                var syncIt = IteratorUtils.createSynchronizedIterator(rangeIterators, addressIndexer);
                return self.iterateSliced(syncIt, function (address, result) {

                    // process the cells
                    builder.createCells(address, result.contents, result.colCount);

                    // count the changed cells, early exit on overflow
                    if (result.countCells) {
                        changedCount += result.colCount;
                        if (changedCount > SheetUtils.MAX_FILL_CELL_COUNT) {
                            return SheetUtils.makeRejected('cells:overflow');
                        }
                    }
                }, { delay: 'immediate' });
            }, { delay: 'immediate' });

            // generate all operations, return the addresses of all changed cell ranges
            promise = promise.then(function () {

                var hyperlinkCollection = sheetModel.getHyperlinkCollection();

                // first, create operations to remove hyperlinks
                if ('' in linkRangesMap) {
                    hyperlinkCollection.generateHyperlinkOperations(generator, linkRangesMap[''].merge(), '');
                    delete linkRangesMap[''];
                }

                // create new hyperlinks for the cells
                _.each(linkRangesMap, function (linkRanges, url) {
                    hyperlinkCollection.generateHyperlinkOperations(generator, linkRanges.merge(), url);
                });

                // generate the cell operations, return addresses of changed ranges to caller
                return builder.finalizeOperations(options);
            });

            return promise.done(function (changedRanges) {
                SheetUtils.log('changed=' + changedRanges);
            });
        });

        /**
         * Generates the cell operations, and the undo operations, to change
         * the contents for equally formatted or filled cell ranges in the
         * sheet, and optionally the appropriate hyperlink operations.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {RangeArray} taggedRanges
         *  The cell ranges to generate operations for. Each range address in
         *  this array may contain the following additional properties:
         *  - {String} [rangeMode='cell:merge']
         *      Specifies which cells will be processed in the range. Usually,
         *      the method getTaggedRanges() should be used to create this
         *      property. See method generateCellContentOperations() for more
         *      details.
         *  - {Object} [fillData]
         *      Additional formatting settings for the range. These data
         *      objects will be collected in an array, and will be passed to
         *      the callback function (see below).
         *
         * @param {Function} callback
         *  A callback function that will be invoked for each cell range of the
         *  partition of the passed cell ranges. Receives the fill data objects
         *  extracted from the original ranges covered by the current partition
         *  range as simple array in the first parameter. MUST return the cell
         *  contents object for the range to be inserted into the internal cell
         *  contents matrix.
         *
         * @param {Object} [options]
         *  Optional parameters that will be passed to the finalize method of
         *  the cell operations builders passed to the callback function. See
         *  CellOperationsBuilder.finalizeOperations() for details.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated.
         */
        var generateCellRangeOperations = (function () {

            // returns the effective (stronger) range mode of the passed range modes
            function mergeRangeMode(rangeMode1, rangeMode2) {
                // short-cut: equal range modes
                if (rangeMode1 === rangeMode2) { return rangeMode1; }
                // ignore range mode 'skip' (fall-back to other range mode)
                if (rangeMode1 === 'skip') { return rangeMode2; }
                if (rangeMode2 === 'skip') { return rangeMode1; }
                // needing to process all cells in a range always wins over every other range mode
                // (cell formatting wins over column and row default formatting)
                if ((rangeMode1 === 'cell:merge') || (rangeMode2 === 'cell:merge')) { return 'cell:merge'; }
                // row formatting wins over column formatting
                return (rangeMode1 === 'col:merge') ? rangeMode2 : rangeMode1;
            }

            // creates and returns  a new row entry in the passed contents matrix
            function addMatrixRow(matrix, count) {
                var rowData = { c: [], r: count };
                matrix.push(rowData);
                return rowData;
            }

            // the actual implementation returned from local scope
            function generateCellRangeOperations(generator, taggedRanges, callback, options) {
                SheetUtils.log('ranges=' + taggedRanges);

                // early exit for empty arrays
                if (taggedRanges.empty()) { return self.createResolvedPromise(new RangeArray()); }

                // the start address of the bounding range, used as anchor cell for the matrix
                var startAddress = taggedRanges.boundary().start;
                // the address of the next free entry in the matrix
                var currAddress = startAddress.clone();
                // do not waste time to collect the fill data of the original ranges, if the callback does not use them
                var collectFillData = callback.length > 0;
                // the contents matrix for all ranges
                var matrix = [];

                // process all row bands, use a time-sliced loop to prevent freezing browser
                var promise = self.iterateArraySliced(taggedRanges.getRowBands({ ranges: true }), function (rowBand) {

                    // add empty rows into the matrix to be skipped
                    if (currAddress[1] < rowBand.first) {
                        addMatrixRow(matrix, rowBand.first - currAddress[1]);
                    }

                    // create a new row in the matrix
                    var rowData = addMatrixRow(matrix, rowBand.size());

                    // convert the ranges covered by the row band to tagged column intervals
                    var colIntervals = IntervalArray.map(rowBand.ranges, function (range) {
                        var interval = range.colInterval();
                        interval.rangeMode = range.rangeMode;
                        interval.fillData = range.fillData;
                        return interval;
                    });

                    // iterate over a (sorted) column partition of the ranges covered by the row band
                    var promise2 = self.iterateArraySliced(colIntervals.partition(), function (colInterval) {

                        // skip the gaps between the ranges
                        if (currAddress[0] < colInterval.first) {
                            rowData.c.push({ r: colInterval.first - currAddress[0] });
                        }

                        // determine the range mode to be used for the current range
                        var rangeMode = colInterval.coveredBy.reduce(function (rangeMode, colInterval) {
                            return mergeRangeMode(rangeMode, colInterval.rangeMode || 'cell:merge');
                        }, 'col:merge');

                        // collect the fill data of all source ranges covered by the current row band and column interval
                        // (performance optimization: do not collect fill data, if callback function does not use them)
                        var fillDataArray = collectFillData ? _.pluck(colInterval.coveredBy, 'fillData').filter(_.identity) : null;

                        // create the cell contents entry for the current range
                        var contents = _.clone(callback(fillDataArray));
                        contents.r = colInterval.size();
                        contents.rangeMode = rangeMode;
                        rowData.c.push(contents);

                        // move current address behind the range processed here
                        currAddress[0] = colInterval.last + 1;

                    }, { delay: 'immediate' });

                    // back to first column for the next row band
                    return promise2.done(function () {
                        currAddress[0] = startAddress[0];
                        currAddress[1] = rowBand.last + 1;
                    });

                }, { delay: 'immediate' });

                // generate the cell operations from the matrix
                return promise.then(function () {
                    return generateCellContentOperations(generator, startAddress, matrix, options);
                });
            }

            return SheetUtils.profileAsyncMethod('CellCollection.generateCellRangeOperations()', generateCellRangeOperations);
        }());

        function generateChainedColIntervalOperations(promise, rangeGroups, callback) {
            return rangeGroups.colRanges ? promise.then(function (changedRanges) {
                var colIntervals = IntervalArray.map(rangeGroups.colRanges, function (range) {
                    var colInterval = range.colInterval();
                    colInterval.fillData = range.fillData;
                    return colInterval;
                });
                return callback(colIntervals).then(function (changedIntervals) {
                    return changedRanges.append(docModel.makeColRanges(changedIntervals));
                });
            }) : promise;
        }

        function generateChainedRowIntervalOperations(promise, rangeGroups, callback) {
            return rangeGroups.rowRanges ? promise.then(function (changedRanges) {
                var rowIntervals = IntervalArray.map(rangeGroups.rowRanges, function (range) {
                    var rowInterval = range.rowInterval();
                    rowInterval.fillData = range.fillData;
                    return rowInterval;
                });
                return callback(rowIntervals).then(function (changedIntervals) {
                    return changedRanges.append(docModel.makeRowRanges(changedIntervals));
                });
            }) : promise;
        }

        function generateChainedCellRangeOperations(promise, callback) {
            return promise.then(function (changedRanges) {
                return callback().then(function (changedRanges2) {
                    changedRanges = changedRanges.append(changedRanges2).merge();
                    SheetUtils.log('changed=' + changedRanges);
                    return changedRanges;
                });
            });
        }

        /**
         * Get the changes for the given SharedFormulaModel
         * @param {SharedFormulaModel} sharedFormula
         * @param {Object} [options]
         *  @param {Boolean} [options.getrange=false]
         *      True if the result contains the new Range for the SharedFormula.
         *  @param {AddressArray} [options.rangeexcludeaddress=[]]
         *      List of addresses which should be excluded to find the new range.
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the sharedForumla is checked for changes.
         *  The Promise returns an Object:
         *   @param {Address|Null} newanchor
         *      the new anchor address for the shared formula. Or null if no new anchor exists.
         *   @param {Array<Address>} the collected addresses if the flag 'collectAddresses' was set.
         *   @param {Range|Null} the new Range of the sharedFormula or Null if the flag 'getrange' is set
         *      to false or there is no cell for the sharedFormula exists.
         */
        function getSharedFormulaChanges(sharedFormula, options) {

            var getRange = Utils.getBooleanOption(options, 'getrange', false);
            var rangeExcludeAddresse = Utils.getObjectOption(options, 'rangeexcludeaddress', []);
            var addresses = [];

            var sharedForumlaRangeIterator = createAddressIterator(sharedFormula.range, { type: 'sharedformula' });
            var anchorIndex = sharedFormula.range.indexAt(sharedFormula.anchor);

            var newanchor = null;
            var startAddress;
            var endAddress;

            var promise = self.iterateSliced(sharedForumlaRangeIterator, function (address) {

                var sharedFormulaCellModel = self.getCellModel(address);
                // Check if the shared formula cell is a cell of the current sharedFormula
                if (sharedFormula.id === sharedFormulaCellModel.si) {
                    if (getRange && !_.find(rangeExcludeAddresse, function (excAddress) { return address.equals(excAddress); })) {
                        if (startAddress) {
                            if (endAddress[0] < address[0]) {
                                endAddress[0] = address[0];
                            }
                            if (endAddress[1] < address[1]) {
                                endAddress[1] = address[1];
                            }

                            if (startAddress[0] > address[0]) {
                                startAddress[0] = address[0];
                            }
                            if (startAddress[1] > address[1]) {
                                startAddress[1] = address[1];
                            }
                        } else {
                            startAddress = address.clone();
                            endAddress = address.clone();
                        }
                    }

                    // Test if it is the new anchor address
                    if (newanchor === null && sharedFormula.range.indexAt(address) > anchorIndex) {
                        newanchor = address.clone();
                        if (!getRange) {
                            return Utils.BREAK;
                        }
                    }
                }
            }, { delay: 'immediate' });

            return promise.then(function () {
                return { newanchor: newanchor, addresses: addresses, range: getRange && startAddress ? new Range(startAddress, endAddress) : null };
            });
        }

        function getAutoFormulaRange(startAddress, direction, skipStart, boundRange) {

            var address1 = null;
            var address2 = null;

            // find the largest range next to the start cell containing numbers only
            var cellIt = self.createLinearAddressIterator(startAddress, direction, { type: 'value', boundRange: boundRange, skipStart: skipStart });
            IteratorUtils.forEach(cellIt, function (address) {

                var value = self.getCellValue(address);

                if (typeof value !== 'number') {
                    if (address1) {
                        return Utils.BREAK;
                    } else {
                        return;
                    }
                }
                if (address2 && ((Math.abs(address2[0] - address[0]) > 1) || (Math.abs(address2[1] - address[1]) > 1))) {
                    return Utils.BREAK;
                }

                if (!address1) { address1 = address; }
                address2 = address;
            });

            if (address1 && address2) {
                var startCol = startAddress[0];
                var startRow = startAddress[1];

                if (direction === 'up') { startRow--; } else { startCol--; }

                return { address1: address1, range: Range.create(startCol, startRow, address2[0], address2[1]) };
            }
        }

        // autofill ------------------------------------------------------------

        // analyze dedicated ranges and decides which type of autoFill should grab here (increase, copy, ...)
        function getAutoFillType(obj, columns, reverse, invertType) {

            var range           = obj.baseRange,
                iterator        = null,
                arrNr           = null,
                parts           = null,
                i               = null,
                partIndex       = null,
                regex           = null,
                lastValue       = null,
                rise            = null;

            if (range.size(!columns) < 3) {
                if (invertType) {
                    return 'copy';
                } else {
                    return 'increase';
                }

            } else {
                iterator        = range.interval(!columns).iterator({ reverse: reverse });
                arrNr           = [];
                parts           = null;
                i               = 0;
                partIndex       = (obj.valueType === 'stringNumber') ? 2 : 1;
                regex           = (obj.valueType === 'stringNumber') ? TRAILING_INT_RE : LEADING_INT_RE;

                IteratorUtils.forEach(iterator, function () {
                    var address     = range.addressAt(i),
                        cell        = self.getCellModel(address),
                        nr          = null;

                    parts = regex.exec(cell.v);
                    nr = parseInt(parts[partIndex], 10);
                    arrNr.push(nr);

                    if (i > 0) {
                        if (rise === null || rise === (nr - lastValue)) {
                            rise = (nr - lastValue);
                        } else {
                            rise = false;
                        }
                    }

                    lastValue = nr;

                    i++;
                });

                var slope = FormulaUtils.slope(arrNr);
                if (rise !== false && rise === slope) {
                    if (invertType) {
                        return 'copy';
                    } else {
                        return 'increase';
                    }

                } else {
                    return 'copy';
                }
            }
        }

        // calculates the average and slope of the given array of numbers
        function calculateAverageAndSlope(arrNumbers, reverse) {
            var slope = null,
                average = null;

            if (arrNumbers.length > 1) {
                var array = (reverse) ? arrNumbers.reverse() : arrNumbers,
                    summe = _.reduce(array, function (memo, cellValue) { return memo + parseFloat(cellValue); }, 0);

                average = (summe / arrNumbers.length);
                slope = FormulaUtils.slope(array);

            } else {
                average = parseFloat(arrNumbers[0]);
                slope = (reverse) ? -1 : 1;
            }
            return {
                slope: slope,
                average: average
            };
        }

        // analyze a range and create a plan for the cell-operations for autofill
        function analyzeForAutoFill(range, columns, reverse, invertType) {
            // detect the type of the value
            function detectValueType(cellModel) {
                if (cellModel.isError()) {
                    return 'error';
                } else if (cellModel.isFormula()) {
                    return 'formula';
                } else if (cellModel.isSharedFormula()) {
                    return 'sharedFormula';
                } else if (cellModel.isNumber()) {
                    var parsedFormat = self.getCellParsedFormat(cellModel.a);
                    if (parsedFormat.category !== 'standard' && parsedFormat.category !== 'custom' && parsedFormat.category !== 'percent') {
                        return parsedFormat.category;
                    } else {
                        return 'number';
                    }
                } else if (cellModel.isText()) {
                    if (listCollection.listContains(cellModel.v)) {
                        return 'definedList';
                    } else if (TRAILING_INT_RE.test(cellModel.v)) {
                        return 'stringNumber';
                    } else if (LEADING_INT_RE.test(cellModel.v)) {
                        return 'numberString';
                    } else {
                        return 'string';
                    }
                } else if (cellModel.isBlank()) {
                    return '';
                } else {
                    return null;
                }
            }

            function analyzeDateTimeInterval(arrDates) {
                var returnObj = false;

                if (arrDates.length > 1) {
                    var i               = 0,
                        interval        = null,
                        unit            = null,

                        intervalDays    = null,
                        intervalMonth   = null,
                        intervalYears   = null,
                        intervalMilli   = null,
                        intervalSeconds = null,
                        intervalMinutes = null,
                        intervalHours   = null,

                        lastUnit        = null,
                        lastInterval    = null,
                        lastDateObj     = null;

                    var linear = _.every(arrDates, function (number) {
                        var date        = numberFormatter.convertNumberToDate(number),
                            momentObj   = moment({ year: date.getFullYear(), month: date.getMonth(), day: date.getDate(), hours: date.getHours(), minutes: date.getMinutes(), seconds: date.getSeconds(), milliseconds: date.getMilliseconds() }).utc();

                        if (i > 0) {
                            intervalMilli   = lastDateObj.diff(momentObj, 'milliseconds', true) * -1;
                            intervalSeconds = lastDateObj.diff(momentObj, 'seconds', true) * -1;
                            intervalMinutes = lastDateObj.diff(momentObj, 'minutes', true) * -1;
                            intervalHours   = lastDateObj.diff(momentObj, 'hours', true) * -1;
                            intervalDays    = lastDateObj.diff(momentObj, 'days', true) * -1;
                            intervalMonth   = lastDateObj.diff(momentObj, 'month', true) * -1;
                            intervalYears   = lastDateObj.diff(momentObj, 'years', true) * -1;

                            if (intervalYears % 1 === 0) {
                                unit = 'years';         interval = intervalYears;
                            } else if (intervalMonth % 1 === 0) {
                                unit = 'months';        interval = intervalMonth;
                            } else if (intervalDays % 1 === 0) {
                                unit = 'days';          interval = intervalDays;
                            } else if (intervalHours % 1 === 0) {
                                unit = 'hours';         interval = intervalHours;
                            } else if (intervalMinutes % 1 === 0) {
                                unit = 'minutes';       interval = intervalMinutes;
                            } else if (intervalSeconds % 1 === 0) {
                                unit = 'seconds';       interval = intervalSeconds;
                            } else if (intervalMilli % 1 === 0) {
                                unit = 'milliseconds';  interval = intervalMilli;
                            }

                            if (lastUnit === null && lastInterval === null) {
                                lastUnit = unit;
                                lastInterval = interval;
                            } else if (lastUnit !== unit || lastInterval !== interval) {
                                return false;
                            }
                        }

                        lastDateObj = momentObj;
                        i++;
                        return true;
                    });

                    if (linear && interval !== null && unit !== null) {
                        returnObj = {
                            interval: interval,
                            unit: unit
                        };
                    }

                } else {
                    returnObj = {
                        interval: 1,
                        unit: 'day'
                    };
                }

                return returnObj;
            }

            function analyzeDefinedListInterval(arr) {
                var lastList = null,
                    lastIndexArray = [],
                    linear = null,
                    interval = null;

                if (arr.length > 1) {
                    _.each(arr, function (val) {
                        var value = val.toLowerCase();
                        listCollection.each(function (list, i) {
                            if (lastList === null && _.contains(list, value)) {
                                lastList = i;
                                lastIndexArray.push(_.indexOf(list, value));

                            } else if (_.contains(list, value) && i === lastList) {
                                lastIndexArray.push(_.indexOf(list, value));

                            } else if (_.contains(list, value)) {
                                lastList = i;
                                lastIndexArray = [];
                                lastIndexArray.push(_.indexOf(list, value));
                            }
                        });
                    });

                    var lastDistance = null,
                        lastIndex = null;
                    linear = _.every(lastIndexArray, function (index) {
                        var currentDistance = null;
                        if (lastDistance === null && lastIndex === null) {
                            lastIndex = index;
                        } else if (lastDistance === null) {
                            lastDistance = (index - lastIndex);
                            if (lastDistance < 0) { lastDistance += listCollection.getList(lastList).length; }
                            lastIndex = index;
                        } else {
                            currentDistance = (index - lastIndex);
                            if (currentDistance < 0) { currentDistance += listCollection.getList(lastList).length; }
                            if (lastDistance !== currentDistance) {
                                return false;
                            }
                            lastDistance = currentDistance;
                            lastIndex = index;
                        }
                        return true;
                    });
                    interval = (linear) ? lastDistance : null;

                } else {
                    var value = arr[0].toLowerCase();
                    listCollection.some(function (list, i) {
                        if (_.contains(list, value)) {
                            lastList = i;
                            lastIndexArray.push(_.indexOf(list, value));
                            return true;
                        }
                    });

                    linear = true;
                    interval = 1;
                }

                return {
                    list: lastList,
                    indexArr: lastIndexArray,
                    linear: linear,
                    interval: interval
                };
            }

            // compares values of the last and current cell to find similar types
            function compareValues(type, last, current) {
                var partsLast = null,
                    partsCurrent = null;

                if (type === 'stringNumber') {
                    partsLast = TRAILING_INT_RE.exec(last);
                    partsCurrent = TRAILING_INT_RE.exec(current);
                    return partsLast[1].toLowerCase() === partsCurrent[1].toLowerCase();

                } else if (type === 'numberString') {
                    partsLast = LEADING_INT_RE.exec(last);
                    partsCurrent = LEADING_INT_RE.exec(current);
                    return partsLast[2].toLowerCase() === partsCurrent[2].toLowerCase();

                } else if (type === 'definedList') {
                    listCollection.each(function (list, listIndex) {
                        if (_.contains(list, last.toLowerCase()))    { partsLast    = listIndex; }
                        if (_.contains(list, current.toLowerCase())) { partsCurrent = listIndex; }
                    });
                    return (partsLast !== null && partsLast === partsCurrent);

                }
                return false;
            }

            var iterator            = range.interval(!columns).iterator({ reverse: reverse }),
                planingResult       = [],
                i                   = 0,
                lastKnownRange      = null,
                lastKnownValueType  = null,
                lastKnownValue      = null;

            IteratorUtils.forEach(iterator, function () {
                var address     = range.addressAt(i),
                    cell        = self.getCellModel(address),
                    valueType   = null;

                if (cell) {
                    valueType = detectValueType(cell);

                    if (lastKnownRange !== null && lastKnownValueType !== null) {
                        if (valueType === lastKnownValueType && ((valueType !== 'stringNumber' && valueType !== 'numberString' && valueType !== 'definedList') || (lastKnownValue !== null && compareValues(valueType, lastKnownValue, cell.v)))) {
                            lastKnownRange.end.move(1, !columns);
                        } else {
                            lastKnownRange = null;
                            lastKnownValueType = null;
                        }
                    }

                    if (lastKnownRange === null) {
                        lastKnownRange = new Range(address.clone(), address.clone());
                    }
                    if (lastKnownValueType === null) {
                        lastKnownValueType = valueType;
                    }

                    lastKnownValue = cell.v;
                } else {
                    lastKnownRange = null;
                    lastKnownValueType = null;
                    lastKnownValue = null;
                }

                planingResult.push({
                    baseCell: address,
                    baseRange: lastKnownRange,
                    valueType: valueType
                });

                i++;
            });

            // analyze the fill type (increase, copy, ...) for each found range
            _.each(planingResult, function (obj) {
                if (obj.valueType === 'stringNumber' || obj.valueType === 'numberString') {
                    obj.fillType = getAutoFillType(obj, columns, reverse, invertType);

                } else if (obj.valueType === 'date' || obj.valueType === 'time' || obj.valueType === 'datetime') {
                    var interval = analyzeDateTimeInterval(_.reduce(self.getRangeContents(obj.baseRange), function (memo, cell) { memo.push(cell.value); return memo; }, []));
                    if (interval !== false && !invertType) {
                        obj.fillType = 'increase';
                        obj.unit = interval.unit;
                        obj.interval = interval.interval;
                    } else {
                        obj.fillType = 'copy';
                    }

                } else if (obj.valueType === 'number') {
                    if ((!range.singleLine(!columns) && !invertType) || (range.singleLine(!columns) && invertType)) {
                        obj.fillType = 'increase';
                    } else {
                        obj.fillType = 'copy';
                    }

                } else if ((obj.valueType === 'formula' || obj.valueType === 'sharedFormula') && !invertType) {
                    obj.fillType = 'relocate';

                } else if (obj.valueType === 'definedList' && !invertType) {
                    var listInterval = analyzeDefinedListInterval(_.reduce(self.getRangeContents(obj.baseRange), function (memo, cell) { memo.push(cell.value); return memo; }, []));

                    if (listInterval.linear) {
                        obj.fillType     = 'increase';
                        obj.list         = listInterval.list;
                        obj.interval     = listInterval.interval;
                        obj.rangeIndexes = listInterval.indexArr;

                    } else {
                        obj.fillType = 'copy';
                    }

                } else {
                    obj.fillType = 'copy';
                }
            });

            // return the planing-result
            return planingResult;
        }

        // create target-range for autofill operations (for normal ranges)
        function createTargetRange(sourceRange, columns, count, reverse) {
            var targetRange = sourceRange.clone();
            var index = columns ? 1 : 0;
            var counter = (targetRange.end[index] - targetRange.start[index]) + 1;

            if (reverse) {
                // prepare the targetRange
                targetRange.start.move(-count, !columns);
                targetRange.end.move(-counter, !columns);

            } else {
                // prepare the targetRange
                targetRange.end.move(count, !columns);
                targetRange.start.move(counter, !columns);
            }

            return targetRange;
        }

        // create target-interval for autofill-operations (for complete cols/rows)
        function createTargetInterval(sourceInterval, count, reverse) {
            var targetInterval = sourceInterval.clone();

            if (reverse) {
                return new Interval((targetInterval.first - count), (targetInterval.first - 1));
            } else {
                return new Interval((targetInterval.last + 1), (targetInterval.last + count));
            }
        }

        // generates autofill-operations for cells
        function generateAutoFillOperations(generator, range, border, count, options) {

            // calculate the result for the autofill-operations
            function calculateResult(plan, round, sourceLineCount) {
                var from                 = self.getCellModel(plan.baseCell),

                    resultValue          = null,
                    formulaTargetAddress = null,
                    resultFormula        = null,
                    string               = null;

                // INCREASE (values like numbers, dates, defined lists, ...)
                if (plan.fillType === 'increase') {
                    // count up a index for each range separately
                    rangeIndex[plan.baseRange.toString()] = (rangeIndex[plan.baseRange.toString()] + 1) || 0;

                    if (plan.valueType === 'date' || plan.valueType === 'time' || plan.valueType === 'datetime') {
                        var cellDate    = numberFormatter.convertNumberToDate(self.getCellModel(plan.baseRange.start).v),
                            momentObj   = moment({ year: cellDate.getFullYear(), month: cellDate.getMonth(), day: cellDate.getDate(), hours: cellDate.getHours(), minutes: cellDate.getMinutes(), seconds: cellDate.getSeconds(), milliseconds: cellDate.getMilliseconds() }).utc(),
                            currIndex   = rangeIndex[plan.baseRange.toString()];

                        if (reverse) {
                            momentObj.subtract(((currIndex + 1) * plan.interval), plan.unit);
                        } else {
                            momentObj.add(((plan.baseRange.size(!columns) + currIndex) * plan.interval), plan.unit);
                        }

                        resultValue = numberFormatter.convertDateToNumber(momentObj.toDate());

                    } else if (plan.valueType === 'definedList') {
                        var list = listCollection.getList(plan.list),
                            currI = (rangeIndex[plan.baseRange.toString()] + 1),
                            newIndex = null;

                        if (reverse) {
                            newIndex = plan.rangeIndexes[0] - (currI * plan.interval);
                            while (newIndex < 0) { newIndex = list.length + newIndex; }

                        } else {
                            newIndex = plan.rangeIndexes[plan.rangeIndexes.length - 1] + (currI * plan.interval);
                            if (newIndex >= list.length) { newIndex %= list.length; }
                        }

                        resultValue = list[newIndex];

                        // make the whole string uppercase, if the initial value is uppercase too
                        if (self.getCellModel(plan.baseRange.start).v.toUpperCase() === self.getCellModel(plan.baseRange.start).v) {
                            resultValue = resultValue.toUpperCase();
                        // make first letter uppercase, if the first initial letter is uppercase too
                        } else if (self.getCellModel(plan.baseRange.start).v[0].toUpperCase() === self.getCellModel(plan.baseRange.start).v[0]) {
                            resultValue = resultValue[0].toUpperCase() + resultValue.slice(1);
                        }

                    } else {
                        var startCellIndex  = ((plan.baseRange.size(!columns) + 1) / 2),
                            multiplier      = null,

                            strIndex        = (plan.valueType === 'stringNumber') ? 1 : 2,
                            nrIndex         = (plan.valueType === 'stringNumber') ? 2 : 1,
                            regex           = (plan.valueType === 'stringNumber') ? TRAILING_INT_RE : LEADING_INT_RE,
                            valueGetter     = (plan.valueType === 'number') ? function (memo, cell) { memo.push(parseFloat(cell.value)); return memo; } : function (memo, cell) { memo.push(parseFloat(regex.exec(cell.value)[nrIndex])); return memo; },
                            calcObj         = calculateAverageAndSlope(_.reduce(self.getRangeContents(plan.baseRange), valueGetter, []), reverse);

                        if (reverse) {
                            multiplier = (((plan.baseRange.size(!columns) + 1) - startCellIndex) + rangeIndex[plan.baseRange.toString()]);
                        } else {
                            multiplier = (((plan.baseRange.size(!columns) + 1) - startCellIndex) + rangeIndex[plan.baseRange.toString()]);
                        }

                        // calculate the new number
                        resultValue = (calcObj.average + (multiplier * calcObj.slope));

                        // when the source-value is a combination of string and number
                        if (plan.valueType === 'stringNumber' || plan.valueType === 'numberString') {
                            // get the string(-part) of the first cell from the baseRange
                            string          = regex.exec(self.getCellModel(plan.baseRange.start).v)[strIndex]; // need to use the first cell, because of the case-sensitifity
                            // add the string to the resultValue
                            resultValue     = (plan.valueType === 'stringNumber') ? string + Math.abs(resultValue) : Math.abs(resultValue) + string;
                        }
                    }

                // RELOCATE (for formulas and sharedFormulas)
                } else if (plan.fillType === 'relocate') {
                    // set the cell value to zero, so that the user sees the "0" shortly till the formulaengine calculates the correct result
                    resultValue = 0;
                    // calculate the target address for relocating the formula
                    formulaTargetAddress = from.a.clone().move(((round * sourceLineCount) * ((reverse) ? -1 : 1)), !columns);
                    // relocate the formula to the target cell
                    resultFormula = getCellModelTokenArray(from).getFormula('op', { refAddress: from.a, targetAddress: formulaTargetAddress });

                // COPY (values)
                } else if (plan.baseRange) {
                    if (plan.valueType === 'formula') {
                        resultFormula = from.f;
                    } else if (plan.valueType === 'sharedFormula') {
                        // set the cell value to zero, so that the user sees the "0" shortly till the formulaengine calculates the correct result
                        resultValue = 0;
                        // relocate the formula to the target cell
                        resultFormula = getCellModelTokenArray(from).getFormula('op');
                    }

                    resultValue = (!_.isNull(from)) ? from.v : null;
                }

                // returns the result within an object
                return {
                    from: from,
                    style: self.getCellStyleId(plan.baseCell),
                    result: {
                        value: resultValue,
                        formula: resultFormula
                    }
                };
            }

            // creates a matrix for the range-autofill-operations
            function createMatrix(sourceRange, index) {

                // adds a cell-config-object to the matrix
                function addToMatrix(cell, row_i, col_i) {
                    var i = (columns) ? row_i : col_i;

                    if (!matrix[i]) { matrix[i] = { c: [] }; }
                    matrix[i].c.push(cell);
                }

                var // target range
                    targetRange     = createTargetRange(sourceRange, columns, count, reverse),
                    // the index of the foreach of the targetRange
                    i               = 0,
                    // the index of the sourceRange-loop
                    round           = 0,
                    // generate plan for iterating over targetRange with help of sourceRange
                    autoFillPlaning = analyzeForAutoFill(sourceRange, columns, reverse, invertType),
                    // the count of the lines from the sourceRange
                    sourceLineCount = sourceRange.size(!columns);

                // in reverse case, switch the base plan to be able to reuse the index-count
                if (reverse) { autoFillPlaning.reverse(); }

                // iterate over all single colums/rows to push new content to target-ranges
                IteratorUtils.forEach(targetRange.iterator({ reverse: reverse }), function () {
                    var sourceIndex     = (i % sourceLineCount),
                        plan            = autoFillPlaning[sourceIndex],
                        result          = null;

                    // count up the roundIndex, everytime sourceIndex is "0"
                    if (sourceIndex === 0) { round++; }

                    // calculate the actual autofill-result
                    result = calculateResult(plan, round, sourceLineCount);

                    // when there is a formula (and no new calculated value), use the old
                    // value for now
                    if (!_.isNull(result.result.formula) && result.result.value === 0) {
                        result.result.value = result.from.v;
                    }

                    // add result to the matrix
                    addToMatrix({
                        f: result.result.formula,
                        s: (!_.isNull(result.style)) ? result.style : null,
                        v: result.result.value
                    }, i, (index - startIndex));

                    i++;
                });
            }

            // creates a contents-array for the autofill-operations
            function createArray(sourceRange) {
                var contents        = [],
                    // target range
                    targetRange     = createTargetRange(sourceRange, columns, count, reverse),
                    // the index of the foreach of the targetRange
                    i               = 0,
                    // the index of the sourceRange-loop
                    round           = 0,
                    // generate plan for iterating over targetRange with help of sourceRange
                    autoFillPlaning = analyzeForAutoFill(sourceRange, columns, reverse, invertType),
                    // the count of the lines from the sourceRange
                    sourceLineCount = sourceRange.size(!columns);

                // in reverse case, switch the base plan to be able to reuse the index-count
                if (reverse) { autoFillPlaning.reverse(); }

                // iterate over all single colums/rows to push new content to target-ranges
                IteratorUtils.forEach(targetRange.iterator({ reverse: reverse }), function () {
                    var sourceIndex     = (i % sourceLineCount),
                        plan            = autoFillPlaning[sourceIndex],
                        result          = null;

                    // count up the roundIndex, everytime sourceIndex is "0"
                    if (sourceIndex === 0) { round++; }

                    // calculate the actual autofill-result
                    result = calculateResult(plan, round, sourceLineCount);

                    // the resulting cell-config-object
                    var cell = {
                        f: result.result.formula,
                        s: (!_.isNull(result.from)) ? result.from.s : null,
                        v: result.result.value
                    };

                    // push cell-config-object to contents-array
                    if (columns) {
                        contents.push({ c: [cell] });
                    } else {
                        if (!contents[0]) { contents.push({ c: [] }); }
                        contents[0].c.push(cell);
                    }

                    i++;
                });
                return contents;
            }

            var // column or row direction
                columns         = /^(top|bottom)$/.test(border),
                // reverse running or not
                reverse         = /^(left|top)$/.test(border),
                // should the type be inverted (copy to increase aso)
                invertType      = Utils.getBooleanOption(options, 'invertType', false),
                // start index of col/row (to create autofill ranges)
                startIndex      = columns ? range.start[0] : range.start[1],
                // complete column(s) selected
                completeCol     = docModel.isColRange(range),
                // complete row(s) selected
                completeRow     = docModel.isRowRange(range),
                // object which holds the index for each separate group within the sourceRange
                rangeIndex      = {},
                // col/row interval
                interval        = range.interval(!columns),
                parallelRanges  = [],
                matrix          = [];

            // if one or more complete col/row is selected for autofill
            if (completeRow || completeCol) {
                var arrContents = {};

                // generate ranges for the parallel iterator
                IteratorUtils.forEach(((!columns) ? colMatrix : rowMatrix).createIndexIterator(interval), function (index) {
                    if (completeRow) {
                        parallelRanges.push(Range.create(0, index, range.end[0], index));
                    } else {
                        parallelRanges.push(Range.create(index, 0, index, range.end[1]));
                    }
                });

                // generate contents-array for operations (with parallel-iterator)
                IteratorUtils.forEach(createParallelAddressIterator(parallelRanges, { type: 'defined' }), function (arrAddresses) {
                    var start    = (completeRow) ? new Address(arrAddresses[0][0], range.start[1]) : new Address(range.start[0], arrAddresses[0][1]),
                        end      = (completeRow) ? new Address(arrAddresses[0][0], range.end[1]) : new Address(range.end[0], arrAddresses[0][1]),
                        // range to work on currently
                        useRange = Range.createFromAddresses(start, end),
                        // get contents-array for range
                        contents = createArray(useRange);

                    // reverse contents-array if necessary
                    if (reverse) { ((!columns) ? contents[0].c : contents).reverse(); }

                    arrContents[createTargetRange(useRange.clone(), columns, count, reverse).start.toString()] = contents;
                });

                // generate cell operations (async)
                return self.iterateSliced(IteratorUtils.createObjectIterator(arrContents), function (contents, options) {
                    return self.generateCellContentOperations(generator, Address.parse(options.key), contents);
                });

            // if an normal range is selected for autofill
            } else {
                // iterate over all ranges and create the matrix
                IteratorUtils.forEach(range.interval(columns).iterator(), function (index) {
                    createMatrix(range.lineRange(columns, (index - startIndex)), index);
                });

                // reverse new cell values if necessary
                if (reverse) {
                    if (!columns) {
                        _.each(matrix, function (data) {
                            data.c.reverse();
                        });
                    } else {
                        matrix.reverse();
                    }
                }

                // generate the cell content operations
                return self.generateCellContentOperations(generator, createTargetRange(range, columns, count, reverse).start, matrix);
            }
        }

        /**
         * Returns the internal contents of this collection, needed for cloning
         * into another collection.
         *
         * @internal
         *  Used during clone construction. DO NOT CALL from external code!
         *
         * @returns {Object}
         *  The internal contents of this collection.
         */
        this._getCloneData = function () {
            return {
                modelMap: modelMap,
                usedRange: usedRange,
                sharedFormulaMap: sharedFormulaMap
            };
        };

        // operation implementations ------------------------------------------

        /**
         * Callback handler for the document operation 'copySheet'. Clones all
         * contents from the passed collection into this collection.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'copySheet' document operation.
         *
         * @param {CellCollection} collection
         *  The source collection whose contents will be cloned into this
         *  collection.
         *
         * @throws {OperationError}
         *  If applying the operation fails, e.g. if a required property is
         *  missing in the operation.
         */
        this.applyCopySheetOperation = function (context, collection) {

            // the internal contents of the source collection
            var cloneData = collection._getCloneData();

            // deeply clone the shared formulas
            sharedFormulaMap = cloneData.sharedFormulaMap.clone(function (sharedFormula) {
                return sharedFormula.clone();
            });

            // clone the cell models of the source collection
            cloneData.modelMap.forEach(function (cellModel) {
                implInsertCellModel(cellModel.clone(sheetModel));
            });

            // copy the cached bounding range of all existing cells
            if (cloneData.usedRange) {
                usedRange = cloneData.usedRange.clone();
            }
        };

        /**
         * Callback handler for the document operation 'changeCells'. Changes
         * the contents and auto-styles of the cells in this collection, and
         * triggers a 'change:cells' event for all changed cells.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'changeCells' document operation.
         *
         * @throws {OperationError}
         *  If applying the operation fails, e.g. if a required property is
         *  missing in the operation.
         */
        this.applyChangeCellsOperation = function (context) {

            // the address of the top-left cell
            var start = context.getAddress('start');
            // the cell contents matrix (array of row data objects)
            var matrix = context.getArr('contents');
            // the maximum column and row indexes in the sheet
            var maxCol = docModel.getMaxCol();
            var maxRow = docModel.getMaxRow();
            // the address of the current cell
            var currAddress = start.clone();
            // descriptor with the addresses of all changed cells
            var changeDesc = new ChangeDescriptor();
            // special handling for ODF
            var odf = docModel.getApp().isODF();

            // creates a cell contents object from the passed raw JSON data
            function createCellContents(jsonData, range) {

                // the contents object expected by the method updateCellModel()
                var contents = {};

                // update cell value (special handling for error codes needed)
                if ('v' in jsonData) {
                    context.ensure(!('e' in jsonData), 'unexpected cell error code');
                    var type = typeof jsonData.v;
                    context.ensure((jsonData.v === null) || (type === 'number') || (type === 'string') || (type === 'boolean'), 'invalid cell value type');
                    contents.v = jsonData.v;
                } else if ('e' in jsonData) {
                    context.ensure((typeof jsonData.e === 'string') && (jsonData.e.length > 0), 'invalid cell error code');
                    contents.v = formulaGrammar.getErrorCode(jsonData.e) || ErrorCode.NA;
                }

                // update cell formula
                if ('f' in jsonData) {
                    context.ensure((jsonData.f === null) || ((typeof jsonData.f === 'string') && (jsonData.f.length > 0)), 'invalid formula expression');
                    contents.f = jsonData.f;
                }

                // identifier of a shared formula
                if ('si' in jsonData) {
                    context.ensure((jsonData.si === null) || (typeof jsonData.si === 'number'), 'invalid shared formula identifier');
                    contents.si = jsonData.si;
                }

                // definition of a shared formula
                if ('sr' in jsonData) {
                    context.ensure(range.single(), 'repetition not allowed for anchor cell of a shared formula');
                    contents.sr = docModel.createRange(jsonData.sr);
                    if ((contents.sr !== null) && ('f' in contents) && ('si' in contents)) {
                        context.ensure(contents.sr, 'invalid bounding range for a shared formula');
                        sharedFormulaMap.insert(contents.si, new SharedFormulaModel(contents.si, contents.f, contents.sr, range.start));
                    }
                }

                // check the identifier and bounding range of a shared formula referred by the cell(s)
                if ('si' in contents && contents.si !== null) {
                    var sharedFormula = sharedFormulaMap.get(contents.si, null);
                    context.ensure(sharedFormula, 'invalid shared formula identifier');
//                    context.ensure(sharedFormula.range.contains(range), 'invalid bounding range in definition of shared formula');
                }

                // update cell auto-style
                if ('s' in jsonData) {
                    context.ensure(typeof jsonData.s === 'string', 'invalid auto-style identifier');
                    contents.s = jsonData.s;
                }

                if ('sa' in jsonData) {
                    contents.sa = jsonData.sa;
                }

                // fail safety: check number of changed cells (prevent freezing browser)
                if (range.cells() > SheetUtils.MAX_FILL_CELL_COUNT) {

                    // ODF import: warn but do not fail for formatted cells without contents
                    if (odf && ('s' in contents) && !('v' in contents) && !('f' in contents) && !('si' in contents) && !('sr' in contents)) {
                        Utils.warn('CellCollection.applyChangeCellsOperation(): Overflow error! Skipping formatting of ' + range.cells() + ' cells.');
                        contents = {};
                    } else {
                        context.error('too many cells');
                    }
                }

                return contents;
            }

            // collects changed cells by value, and by auto-style (performance: do not collect during import)
            var collectChangedAddresses = context.importing ? _.noop : function (address, changeFlags) {
                changeDesc.addAddress(address, changeFlags);
            };

            // check that the 'contents' operation property is an array of row objects
            context.ensure(matrix.every(_.isObject), 'invalid row entries in cell matrix');

            // process all rows in the contents array
            matrix.forEach(function (rowData, index) {

                // number of repetitions for this row entry
                var rowRepeat = Utils.getIntegerOption(rowData, 'r', 1);
                context.ensure(currAddress[1] + rowRepeat - 1 <= maxRow, 'too many rows in cell matrix');

                // skip entire rows without 'c' property
                if (!('c' in rowData)) {
                    currAddress[1] += rowRepeat;
                    return;
                }

                // the 'c' property may be the array index of another row data object with 'c' property being an array
                var rowContents = rowData.c;
                if (_.isNumber(rowContents)) {
                    context.ensure((0 <= rowContents) && (rowContents < index), 'invalid array index to row data');
                    rowContents = matrix[rowContents].c;
                }

                // now, the 'c' property must be an array
                context.ensure(_.isArray(rowContents), 'cell data must be an array');

                // process all cell entries in the row array
                rowContents.forEach(function (jsonData) {
                    context.ensure(_.isObject(jsonData), 'cell entry must be an object');

                    // number of repetitions for this cell
                    var colRepeat = Utils.getIntegerOption(jsonData, 'r', 1);

                    // check validity of the covered column interval
                    context.ensure(currAddress[0] + colRepeat - 1 <= maxCol, 'row element in cell contents too long');

                    // immediately skip cells that will not be modified at all (without loop)
                    if (CellOperationsBuilder.isEmptyJSON(jsonData)) {
                        currAddress[0] += colRepeat;
                        return;
                    }

                    // build the cell range covered by the current JSON entry, update the current address
                    var range = Range.createFromAddressAndSize(currAddress, colRepeat, rowRepeat);
                    currAddress[0] += colRepeat;

                    // explicitly delete cell models, if the 'u' property is set
                    if (jsonData.u === true) {
                        // the ranges MUST be iterated in reversed order, to not disturb the cell matrix iterators while deleting the models
                        IteratorUtils.forEach(createAddressIterator(range, { type: 'defined', reverse: true }), function (address) {
                            var changeFlags = deleteCellModel(address);
                            collectChangedAddresses(address, changeFlags);
                        });
                        return;
                    }

                    // the contents object expected by the method updateCellModel()
                    var contents = createCellContents(jsonData, range);
                    if (_.isEmpty(contents)) { return; }

                    // create/update the cell models, collect addresses of all changed cells
                    IteratorUtils.forEach(range.iterator(), function (address) {
                        var changeFlags = updateCellModel(address, contents);
                        collectChangedAddresses(address, changeFlags);
                    });
                });

                // update the current address
                currAddress[0] = start[0];
                currAddress[1] += rowRepeat;
            });

            // update the used area of the sheet
            updateUsedArea();

            // notify change event listeners
            if (!changeDesc.empty()) {
                var results = context.getOptStr('scope') === 'filter';
                this.trigger('change:cells', changeDesc, context.external, results);
            }
        };

        /**
         * Callback handler for the document operation 'moveCells'.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'moveCells' document operation.
         *
         * @throws {OperationError}
         *  If applying the operation fails, e.g. if a required property is
         *  missing in the operation.
         */
        this.applyMoveCellsOperation = function (context) {

            // the range to be inserted or cleared
            var range = context.getRange();
            // the move direction
            var direction = context.getEnum('dir', /^(left|right|up|down)$/);
            // whether to move the cells through columns
            var columns = (direction === 'left') || (direction === 'right');
            // whether to insert cells (move towards end of sheet)
            var insert = (direction === 'right') || (direction === 'down');
            // the move descriptor containing all needed data
            var moveDesc = new MoveDescriptor(docModel, range.interval(!columns), range.interval(columns), columns, insert);

            // first, delete all existing cell models in 'blankRange' (the cells MUST be iterated
            // in reversed order, to not disturb the cell matrix iterators while delting the models)
            var blankRange = moveDesc.createRange(moveDesc.deleteIntervals.first());
            var blankIterator = createAddressIterator(blankRange, { type: 'defined', reverse: true });
            IteratorUtils.forEach(blankIterator, function (address, result) {
                implRemoveCellModel(result.model);
            });

            // when deleting a cell range (move towards the beginning), there may not be anything left to move
            if (!moveDesc.moveFromIntervals.empty()) {

                // the source range containing cell models that need to be moved (may be empty)
                var moveRange = moveDesc.createRange(moveDesc.moveFromIntervals.first());
                // a temporary array used to collect moved cell models (prevents overwriting existing models)
                var tempModels = [];
                // the signed distance to modify the cell address index
                var moveDist = range.size(columns) * (insert ? 1 : -1);
                // column matrix does not need to be updated, when moving cells up/down;
                // or when moving the cells in all rows of the entire used area to the left/right
                var usedRowInterval = columns ? rowMatrix.getUsedInterval() : null;
                var skipColMatrix = !usedRowInterval || moveRange.rowInterval().contains(usedRowInterval);
                // row matrix does not need to be updated, when moving cells to left/right;
                // or when moving the cells in all columns of the entire used area up/down
                var usedColInterval = !columns ? colMatrix.getUsedInterval() : null;
                var skipRowMatrix = !usedColInterval || moveRange.colInterval().contains(usedColInterval);

                // move all cell models into the temporary map (the cells MUST be iterated in reversed
                // order, to not disturb the cell matrix iterator while removing the models)
                IteratorUtils.forEach(createAddressIterator(moveRange, { type: 'defined', reverse: true }), function (address, result) {

                    // remove the cell from the containers, before the address will be changed
                    var cellModel = result.model;
                    implRemoveCellModel(cellModel, skipColMatrix, skipRowMatrix);

                    // adjust the cell address, and store the cell in the temporary array
                    cellModel.a.move(moveDist, columns);
                    tempModels.push(cellModel);
                });

                // insert all moved cell models back into the containers
                tempModels.forEach(function (cellModel) {
                    implInsertCellModel(cellModel, skipColMatrix, skipRowMatrix);
                });
            }

            // update the used area of the sheet
            updateUsedArea();

            // notify move event listeners
            if (!context.importing) {
                this.trigger('move:cells', moveDesc, context.external);
            }
        };

        // protected methods --------------------------------------------------

        /**
         * Post-processing of the cell collection, after all import operations
         * have been applied successfully.
         *
         * @internal
         *  Called from the application import process. MUST NOT be called from
         *  external code.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the cell collection has been
         *  post-processed successfully; or rejected when any error has
         *  occurred.
         */
        this.postProcessCells = function () {

            // query token arrays of all formula cells, this causes to parse all formula expressions
            return this.iterateSliced(modelMap.iterator(), getCellModelTokenArray);
        };

        // public methods -----------------------------------------------------

        /**
         * Returns the current zero-based index of the sheet containing this
         * cell collection.
         *
         * @returns {Number}
         *  The current zero-based index of the sheet containing this cell
         *  collection.
         */
        this.getSheetIndex = function () {
            return sheetModel.getIndex();
        };

        /**
         * Returns the number of columns in the used area of the sheet.
         *
         * @returns {Number}
         *  The number of columns in the used area of the sheet, starting from
         *  cell A1. If the returned number is zero, this cell collection is
         *  completely empty.
         */
        this.getUsedCols = function () {
            return usedRange ? (usedRange.end[0] + 1) : 0;
        };

        /**
         * Returns the number of rows in the used area of the sheet.
         *
         * @returns {Number}
         *  The number of rows in the used area of the sheet, starting from
         *  cell A1. If the returned number is zero, this cell collection is
         *  completely empty.
         */
        this.getUsedRows = function () {
            return usedRange ? (usedRange.end[1] + 1) : 0;
        };

        /**
         * Returns the number of columns or rows in the used area of the sheet.
         *
         * @param {Boolean} columns
         *  Whether to return the number of used columns (true), or the number
         *  of used rows (false).
         *
         * @returns {Number}
         *  The number of columns/rows in the used area of the sheet. If the
         *  returned number is zero, this cell collection is completely empty.
         */
        this.getUsedCount = function (columns) {
            return columns ? this.getUsedCols() : this.getUsedRows();
        };

        /**
         * Returns the address of the last used cell in the sheet.
         *
         * @returns {Address}
         *  The address of the last used cell in the sheet, or the address A1,
         *  if this collection is empty.
         */
        this.getLastUsedCell = function () {
            return usedRange ? usedRange.end.clone() : Address.A1.clone();
        };

        /**
         * Returns the range address of the used area in the sheet.
         *
         * @returns {Range}
         *  The range address of the used area in the sheet, always starting at
         *  cell A1. If this collection is completely empty, returns a range
         *  address covering the cell A1 only.
         */
        this.getUsedRange = function () {
            return new Range(Address.A1.clone(), this.getLastUsedCell());
        };

        /**
         * Returns the identifier for the auto-style of the row or column to be
         * used for an undefined cell at the specified position.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @param {Boolean} [ignoreRowStyle=false]
         *  If set to true, an existing row auto-style will be ignored, and
         *  this method effectively returns the default column auto-style of
         *  the specified cell.
         *
         * @returns {String}
         *  The identifier of the row auto-style, if the specified row contains
         *  the 'customFormat' flag (and the ignore flag is not set), otherwise
         *  the identifier of the column auto-style.
         */
        this.getDefaultStyleId = function (address, ignoreRowStyle) {
            var rowDesc = ignoreRowStyle ? null : rowCollection.getEntry(address[1]);
            return (rowDesc && rowDesc.merged.customFormat) ? rowDesc.style : colCollection.getEntry(address[0]).style;
        };

        /**
         * Returns the internal cell model with the specified cell address from
         * the collection.
         *
         * @attention
         *  The method is provided for optimized access to internal cell data.
         *  The cell model MUST NOT BE CHANGED in order to retain the internal
         *  consistency of the cell collection!
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {CellModel|Null}
         *  The cell model; or null, if the cell does not exist in this cell
         *  collection. MUST NOT be changed!
         */
        this.getCellModel = function (address) {
            return modelMap.get(address.key(), null);
        };

        /**
         * Returns whether the specified cell is blank (no result value). A
         * blank cell may contain formatting attributes.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Boolean}
         *  Whether the specified cell is blank.
         */
        this.isBlankCell = function (address) {
            return !CellModel.hasValue(this.getCellModel(address));
        };

        /**
         * Returns whether the specified cell is a formula cell.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Boolean}
         *  Whether the specified cell is a formula cell.
         */
        this.isFormulaCell = function (address) {
            return CellModel.hasFormula(this.getCellModel(address));
        };

        /**
         * Returns whether the specified cell is a formula or shared formula cell or ...formula cell.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Boolean}
         *  Whether the specified cell is a formula cell.
         */
        this.isAnyFormulaCell = function (address) {
            return CellModel.hasAnyFormula(this.getCellModel(address));
        };

        /**
         * Returns whether the specified cell is visible (located in a visible
         * column, AND a visible row).
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Boolean}
         *  Whether the specified cell is visible.
         */
        this.isVisibleCell = function (address) {
            return colCollection.isEntryVisible(address[0]) && rowCollection.isEntryVisible(address[1]);
        };

        /**
         * Returns the value (or formula result) of the specified cell.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Number|String|Boolean|ErrorCode|Null}
         *  The current value (or formula result) of the specified cell.
         */
        this.getCellValue = function (address) {
            var cellModel = this.getCellModel(address);
            return cellModel ? cellModel.v : null;
        };

        /**
         * Returns the default display string of the specified cell.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {String|Null}
         *  The default display string of the specified cell (regardless of the
         *  column width); or null, if the cell cannot display its value (e.g.
         *  due to an invalid number format).
         */
        this.getCellDisplayString = function (address) {
            var cellModel = this.getCellModel(address);
            return cellModel ? cellModel.d : '';
        };

        /**
         * Returns the token array of the formula cell at the specified
         * address.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {TokenArray|Null}
         *  The token array of the specified cell; or null, if the cell does
         *  not contain a formula.
         */
        this.getCellTokenArray = function (address) {
            return getCellModelTokenArray(this.getCellModel(address));
        };

        /**
         * Returns the formula expression of the cell at the specified address.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @param {String} grammarId
         *  The identifier of the formula grammar for the formula expression.
         *
         * @returns {String|Null}
         *  The formula expression of the specified cell; or null, if the cell
         *  does not contain a formula.
         */
        this.getCellFormula = function (address, grammarId) {

            // no formulas in undefined cells
            var cellModel = this.getCellModel(address);
            if (!cellModel) { return null; }

            // resolve the formula expression of a shared formula
            if (cellModel.isSharedFormula()) {
                var sharedFormulaDesc = sharedFormulaMap.get(cellModel.si);
                if (!sharedFormulaDesc) {
                    Utils.error('CellCollection.getCellFormula(): missing shared formula definition');
                    return null;
                }
                var tokenArray = new TokenArray(sheetModel, 'cell');
                tokenArray.parseFormula('op', sharedFormulaDesc.formula, { refAddress: sharedFormulaDesc.anchor });
                return tokenArray.getFormula(grammarId, { refAddress: sharedFormulaDesc.anchor, targetAddress: address });
            }

            // resolve the formula expression directly from the cell model
            return cellModel.getFormula(sheetModel, grammarId);
        };

        /**
         * Returns the identifier of the effective auto-style used by the cell
         * at the specified address.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {String}
         *  The identifier of the cell auto-style, if the cell is defined,
         *  otherwise the default auto-style of the row or column.
         */
        this.getCellStyleId = function (address) {
            var cellModel = this.getCellModel(address);
            return cellModel ? cellModel.s : this.getDefaultStyleId(address);
        };

        /**
         * Returns the effective merged attribute set of the cell at the
         * specified address.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Object}
         *  The merged attribute set of the cell, if it is defined, otherwise
         *  the attribute set of the default auto-style of the row or column.
         */
        this.getCellAttributeSet = function (address) {
            return autoStyles.getMergedAttributeSet(this.getCellStyleId(address));
        };

        /**
         * Returns the effective text orientation settings of the cell at the
         * specified address.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {Object}
         *  A descriptor containing various text orientation properties. See
         *  method SheetUtils.getTextOrientation() for details.
         */
        this.getCellTextOrientation = function (address) {
            var cellModel = this.getCellModel(address);
            var attrSet = this.getCellAttributeSet(address);
            return SheetUtils.getTextOrientation(cellModel ? cellModel.v : null, attrSet.cell.alignHor, cellModel ? cellModel.d : '');
        };

        /**
         * Returns the effective parsed number format of the cell at the
         * specified address.
         *
         * @param {Address} address
         *  The address of the cell.
         *
         * @returns {ParsedFormat}
         *  The parsed number format of the cell, if it is defined, otherwise
         *  the parsed format of the default auto-style of the row or column.
         */
        this.getCellParsedFormat = function (address) {
            return autoStyles.getParsedFormat(this.getCellStyleId(address));
        };

        /**
         * Returns the URL of a hyperlink as returned by a cell formula (via
         * the function HYPERLINK) at the specified address.
         *
         * @param {Address} address
         *  The address of a cell.
         *
         * @returns {String|Null}
         *  The URL of a hyperlink as returned by a cell formula (via the
         *  function HYPERLINK) at the specified address.
         */
        this.getFormulaURL = function (address) {
            var cellModel = this.getCellModel(address);
            return cellModel ? cellModel.url : null;
        };

        /**
         * Returns the effective URL of a hyperlink at the specified address.
         * If the cell contains a regular hyperlink, and a cell formula with a
         * HYPERLINK function, the regular hyperlink will be preferred. See
         * description of the methods CellCollection.getCellURL(), and
         * CellCollection.getFormulaURL() for more details.
         *
         * @param {Address} address
         *  The address of a cell.
         *
         * @returns {String|Null}
         *  The effective URL of a hyperlink at the specified address.
         */
        this.getEffectiveURL = function (address) {
            var hyperlinkCollection = sheetModel.getHyperlinkCollection();
            var cellURL = hyperlinkCollection.getCellURL(address);
            return _.isString(cellURL) ? cellURL : this.getFormulaURL(address);
        };

        // range iterators ----------------------------------------------------

        /**
         * Creates an iterator that generates the cell addresses for all, or
         * specific, cells contained in the passed cell ranges.
         *
         * @param {RangeArray|Range} ranges
         *  An array of range addresses, or a single cell range address, used
         *  to generate cell addresses from.
         *
         * @param {Object} [options]
         *  Optional parameters. Supports all options supported by the method
         *  getIteratorRanges() to define a custom start position inside the
         *  passed cell ranges (i.e. the options 'reverse', 'startIndex',
         *  'startAddress', 'skipStart', and 'wrap'), and the following
         *  additional options:
         *  @param {String} [options.type='all']
         *      Specifies which type of cells will be covered by the iterator.
         *      Must be one of the following values:
         *      - 'all': Generates the addresses for all defined and undefined
         *          cells in the passed ranges.
         *      - 'defined': Generates the addresses for all defined cells that
         *          really exist in this collection, regardless of their
         *          content value (including blank but formatted cells).
         *      - 'value': Generates the addresses for all non-blank cells
         *          (with any content value), regardless if the value is given
         *          literally, or is the result of a formula.
         *      - 'formula': Generates the addresses of formula cells only.
         *  @param {Boolean|String} [options.visible=false]
         *      Specifies how to handle hidden columns and rows:
         *      - true (boolean): Only visible cells (cells in visible columns,
         *          AND visible rows) will be covered by the iterator.
         *      - 'columns': Cells in hidden columns will be skipped, but cells
         *          in visible columns and hidden rows will be visited.
         *      - 'rows': Cells in hidden rows will be skipped, but cells in
         *          visible rows and hidden columns will be visited.
         *      - false (boolean, default): By default, all visible and hidden
         *          cells in the passed cell ranges will be covered.
         *  @param {Boolean} [options.covered=false]
         *      If set to true, visits all cells covered and hidden by a merge
         *      range. By default, covered cells in merged ranges will not be
         *      visited.
         *  @param {Boolean} [options.reverse=false]
         *      If set to true, the ranges AND the cell addresses in each range
         *      will be processed in reversed order.
         *
         * @returns {Object}
         *  An iterator object with the method next(). The result object of the
         *  iterator will contain the following value properties on success:
         *  - {Address} value
         *      The address of the cell currently visited.
         *  - {Range} orig
         *      The address of the original cell range (from the passed range
         *      array) containing the current cell address.
         *  - {Number} index
         *      The array index of the original cell range contained in the
         *      result property 'orig'.
         */
        this.createAddressIterator = function (ranges, options) {

            // collection of merged ranges in the active sheet
            var mergeCollection = sheetModel.getMergeCollection();
            // whether to visit visible columns and/or rows only
            var visible = getVisibleMode(options);
            // the shrink method for merged ranges
            var shrinkMethod = visible.cols ? (visible.rows ? 'both' : 'columns') : (visible.rows ? 'rows' : null);
            // whether to visit cells covered (hidden) by merged ranges
            var covered = Utils.getBooleanOption(options, 'covered', false);
            // whether to iterate in reversed order
            var reverse = Utils.getBooleanOption(options, 'reverse', false);

            // returns the appropriate address iterator for the passed cell range
            function createIteratorForRange(range) {

                // create an iterator that visits the cell addresses in the range
                var iterator = createAddressIterator(range, options);
                if (covered || mergeCollection.isEmpty()) { return iterator; }

                // returns the merged ranges overlapping the current range on first invocation (for performance:
                // if the range does not contain any matching cells, the merged ranges will not be collected at all)
                var getMergedRanges = _.once(function () {

                    // the merged ranges covering the cell range currently iterated
                    var mergedRanges = mergeCollection.getMergedRanges(range);

                    // shrink merged ranges to their visible parts in visible cells mode
                    if (shrinkMethod && !mergedRanges.empty()) {
                        mergedRanges = RangeArray.map(mergedRanges, function (mergedRange) {
                            mergedRange = sheetModel.shrinkRangeToVisible(mergedRange, shrinkMethod);
                            // remove resulting single-cell ranges from the array
                            return (mergedRange && !mergedRange.single()) ? mergedRange : null;
                        });
                    }

                    return mergedRanges;
                });

                // filter cells that are covered by merged ranges
                return IteratorUtils.createFilterIterator(iterator, function (address) {
                    var mergedRange = getMergedRanges().findByAddress(address);
                    return !mergedRange || mergedRange.startsAt(address);
                });
            }

            // split and reorder the passed ranges according to the passed options (the new ranges will contain
            // the additional properties 'orig' and 'index'), and create an array iterator for these ranges
            var arrayIt = getIteratorRanges(ranges, options).iterator({ reverse: reverse });

            // create the resulting iterator, combining the array iterator with an inner address iterator for each range
            return IteratorUtils.createNestedIterator(arrayIt, createIteratorForRange, function (arrayResult, rangeResult) {
                // the range from the range array iterator (contains original range and index)
                var iterRange = arrayResult.value;
                return { value: rangeResult.value, orig: iterRange.orig, index: iterRange.index };
            });
        };

        /**
         * Creates an iterator that generates the cell addresses for all, or
         * specific, cells contained in a single column or row, while moving
         * away from the start cell into the specified directions.
         *
         * @param {Address} address
         *  The address of the cell to start the iteration process from (the
         *  option 'skipStart' can be used, if iteration shall start with the
         *  nearest available neighbor of the start cell, and not with the cell
         *  itself).
         *
         * @param {String} directions
         *  Specifies how to move to the next cell addresses while iterating.
         *  Supported values are 'up', 'down', 'left', or 'right', or any space
         *  separated list of these values. Multiple directions will be
         *  processed in the same order as specified in this parameter.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {String} [options.type='all']
         *      Specifies which type of cells will be covered by the iterator.
         *      See method CellCollection.createAddressIterator() for details.
         *  @param {Boolean|String} [options.visible=false]
         *      Specifies how to handle hidden columns and rows. See method
         *      CellCollection.createAddressIterator() for details.
         *  @param {Boolean} [options.covered=false]
         *      If set to true, visits all cells covered and hidden by a merge
         *      range. By default, covered cells in merged ranges will not be
         *      visited.
         *  @param {Range} [options.boundRange]
         *      If specified, the line iterator will be restricted to that cell
         *      range. If omitted, the entire area of the sheet will be used.
         *  @param {Boolean} [options.skipStart=false]
         *      If set to true, the iterator will start at the nearest neighbor
         *      of the cell specified in the 'address' parameter instead of
         *      that cell (nearest visible cell, if option 'visible' has been
         *      set).
         *
         * @returns {Object}
         *  An iterator object with the method next(). The result object of the
         *  iterator will contain the following value properties on success:
         *  - {Address} value
         *      The daddress of the cell currently visited.
         *  - {String} direction
         *      The position of the current cell, relative to the start cell
         *      (one of the directions passed to the parameter 'direction' of
         *      this method).
         */
        this.createLinearAddressIterator = function (address, directions, options) {

            // the bounding range to restrict the visited area
            var boundRange = Utils.getObjectOption(options, 'boundRange', docModel.getSheetRange());
            // number of cells to eb skipped before iteration starts
            var skipCount = Utils.getBooleanOption(options, 'skipStart', false) ? 1 : 0;

            // prepare the cell ranges to be visited
            var ranges = RangeArray.map(directions.split(/\s+/), function (direction) {

                // whether to iterate with variable column index
                var columns = /^(left|right)$/.test(direction);
                // whether to iterate in reversed order (towards first column/row)
                var reverse = /^(up|left)$/.test(direction);

                // check passed direction identifier
                if (!columns && !reverse && (direction !== 'down')) {
                    Utils.error('CellCollection.createLinearAddressIterator(): invalid direction "' + direction + '"');
                    return null;
                }

                // the entire column/row range containing the passed address
                var range = docModel.makeFullRange(address.get(!columns), !columns);

                // reduce range to leading or trailing part, according to start address and direction
                if (reverse) {
                    range.setEnd(address.get(columns) - skipCount, columns);
                } else {
                    range.setStart(address.get(columns) + skipCount, columns);
                }

                // skip the current direction, if the range becomes invalid (e.g. 'up' in first row);
                // reduce the range to the bounding range (this may invalidate the range too)
                range = (range.getStart(columns) <= range.getEnd(columns)) ? boundRange.intersect(range) : null;
                if (!range) { return null; }

                // put additional information into the range
                range.direction = direction;
                range.columns = columns;
                range.reverse = reverse;
                return range;
            });

            // create the nested iterator, combining an outer array iterator with an inner address iterator for each range
            return IteratorUtils.createNestedIterator(ranges.iterator(), function (range) {
                return self.createAddressIterator(range, _.extend({}, options, { reverse: range.reverse }));
            }, function (outerResult, innerResult) {
                innerResult.direction = outerResult.value.direction;
                return innerResult;
            });
        };

        /**
         * Creates an iterator that visits the equally formatted parts in the
         * passed cell ranges.
         *
         * @param {RangeArray|Range} ranges
         *  An array of range addresses, or a single cell range address, to be
         *  visited by the iterator.
         *
         * @param {Object} [options]
         *  Optional parameters. Supports all options supported by the method
         *  getIteratorRanges() to define a custom start position inside the
         *  passed cell ranges (i.e. the options 'reverse', 'startIndex',
         *  'startAddress', 'skipStart', and 'wrap'), and the following
         *  additional options:
         *  @param {Boolean} [options.visible=false]
         *      If set to true, only visible cells (cells in visible columns,
         *      AND visible rows) will be covered by the iterator. By default,
         *      all visible and hidden cells will be covered.
         *  @param {Boolean} [options.covered=false]
         *      If set to true, visits all cells covered and hidden by a merge
         *      range. By default, covered cells in merged ranges will not be
         *      visited.
         *  @param {Boolean} [options.reverse=false]
         *      If set to true, the ranges AND the cell addresses in each range
         *      will be processed in reversed order.
         *
         * @returns {Object}
         *  An iterator object with the method next(). The result object of the
         *  iterator will contain the following value properties on success:
         *  - {Range} value
         *      The current cell range address.
         *  - {String} style
         *      The auto-style identifier for all cells in the range.
         *  - {Range} orig
         *      The address of the original cell range (from the passed range
         *      array) containing the current cell range.
         *  - {Number} index
         *      The array index of the original cell range contained in the
         *      result property 'orig'.
         */
        this.createStyleIterator = function (ranges, options) {

            // collection of merged ranges in the active sheet
            var mergeCollection = sheetModel.getMergeCollection();
            // whether to visit visible cells only
            var visible = Utils.getBooleanOption(options, 'visible', false);
            // whether to visit cells covered (hidden) by merged ranges
            var covered = Utils.getBooleanOption(options, 'covered', false);
            // whether to iterate in reversed order
            var reverse = Utils.getBooleanOption(options, 'reverse', false);

            // Creates a style iterator for a single cell. Used as short-cut for single cells to improve
            // performance, to prevent to create all the column/row intervals, sub-iterators, etc.
            function createStyleIteratorForAddress(address) {
                var skipCell = !covered && mergeCollection.isHiddenCell(address);
                return skipCell ? IteratorUtils.ALWAYS_DONE_ITERATOR : IteratorUtils.createSingleIterator({
                    value: new Range(address.clone()),
                    style: self.getCellStyleId(address)
                });
            }

            // Creates an iterator that tries to combine existing equally formatted cells into sub-ranges.
            // The result objects of the iterator will contain the following properties:
            // - {IntervalArray} colIntervals
            //      The column intervals of all equally formatted ranges of the iterator step; each
            //      interval in the array with the additional property 'style'.
            // - {Interval} rowInterval
            //      The row interval containing all equally formatted ranges of the iterator step.
            function createIteratorForExistingCells(range) {

                // creates a column interval for the cell in the passed result object of a cell model iterator
                function createColInterval(result) {
                    var colInterval = new Interval(result.model.a[0]);
                    colInterval.style = result.model.s;
                    return colInterval;
                }

                // creates and initializes the property 'colIntervals' in the passed result if missing
                function initializeColIntervals(result) {
                    return result.colIntervals || (result.colIntervals = new IntervalArray(createColInterval(result)));
                }

                // creates and initializes the property 'rowInterval' in the passed result if missing
                function initializeRowInterval(result) {
                    return result.rowInterval || (result.rowInterval = new Interval(result.model.a[1]));
                }

                // an iterator that visits all existing cells in the range
                var iterator = createAddressIterator(range, { type: 'defined', visible: visible, reverse: reverse });

                // filter for existing cells with an auto-style that differs from its active column/row style
                iterator = IteratorUtils.createFilterIterator(iterator, function (address, result) {
                    return !autoStyles.areEqualStyleIds(result.model.s, self.getDefaultStyleId(address));
                });

                // a combining iterator that reduces all existing formatted cells of a single row into a column interval array
                // with auto-style identifiers (inserted as new property 'colIntervals' into the iterator result)
                iterator = IteratorUtils.createReduceIterator(iterator, function (result1, result2) {

                    // immediately return if the new cell model is located in another row (iterator will actually step to 'result1')
                    var cellModel = result2.model;
                    if (result1.model.a[1] !== cellModel.a[1]) { return; }

                    // get or create the column interval array (will be initialized with column index of result1 if missing)
                    var colIntervals = initializeColIntervals(result1);

                    // try to extend the interval array (adjacent cell with equal formatting)
                    var colInterval = reverse ? colIntervals.first() : colIntervals.last();
                    if (autoStyles.areEqualStyleIds(colInterval.style, cellModel.s)) {
                        if (!reverse && (colInterval.last + 1 === cellModel.a[0])) {
                            colInterval.last += 1;
                            return result1;
                        }
                        if (reverse && (cellModel.a[0] + 1 === colInterval.first)) {
                            colInterval.first -= 1;
                            return result1;
                        }
                    }

                    // create an interval for the new cell (push or unshift according to direction)
                    colInterval = createColInterval(result2);
                    if (reverse) { colIntervals.unshift(colInterval); } else { colIntervals.push(colInterval); }
                    return result1;
                });

                // a combining iterator that tries to collapse multiple rows with equally formatted existing cells
                iterator = IteratorUtils.createReduceIterator(iterator, function (result1, result2) {

                    // initialize the row interval in the first iterator result if missing
                    var rowInterval = initializeRowInterval(result1);

                    // the second iterator result must represent a row that can extend the row interval
                    if (reverse ? (result2.model.a[1] + 1 !== rowInterval.first) : (rowInterval.last + 1 !== result2.model.a[1])) { return null; }

                    // create missing column intervals in both result objects for comparison
                    var colIntervals1 = initializeColIntervals(result1);
                    var colIntervals2 = initializeColIntervals(result2);

                    // compare the positions and auto-style identifiers of both column interval arrays
                    var equalIntervals = (colIntervals1.length === colIntervals2.length) && colIntervals1.every(function (colInterval1, index) {
                        var colInterval2 = colIntervals2[index];
                        return colInterval1.equals(colInterval2) && autoStyles.areEqualStyleIds(colInterval1.style, colInterval2.style);
                    });

                    // column intervals are equal: expand the row interval, and continue to search for adjacent rows
                    if (equalIntervals) {
                        if (reverse) { rowInterval.first -= 1; } else { rowInterval.last += 1; }
                        return result1;
                    }
                });

                // ensure that all iterator results contain column and row intervals (reduce iterators may skip single entries)
                return IteratorUtils.createTransformIterator(iterator, function (value, result) {
                    initializeColIntervals(result);
                    initializeRowInterval(result);
                    return result;
                });
            }

            // returns an interval array for the passed column interval (or the visible parts) to be used for rows with
            // an own active auto-style (rows with 'customFormat' flag set to true), and caches it for the iterator
            var getColIntervals = _.memoize(function (colInterval, rowStyleId) {
                var colIntervals = visible ? colCollection.getVisibleIntervals(colInterval) : new IntervalArray(colInterval.clone());
                return Utils.addProperty(colIntervals, 'style', rowStyleId);
            }, function (colInterval, rowStyleId) { return colInterval.toString() + ':' + rowStyleId; });

            // collect the default column auto-styles of the column interval to be used for unformatted rows that show
            // the default column auto-styles (rows with 'customFormat' flag set to false), and caches it for the iterator
            var getColStyleIntervals = _.memoize(function (colInterval) {
                var iterator = colCollection.createStyleIterator(colInterval, { visible: visible });
                return IteratorUtils.reduce(new IntervalArray(), iterator, function (intervals, interval, result) {
                    interval.style = result.style;
                    intervals.push(interval);
                    return intervals;
                });
            }, function (colInterval) { return colInterval.toString(); });

            // collect the row bands of the merged ranges covering the passed cell range
            function getMergedRowBands(range) {

                // get merged ranges in the current range
                var mergedRanges = mergeCollection.getMergedRanges(range);
                if (mergedRanges.empty()) { return null; }

                // skip all cells but the top-left cell of the merged ranges
                var skipRanges = new RangeArray();
                mergedRanges.forEach(function (mergedRange) {
                    if (!mergedRange.singleCol()) { skipRanges.push(Range.create(mergedRange.start[0] + 1, mergedRange.start[1], mergedRange.end[0], mergedRange.start[1])); }
                    if (!mergedRange.singleRow()) { skipRanges.push(Range.create(mergedRange.start[0], mergedRange.start[1] + 1, mergedRange.end[0], mergedRange.end[1])); }
                });

                // reduce the ranges to the visible areas if the 'visible' flag is set, reduce to the passed cell range address
                if (visible) { skipRanges = RangeArray.map(skipRanges, sheetModel.shrinkRangeToVisible.bind(sheetModel)); }
                skipRanges = skipRanges.intersect(range);
                if (skipRanges.empty()) { return null; }

                // return the row bands containing the column intervals
                return skipRanges.getRowBands({ intervals: true });
            }

            // reduce the passed column intervals with auto-style identifiers to the parts not covered by merged ranges
            function reduceStyleColIntervals(styleColIntervals, skipColIntervals) {
                var partColIntervals = styleColIntervals.concat(skipColIntervals).partition();
                return partColIntervals.reject(function (partColInterval) {
                    return partColInterval.coveredBy.some(function (origColInterval) {
                        if (!('style' in origColInterval)) { return true; }
                        partColInterval.style = origColInterval.style;
                        delete partColInterval.coveredBy;
                    });
                });
            }

            // returns a style iterator for the passed equally formatted and completely visivle row interval,
            // and the entire column interval (with hidden columns that need to be ignored in visible mode)
            function createStyleIteratorForRowInterval(colInterval, rowInterval, rowStyleId) {

                // short-cut for single cells (do not create all the column/row intervals etc.)
                if (colInterval.single() && rowInterval.single()) {
                    return createStyleIteratorForAddress(new Address(colInterval.first, rowInterval.first));
                }

                // an iterator that provides row intervals with existing, equally-formatted cells (as column intervals)
                var range = Range.createFromIntervals(colInterval, rowInterval);
                // the column intervals with auto-style identifiers (use column auto-styles, if the row interval is not formatted)
                var colIntervals = (rowStyleId === null) ? getColStyleIntervals(colInterval) : getColIntervals(colInterval, rowStyleId);
                // an iterator that provides row intervals with existing, equally-formatted cells (as column intervals)
                var cellsIt = createIteratorForExistingCells(range);
                // the iterator provides settings for a row interval with existing formatted cells
                var cellsResult = cellsIt.next();
                // the row bands of the merged ranges covering the current range to be skipped
                var mergedRowBands = covered ? null : getMergedRowBands(range);
                // the current row index (for the gaps between row intervals of existing cells)
                var currRow = reverse ? rowInterval.last : rowInterval.first;

                // visit the row intervals with existing cells, and the gaps between these row intervals (blank rows)
                var rowIntervalIt = IteratorUtils.createWhileIterator(function () {

                    // the iterator is done if the current row moves outside the processed row interval
                    if (!rowInterval.containsIndex(currRow)) { return { done: true }; }

                    // the next row interval to be visited (either a gap between cells, or a row interval with cells)
                    var currRowInterval = null;
                    // the column intervals with style identifiers
                    var styleColIntervals = null;

                    // find the next row interval to be processed, and build the column intervals with auto-style identifiers
                    if (cellsResult.done) {
                        // no more formatted cells available: process the remaining rows to the end (reverse mode: to the beginning) of the entire row interval
                        currRowInterval = reverse ? new Interval(rowInterval.first, currRow) : new Interval(currRow, rowInterval.last);
                        styleColIntervals = colIntervals;
                    } else if (!cellsResult.rowInterval.containsIndex(currRow)) {
                        // 'currRow' is above the next formatted cells (reverse mode: below the cells): process the gap of blank rows
                        currRowInterval = reverse ? new Interval(cellsResult.rowInterval.last + 1, currRow) : new Interval(currRow, cellsResult.rowInterval.first - 1);
                        styleColIntervals = colIntervals;
                    } else {
                        // process a row interval that contains existing formatted cells
                        currRowInterval = reverse ? new Interval(cellsResult.rowInterval.first, currRow) : new Interval(currRow, cellsResult.rowInterval.last);
                        styleColIntervals = colCollection.mergeStyleIntervals(colIntervals, cellsResult.colIntervals);
                    }

                    // shorten the current row interval according to the next merged row band
                    var nextMergedRowBand = !mergedRowBands ? null : reverse ? mergedRowBands.last() : mergedRowBands.first();
                    if (nextMergedRowBand) {
                        if (nextMergedRowBand.containsIndex(currRow)) {
                            // current row interval starts in a row band with merged ranges
                            if (reverse) {
                                currRowInterval.first = Math.max(currRowInterval.first, nextMergedRowBand.first);
                            } else {
                                currRowInterval.last = Math.min(currRowInterval.last, nextMergedRowBand.last);
                            }
                            // reduce the column intervals to the parts that are not covered by merged ranges
                            styleColIntervals = reduceStyleColIntervals(styleColIntervals, nextMergedRowBand.intervals);
                        } else {
                            // current row interval starts before a row band with merged ranges (reverse mode: after a merged row band)
                            if (reverse) {
                                currRowInterval.first = Math.max(currRowInterval.first, nextMergedRowBand.last + 1);
                            } else {
                                currRowInterval.last = Math.min(currRowInterval.last, nextMergedRowBand.first - 1);
                            }
                        }
                    }

                    // index of the next row to be processed
                    currRow = reverse ? (currRowInterval.first - 1) : (currRowInterval.last + 1);

                    // fetch next cell result, if the row interval of the current result has been processed
                    if (!cellsResult.done && (reverse ? (currRow < cellsResult.rowInterval.first) : (cellsResult.rowInterval.last < currRow))) {
                        cellsResult = cellsIt.next();
                    }

                    // delete the row band for merged columns, if it has been processed
                    if (nextMergedRowBand && (reverse ? (currRow < nextMergedRowBand.first) : (nextMergedRowBand.last < currRow))) {
                        if (reverse) { mergedRowBands.pop(); } else { mergedRowBands.shift(); }
                    }

                    return { value: currRowInterval, colIntervals: styleColIntervals };
                });

                return IteratorUtils.createNestedIterator(rowIntervalIt, function (subRowInterval, result) {
                    var arrayIt = IteratorUtils.createArrayIterator(result.colIntervals, { reverse: reverse });
                    return IteratorUtils.createTransformIterator(arrayIt, function (subColInterval) {
                        return { value: Range.createFromIntervals(subColInterval, subRowInterval), style: subColInterval.style };
                    });
                });
            }

            // creates an iterator that visits all equally formatted sub-ranges in the passed range
            function createStyleIteratorForRange(range) {

                // Reduce the passed range to the visible area if the 'visible' flag is set. Continue with
                // the next cell range, if no more visible cells are left (the range shrinks to null).
                if (visible && !(range = sheetModel.shrinkRangeToVisible(range))) { return null; }
                // short-cut for single cells (do not create all the column/row intervals etc.)
                if (range.single()) { return createStyleIteratorForAddress(range.start); }

                // create a nested iterator that returns the results of a style iterator for each row interval
                var rowStyleIt = rowCollection.createStyleIterator(range.rowInterval(), options);
                return IteratorUtils.createNestedIterator(rowStyleIt, function (rowInterval, result) {
                    return createStyleIteratorForRowInterval(range.colInterval(), rowInterval, result.style);
                });
            }

            // split and reorder the passed ranges according to the passed options (the new ranges will contain
            // the additional properties 'orig' and 'index'), and create an array iterator for these ranges
            var arrayIt = getIteratorRanges(ranges, options).iterator({ reverse: reverse });

            // create the resulting iterator, combining the array iterator with an inner style iterator for each range
            return IteratorUtils.createNestedIterator(arrayIt, createStyleIteratorForRange, function (arrayResult, styleResult) {
                // the range from the range array iterator (contains original range and index)
                var range = arrayResult.value;
                return { value: styleResult.value, style: styleResult.style, orig: range.orig, index: range.index };
            });
        };

        /**
         * Returns an iterator that provides the subtotal results of specific
         * subranges and/or cells contained in the passed ranges.
         *
         * @param {RangeArray|Range} ranges
         *  An array of range addresses, or a single cell range address, whose
         *  subtotals will be returned.
         *
         * @returns {SubtotalResult|jQuery.Promise}
         *  The collected subtotal results of the cell ranges in synchronous
         *  mode (option 'async' set to false), or an abortable promise that
         *  will be resolved after the background loop has finished collecting
         *  the subtotals.
         */
        this.createSubtotalIterator = function (ranges) {

            // restrict the used range to its visible part (subtotals are collected from visible cells only)
            var visibleUsedRange = usedRange ? sheetModel.shrinkRangeToVisible(usedRange) : null;
            // restrict the passed ranges to the visible used range of this collection
            ranges = visibleUsedRange ? RangeArray.get(ranges).intersect(visibleUsedRange) : null;

            // return immediately if the ranges do not cover any existing cells
            if (!ranges || ranges.empty()) { return IteratorUtils.createSingleIterator(new SubtotalResult()); }

            // get the column intervals of the ranges covering the used range vertically
            var colIntervals = RangeArray.filter(ranges, function (range) {
                return (range.start[1] === usedRange.start[1]) && (usedRange.end[1] === range.end[1]);
            }).colIntervals().merge();

            // get the row intervals of the ranges covering the used range horizontally
            var rowIntervals = RangeArray.filter(ranges, function (range) {
                return (range.start[0] === usedRange.start[0]) && (usedRange.end[0] === range.end[0]);
            }).rowIntervals().merge();

            // decide whether to use the cached subtotals of the column or row intervals (more covered cells)
            var coveredCellsByCols = colIntervals.size() * usedRange.rows();
            var coveredCellsByRows = rowIntervals.size() * usedRange.cols();
            if (coveredCellsByRows < coveredCellsByCols) {
                rowIntervals.splice(0);
                ranges = ranges.difference(RangeArray.createFromIntervals(colIntervals, usedRange.rowInterval()));
            } else {
                colIntervals.splice(0);
                ranges = ranges.difference(RangeArray.createFromIntervals(usedRange.colInterval(), rowIntervals));
            }

            // Get the subtotal results of the column/row intervals from the matrix (cached internally), and remove
            // the column or row ranges from the passed ranges (the remaining ranges will be iterated individually).
            var matrixIt = null;
            if (!colIntervals.empty()) {
                matrixIt = colMatrix.createSubtotalIterator(colIntervals);
                ranges = ranges.difference(RangeArray.createFromIntervals(colIntervals, usedRange.rowInterval()));
            } else if (!rowIntervals.empty()) {
                matrixIt = rowMatrix.createSubtotalIterator(rowIntervals);
                ranges = ranges.difference(RangeArray.createFromIntervals(usedRange.colInterval(), rowIntervals));
            }

            // Create a cell iterator (do not collect the subtotals of a cell multiple times, always visit the cells
            // covered by merged ranges which is faster and matches the behavior of Excel where covered cells are
            // always empty, and OpenOffice where covered cells are always counted into the subtotals).
            var cellIt = null;
            if (!ranges.empty()) {
                cellIt = createAddressIterator(ranges.merge(), { type: 'value', visible: true });
                cellIt = IteratorUtils.createTransformIterator(cellIt, function (address, result) { return { value: result.model.v }; });
            }

            // create a combined iterator that collects cached subtotals, and separate cell values
            return (matrixIt && cellIt) ? IteratorUtils.createCombinedIterator(matrixIt, cellIt) : (matrixIt || cellIt || IteratorUtils.ALWAYS_DONE_ITERATOR);
        };

        // range data access --------------------------------------------------

        /**
         * Finds the address of the first available cell of a specific type in
         * the passed cell ranges.
         *
         * @param {RangeArray|Range} ranges
         *  An array of range addresses, or a single cell range address, to be
         *  searched.
         *
         * @param {Object} [options]
         *  Optional parameters. Supports all options also supported by the
         *  method CellCollection.createAddressIterator().
         *
         * @returns {Address|Null}
         *  The address of the first available cell matching the specified cell
         *  type; or null, if no cell has been found.
         */
        this.findFirstCell = function (ranges, options) {
            // find the first cell in the search ranges
            var result = this.createAddressIterator(ranges, options).next();
            return result.done ? null : result.value;
        };

        /**
         * Returns the address of the next available cell of a specific type
         * (any blank cells, defined cells only, or content cells only) in the
         * specified direction.
         *
         * @param {Address} address
         *  The address of the cell whose nearest adjacent cell will be
         *  searched.
         *
         * @param {String} direction
         *  The direction to look for the cell. Must be one of the values
         *  'left', 'right', 'up', or 'down'.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {String} [options.type='all']
         *      Specifies the type of the next cell to be searched. See method
         *      CellCollection.createAddressIterator() for details.
         *  @param {Boolean} [options.visible=false]
         *      If set to true, only a visible cell (in a visible column, AND a
         *      visible row) will be found. By default, all visible and hidden
         *      cells will be searched.
         *  @param {Range} [options.boundRange]
         *      If specified, the search area will be restricted to that cell
         *      range. If omitted, the entire area of the sheet will be used.
         *
         * @returns {Address|Null}
         *  The address of the next available cell matching the specified cell
         *  type; or null, if no cell has been found.
         */
        this.findFirstCellLinear = function (address, direction, options) {
            // get the first result object of a linear iterator into the specified direction
            var result = this.createLinearAddressIterator(address, direction, _.extend({}, options, { skipStart: true })).next();
            return result.done ? null : result.value;
        };

        /**
         * Returns whether all passed ranges are blank (no values). Ignores
         * cell formatting attributes.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *
         * @returns {Boolean}
         *  Whether all cells in the passed ranges are blank.
         */
        this.areRangesBlank = function (ranges) {
            // return whether a cell iterator does not find a single value cell
            ranges = RangeArray.get(ranges).merge();
            return createAddressIterator(ranges, { type: 'value' }).next().done;
        };

        /**
         * Returns the cellModel of all cells in the passed cell ranges.
         *
         * @param {RangeArray|Range} ranges
         *  An array of range addresses, or a single cell range address, whose
         *  cell contents will be returned.
         *
         * @returns {Array<CellModel>}
         *  The models of the cells in the passed ranges. The result will be
         *  an array of cell model objects.
         */
        this.getRangeCellModels = function (ranges) {
            var contents = [];

            RangeArray.forEach(ranges, function (range) {
                var addressIt = self.createAddressIterator(range);
                IteratorUtils.forEach(addressIt, function (address) {
                    var cellModel = self.getCellModel(address);
                    contents.push(cellModel || { u: true });
                });
            });

            return contents;
        };

        /**
         * Returns the contents of all cells in the passed cell ranges.
         *
         * @param {RangeArray|Range} ranges
         *  An array of range addresses, or a single cell range address, whose
         *  cell contents will be returned.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Boolean} [options.visible=false]
         *      If set to true, only visible cells (cells in visible columns,
         *      AND in visible rows) will be included in the result. Otherwise,
         *      all visible and hidden cells will be returned.
         *  @param {Boolean} [options.covered=false]
         *      If set to true, includes all cells covered (hidden) by a merge
         *      range into the result. By default, covered cells in merged
         *      ranges will not be included.
         *  @param {Boolean} [options.attributes=false]
         *      If set to true, the result will contain the identifier of the
         *      cell auto-style, the merged formatting attributes, and the
         *      parsed number format of the cells.
         *  @param {Boolean} [options.display=false]
         *      If set to true, the result will contain the formatted display
         *      strings of the cells in the property 'display'.
         *  @param {Boolean} [options.compressed=false]
         *      If set to true, the result array will be optimized: Consecutive
         *      cells with equal contents and formatting will be represented by
         *      a single array element with an additional property 'count'.
         *      Especially useful, if large ranges located in the unused area
         *      of the sheet are queried.
         *  @param {Number} [options.maxCount]
         *      If specified, the maximum number of cells that will be returned
         *      in the result, regardless how large the passed ranges are.
         *
         * @returns {Array<Object>}
         *  The contents of the cells in the passed ranges. The result will be
         *  an array of cell content objects. Each cell content object will
         *  contain the following properties:
         *  - {Number|String|Boolean|ErrorCode|Null} value
         *      The cell value, or formula result (null for blank cells).
         *  - {Object} [style]
         *      The identifier of the cell auto-style. Will only be set, if the
         *      option 'attributes' has been set to true (see above).
         *  - {Object} [attributes]
         *      The merged formatting attributes of the cell, resolved from the
         *      cell auto-style. Will only be set, if the option 'attributes'
         *      has been set to true (see above).
         *  - {ParsedFormat} [format]
         *      The parsed number format code of the cell. Will only be set, if
         *      the option 'attributes' has been set to true (see above).
         *  - {String|Null} [display]
         *      The display string of the cell, according to its current value
         *      and number format. May be null, if the value cannot be
         *      formatted with the current number format. Will only be set, if
         *      the option 'attributes' has been set to true (see above).
         *  - {Number} [count]
         *      Always set in compressed mode (see option 'compressed' above).
         *      Contains the number of consecutive cells with equal contents
         *      and formatting represented by this array element.
         *  The cells in the array will be in order of the passed ranges. Cells
         *  in each single range will be collected row-by-row. The result array
         *  will contain the additional property 'count' that represents the
         *  total number of cells contained in the result (this value will be
         *  different to the length of the array in compressed mode).
         */
        this.getRangeContents = function (ranges, options) {

            // compressed mode
            var compressed = Utils.getBooleanOption(options, 'compressed', false);
            // the maximum number of cells to be included in the result
            var maxCount = Utils.getIntegerOption(options, 'maxCount', null, 1);
            // the result array returned by this method
            var contents = [];
            // the number of cells already inserted into the result (differs to array length in compressed mode)
            contents.count = 0;

            // creates a new result entry with value property, and optional display property
            var makeEntry = Utils.getBooleanOption(options, 'display', false) ?
                function (value, display) { return { value: value, display: display }; } :
                function (value) { return { value: value }; };

            // inserts the passed result entry into the array (handles compressed mode)
            function pushEntry(entry, count) {

                // reduce the passed count, if the limit will be reached
                if (maxCount) { count = Math.min(count, maxCount - contents.count); }
                if (count === 0) { return; }

                // update the total count of cells in the result array
                contents.count += count;

                // the preceding result, for compressed mode
                var prevEntry = compressed ? _.last(contents) : null;

                // increase count of previous entry, if the new entry is equal
                if (prevEntry) {
                    entry.count = prevEntry.count;
                    if ((prevEntry.value === entry.value) && (prevEntry.style === entry.style) && (prevEntry.display === entry.display)) {
                        prevEntry.count += count;
                        return;
                    }
                }

                // insert the new entry into the array (prepare count for compressed mode)
                if (compressed) {
                    entry.count = count;
                    contents.push(entry);
                } else {
                    for (; count > 0; count -= 1) { contents.push(entry); }
                }
            }

            // pushes result entries for blank cells without formatting attributes
            function pushBlanks(count) {
                pushEntry(makeEntry(null, ''), count);
            }

            // collects all cells in the passed range without formatting attributes
            function collectWithoutAttributes(range) {

                // linear index of next expected cell in the range (used to fill blank cells)
                var blankIndex = 0;

                // fill the result array with existing cells and preceding blank cells
                var addressIt = self.createAddressIterator(range, _.extend({}, options, { type: 'value' }));
                IteratorUtils.forEach(addressIt, function (address) {

                    // the linear index of the address in the range
                    var index = range.indexAt(address);
                    // the cell model at the specified address (exists always)
                    var cellModel = self.getCellModel(address);

                    // insert blank cell entries before the current cell model, and the cell model
                    pushBlanks(index - blankIndex);
                    pushEntry(makeEntry(cellModel.v, cellModel.d), 1);
                    blankIndex = index + 1;

                    // early exit, if the limit has been reached
                    if (maxCount === contents.count) { return Utils.BREAK; }
                });

                // fill up following blank cells
                pushBlanks(range.cells() - blankIndex);
            }

            // collects all cells in the passed range with formatting attributes
            function collectWithAttributes(range) {
                // TODO: use a cell iterator that collects equally formatted blank ranges
                IteratorUtils.forEach(self.createAddressIterator(range, options), function (address) {
                    var entry = makeEntry(self.getCellValue(address), self.getCellDisplayString(address));
                    entry.style = self.getCellStyleId(address);
                    entry.attributes = autoStyles.getMergedAttributeSet(entry.style);
                    entry.format = autoStyles.getParsedFormat(entry.style);
                    pushEntry(entry, 1);
                    // early exit, if the limit has been reached
                    if (maxCount === contents.count) { return Utils.BREAK; }
                });
            }

            // collect all cell contents
            var addAttributes = Utils.getBooleanOption(options, 'attributes', false);
            RangeArray.forEach(ranges, addAttributes ? collectWithAttributes : collectWithoutAttributes);
            return contents;
        };

        /**
         * Returns the content range (the entire range containing consecutive
         * content cells) surrounding the specified cell range.
         *
         * @param {Range} range
         *  The address of the cell range whose content range will be searched.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {String} [options.directions='up down left right']
         *      All directions the passed range will be expanded to, as
         *      space-separated token list with the tokens 'up', 'down',
         *      'left', and 'right'. If omitted, the range will be expanded to
         *      all four directions.
         *
         * @returns {Range}
         *  The address of the content range of the specified cell range.
         */
        this.findContentRange = SheetUtils.profileMethod('CellCollection.findContentRange()', function (range, options) {

            // collection of merged ranges in the active sheet
            var mergeCollection = sheetModel.getMergeCollection();
            // the resulting content range (start by expanding to merged ranges)
            var currRange = mergeCollection.expandRangeToMergedRanges(range);
            // the result of the last iteration, to detect exiting the loop
            var lastRange = null;
            // all directions the range will be expanded to
            var directions = Utils.getTokenListOption(options, 'directions', ['up', 'down', 'left', 'right']);

            // returns the next column/row index outside the passed cell range
            function getNextIndex(srcRange, forward, columns) {
                // index of the next column/row outside the range
                var startIndex = forward ? (srcRange.getEnd(columns) + 1) : (srcRange.getStart(columns) - 1);
                // return the index if it is valid, otherwise -1
                return (forward ? (startIndex > docModel.getMaxIndex(columns)) : (startIndex < 0)) ? -1 : startIndex;
            }

            // array iterator function, must be defined outside the do-while loop
            function expandRangeBorder(direction, index) {

                // whether to move towards the end of the sheet
                var forward = /^(down|right)$/.test(direction);
                // whether to move through columns in the same row
                var columns = /^(left|right)$/.test(direction);
                // the descriptor of the next column/row
                var nextIndex = getNextIndex(currRange, forward, columns);

                // do nothing, if the range already reached the sheet borders
                if (nextIndex < 0) {
                    directions.splice(index, 1);
                    return false;
                }

                // build a single column/row range next to the current range
                var boundRange = currRange.clone();
                boundRange.setBoth(nextIndex, columns);

                // expand sideways, unless these directions are not active anymore
                if (_.contains(directions, columns ? 'up' : 'left')) {
                    nextIndex = getNextIndex(boundRange, false, !columns);
                    if (nextIndex >= 0) { boundRange.setStart(nextIndex, !columns); }
                }
                if (_.contains(directions, columns ? 'down' : 'right')) {
                    nextIndex = getNextIndex(boundRange, true, !columns);
                    if (nextIndex >= 0) { boundRange.setEnd(nextIndex, !columns); }
                }

                // find a content cell next to the current range
                var result = createAddressIterator(boundRange, { type: 'value' }).next();
                if (result.done) { return; }

                // expand the current range to the found content cell
                var currAddress = forward ? currRange.end : currRange.start;
                currAddress.set(result.value.get(columns), columns);

                // performance: expand further into the direction as long as there are adjacent content cells
                var iterator = self.createLinearAddressIterator(result.value, direction, { skipStart: true });
                IteratorUtils.some(iterator, function (address) {
                    if (self.isBlankCell(address)) { return true; }
                    currAddress.set(address.get(columns), columns);
                });
            }

            // expand in all four directions until the range does not change anymore
            do {

                // store result of last iteration for comparison below
                lastRange = currRange.clone();

                // expand current range into all remaining directions
                Utils.iterateArray(directions, expandRangeBorder, { reverse: true });

                // expand the range to include merged ranges
                currRange = mergeCollection.expandRangeToMergedRanges(currRange);

            } while ((directions.length > 0) && currRange.differs(lastRange));

            // return the resulting range
            return currRange;
        });

        /**
         * Returns the address of the first or last cell with the specified
         * formatting attributes in the passed cell ranges, regardless whether
         * the cell is actually defined, or the attribute values have been
         * derived from column or row attributes. Uses an optimized search
         * algorithm for the empty areas between the defined cells, regarding
         * intervals of equally formatted columns and rows, and the position of
         * merged ranges covering these column/row default formatting.
         *
         * @param {RangeArray|Range} ranges
         *  An array of range addresses, or a single cell range address, whose
         *  cells will be searched for the attributes.
         *
         * @param {Object} attributeSet
         *  The incomplete attribute set to be matched against the formatting
         *  attributes in the passed cell ranges.
         *
         * @param {Object} [options]
         *  Optional parameters. Supports all options that are supported by the
         *  method CellCollection.createStyleIterator().
         *
         * @returns {Object|Null}
         *  A result descriptor for the first (or last, depending on direction)
         *  cell with matching formatting attributes; or null, if no such cell
         *  exists in the passed cell ranges. The cell descriptor will contain
         *  the following properties:
         *  - {Address} address
         *      The address of the first cell with the matching attributes.
         *  - {Range} range
         *      The original range from the passed range array containing the
         *      cell with the matching attributes.
         *  - {Number} index
         *      The array index of the cell range contained in the property
         *      'range'.
         */
        this.findCellWithAttributes = SheetUtils.profileMethod('CellCollection.findCellWithAttributes()', function (ranges, attributeSet, options) {

            // whether to find the last matching cell
            var reverse = Utils.getBooleanOption(options, 'reverse', false);
            // create a style iterator that visits equally formatted cell ranges in row bands
            var iterator = this.createStyleIterator(ranges, options);

            // filter for ranges with matching formatting attributes
            iterator = IteratorUtils.createFilterIterator(iterator, function (range, result) {
                var cellAttributeSet = autoStyles.getMergedAttributeSet(result.style);
                return AttributeUtils.matchesAttributesSet(cellAttributeSet, attributeSet);
            });

            // fetch the first result from the iterator
            var result = iterator.next();
            if (result.done) { return null; }

            // return a result descriptor on success
            var address = reverse ? result.value.end : result.value.start;
            return { address: address.clone(), range: result.orig, index: result.index };
        });

        /**
         * Returns whether the given range contains a header-line.
         *
         * @param {Range} range
         *  A single cell range address, whose cells will be sorted.
         *
         * @param {String} direction
         *  'horizontal' / 'vertical' direction of sorting.
         *
         * @returns {Boolean}
         *  'true' if the range contains a headline, 'flase' if not
         */
        this.hasRangeHeadline = function (range, direction) {

            // return the type of the cell-value
            function detectValueType(cellModel) {
                if (!cellModel) { return null; }

                if (cellModel.isFormula()) {
                    return 'formula';

                } else if (cellModel.isNumber()) {
                    if (self.getCellParsedFormat(cellModel.a).category === 'date') {
                        return 'date';
                    } else {
                        return 'number';
                    }

                } else if (cellModel.isText()) {
                    if (listCollection.listContains(cellModel.v)) {
                        return 'definedList';
                    } else {
                        return 'string';
                    }

                } else if (cellModel.isBlank()) {
                    return '';

                } else {
                    return null;
                }
            }

            // return an array of types from a range
            function typeDetection(range) {
                var result = [];
                IteratorUtils.forEach(range.iterator(), function (address) {
                    result.push(detectValueType(self.getCellModel(address)));
                });
                return result;
            }

            // decides whether a header-line is present, or not
            function hasHeadline(headType, contentTypes) {
                if (
                    (headType === 'definedList'         && _.uniq(contentTypes).length >= 1)
                    ||
                    (headType === 'date'                && _.uniq(contentTypes).length >= 1)
                    ||
                    (headType === null                  && _.uniq(contentTypes).length >= 1)
                    ||
                    (_.uniq(contentTypes).length === 1  && contentTypes[0] !== headType)
                ) {
                    return 'yes';

                } else if (
                    (headType === 'number'  && _.uniq(contentTypes).length >= 1)
                    ||
                    (headType === 'string'  && _.uniq(contentTypes).length === 1 && contentTypes[0] === 'string')
                    ||
                    (headType === 'number'  && _.uniq(contentTypes).length === 1 && contentTypes[0] === 'number')
                    ||
                    (headType === 'date'    && _.uniq(contentTypes).length === 1 && contentTypes[0] === 'date')
                    ||
                    (headType === 'formula' && _.uniq(contentTypes).length === 1 && contentTypes[0] === 'formula')
                    ||
                    (headType === 'formula' && _.uniq(contentTypes).length >= 1)
                ) {
                    return 'no';

                } else {
                    return 'maybe';
                }
            }

            var columns                 = (direction === 'vertical'),
                startLine               = columns ? range.start[0] : range.start[1],
                startColRow             = columns ? range.start[1] : range.start[0],

                tableCollection         = sheetModel.getTableCollection(),
                arrTables               = tableCollection.findTables(range),

                arrHeadTypes            = [],
                arrContentTypes         = [],

                states                  = { yes: 0, no: 0, maybe: 0 };

            // when the selection covers exactly one table, go on ...
            //   ... otherwise, it isn't relevant whether the tables have headerRows or not
            if (arrTables.length === 1) {
                if (arrTables[0].getRange().equals(range)) {
                    return arrTables[0].hasHeaderRow();
                }
            }

            // creates an object which counts 'yes', 'no' and 'maybe' to decide whether
            // we have a headline or not
            IteratorUtils.forEach(range.interval(columns).iterator(), function (index) {
                var currentRange = range.lineRange(columns, (index - startLine)),
                    head         = currentRange.start,
                    content      = currentRange.clone().setStart((startColRow + 1), !columns);

                var headType = detectValueType(self.getCellModel(head));
                arrHeadTypes.push(headType);

                var contentTypes = typeDetection(content);
                arrContentTypes = arrContentTypes.concat(contentTypes);

                if (hasHeadline(headType, contentTypes) === 'yes') {
                    states.yes++;

                } else if (hasHeadline(headType, contentTypes) === 'no') {
                    states.no++;

                } else {
                    states.maybe++;
                }
            });

            // if the leading line is empty, there is no header
            if (_.uniq(arrHeadTypes).length === 1 && arrHeadTypes[0] === null) {
                return false;
            }

            // when 'yes' predominates, return 'true'
            if (states.yes > states.no) {
                return true;

            // when 'no' predominates
            } else if (states.no > states.yes) {
                // check 'yes' + 'maybe' against 'no'
                if (states.yes + states.maybe > states.no) {
                    return true;
                } else {
                    return false;
                }

            // otherwise
            } else {
                // check 'maybe'
                if (states.maybe > 0) {
                    return true;
                } else {
                    return false;
                }
            }
        };

        /**
         * Returns the source range for an automatically generated subtotal
         * formula for a single cell.
         *
         * @param {Address} address
         *  The address of the target cell for the subtotal formula.
         *
         * @returns {Range|Null}
         *  The source range for a subtotal formula for the specified target
         *  cell; or null, if no such range could be found.
         */
        this.findAutoFormulaRange = function (address) {

            var top = getAutoFormulaRange(address, 'up', true);
            var left = getAutoFormulaRange(address, 'left', true);

            // prefer the range located nearer to the target cell
            if (top && left) {

                var topRange = top.range;
                var leftRange = left.range;
                var preferLeft = Math.abs(address[0] - left.address1[0]) < Math.abs(address[1] - top.address1[1]);
                return preferLeft ? leftRange : topRange;
            } else if (top) {
                return top.range;
            } else if (left) {
                return left.range;
            } else {
                return null;
            }
        };

        // operation generators -----------------------------------------------

        /**
         * Converts the passed text to a cell contents object as expected by
         * the other operation generator methods of this class.
         *
         * @param {String} parseText
         *  The text to be parsed to a cell contents object. The text may
         *  represent a plain or formatted number, a boolean value, an error
         *  code, or a formula expression.
         *
         * @param {Address} refAddress
         *  The reference address used to interpret a formula expression.
         *
         * @returns {Object}
         *  The result object, with the following properties:
         *  - {Number|String|Boolean|ErrorCode|Null} v
         *      The resulting value with correct data type, or the calculated
         *      result of a formula expression.
         *  - {String|Null} f
         *      The formula expression, translated to the native formula
         *      grammar used in document operations, without leading equality
         *      sign; or null, if the passed text is not a formula expression.
         *  - {Object|Null} result
         *      The result descriptor of the formula interpreter, as returned
         *      by the method FormulaInterpreter.interpretTokens(); or null, if
         *      the passed text is not a formula expression.
         *  - {String|Number} [format]
         *      A specific number format code (as string), or the identifier of
         *      a preset format code (as integer), associated to the parsed
         *      number, or the result of the formula. If omitted, the number
         *      format shall not be modified.
         *  - {String} [url]
         *      The trimmed URL, if the passed text represents a valid URL; or
         *      an empty string, if the cell currently contains a hyperlink,
         *      and the passed text is an empty string (causes to delete the
         *      existing hyperlink).
         */
        this.parseCellValue = function (parseText, refAddress) {

            // the resulting contents object
            var contents = { v: parseText, f: null, result: null };

            // do not parse the text for the number format category 'text'
            var category = this.getCellParsedFormat(refAddress).category;
            if (category === 'text') { return contents; }

            // handle specific strings not to be recognized as formulas
            if (/^[-+=]+$/.test(parseText)) { parseText = '\'' + parseText; }

            // try to parse the text as (formatted) value (empty strings will result in value null)
            var parseResult = numberFormatter.parseFormattedValue(parseText, { blank: true });
            contents.v = parseResult.value;

            // the identifier or format code of a number format (may be changed via formula result)
            var parsedFormat = numberFormatter.getParsedFormat(parseResult.format);

            // If the parsed value is still a string (no conversion to numbers etc. possible), and the original
            // string starts with specific characters, parse and interpret a formula expression in UI grammar.
            // This check excludes simple numbers with leading sign, e.g. '-1234' (which have been converted to
            // numbers already), but includes other formula expressions starting with a sign character instead
            // of an equality sign, e.g. '-A1+A2'.
            if (_.isString(contents.v) && /^[-+=]./.test(parseText)) {

                // parse the formula in UI grammar, and translate to native grammar
                var tokenArray = new TokenArray(sheetModel, 'cell');
                tokenArray.parseFormula('ui', parseText.replace(/^=/, ''), { autoCorrect: true, refAddress: refAddress });
                contents.f = tokenArray.getFormula('op', { refAddress: refAddress, targetAddress: refAddress });

                // calculate the formula result, handle formula errors
                var result = contents.result = tokenArray.interpretFormula('val', { refAddress: refAddress, targetAddress: refAddress });
                contents.v = result.value;

                // the formula engine may provide an arbitrary number format with the result
                if ('format' in result) { parsedFormat = result.format; }
            }

            // add a formatting attribute for the number format, if the cell's category changes, but ignore
            // the standard number formats, and ignore changes inside any of the date/time categories
            // (e.g. do not change a date cell to a time cell)
            var formatCode = parsedFormat ? getNewCellFormatCode(refAddress, parsedFormat) : null;
            if (formatCode) { contents.format = formatCode; }

            // try to parse a URL from the passed text
            if (contents.v === null) {
                // delete the URL explicitly (via an empty string), if text is cleared
                if (sheetModel.getHyperlinkCollection().getCellURL(refAddress)) { contents.url = ''; }
            } else if (!contents.f && _.isString(contents.v)) {
                var url = HyperlinkUtils.checkForHyperlink(contents.v);
                if (_.isString(url) && (url.length > 0)) { contents.url = url; }
            }

            return contents;
        };

        /**
         * Generates the cell operations, and the undo operations, to change
         * the specified contents for a single cell, or multiple cells in the
         * sheet, and optionally the appropriate hyperlink operations.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {Address} address
         *  The address of the top-left cell to be modified.
         *
         * @param {Object|Array<Object>} contents
         *  An object with the new properties to be applied at a single cell,
         *  or an array of row descriptors with cell property objects to change
         *  multiple cells at once. The row descriptors are objects with the
         *  following properties:
         *  - {Array<Object>} [rowData.c]
         *      An array of cell contents objects (see below for details). Can
         *      be omitted to skip one row, or multiple rows.
         *  - {Number} [rowData.r=1]
         *      The row repetition count. If set to a value greater than one,
         *      multiple consecutive rows will be changed simultaneously with
         *      the row descriptor.
         *  A cell content object (either passed directly to this method, or as
         *  element of the 'c' array property of a row descriptor) supports the
         *  following properties:
         *  - {Number|String|Boolean|ErrorCode|Null} [contents.v]
         *      The new value of the cell. The value null represents a blank
         *      cell. If omitted, the current cell value will not be changed.
         *  - {String|Null} [contents.f]
         *      The new formula expression for the cell (in the native formula
         *      grammar of the file format). If set to null, the current
         *      formula will be removed from the cell. If omitted, the current
         *      formula will not be changed. If the 'v' property has been set
         *      to null (clear the cell value), this property MUST be set to
         *      null too in order to delete the formula (formulas must have a
         *      non-null result value).
         *  - {String} [contents.s]
         *      The name of a new auto-style to be applied at the cell. If this
         *      property and the properties 'a' and 'format' have been omitted,
         *      the current auto-style of the cell will not be changed.
         *  - {Object} [contents.a]
         *      An attribute set with explicit formatting attributes to be
         *      applied at the cell. If this property, and the properties 's'
         *      and 'format' have been omitted, the current auto-style of the
         *      cell will not be changed.
         *  - {Number|String} [contents.format]
         *      The identifier of a number format, or an explicit format code,
         *      to be inserted into the auto-style of the cell. If this
         *      property, and the properties 's' and 'a' have been omitted, the
         *      current auto-style of the cell will not be changed.
         *  - {Boolean} [contents.u=false]
         *      If set to true, the cell will be deleted explicitly (removed
         *      completely from the cell collection). The previous properties
         *      for changing the cell contents or formatting MUST NOT be used
         *      in this case.
         *  - {Number} [contents.r=1]
         *      The repetition count causing to change the specified number of
         *      consecutive cells in the same row.
         *  - {String} [contents.url]
         *      If set to a non-empty string, a hyperlink for the cell will be
         *      created. If set to an empty string, an existing URL will be
         *      removed. If multiple content objects in a two-dimensional array
         *      contain URLs, they will be merged to coherent cell ranges.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated.
         */
        this.generateCellContentOperations = function (generator, address, contents) {

            // convert a single object to a contents matrix
            var matrix = _.isArray(contents) ? contents : [{ c: [contents] }];

            // let the private method do all the work
            return generateCellContentOperations(generator, address, matrix);
        };

        /**
         * Generates the cell operations, and the undo operations, to fill the
         * passed cell ranges in the sheet with some individual contents.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *  Each range object MUST contain an additional property 'fillData'
         *  with cell properties to be applied at all cells in the respective
         *  range. See method CellCollection.generateFillOperations() for more
         *  details about the properties. Overlapping ranges MUST NOT contain
         *  the same cell properties with different values (e.g.: set value 1
         *  or value 2 to the same cells in overlapping ranges), otherwise the
         *  result will be undefined. However, incomplete cell property objects
         *  of overlapping ranges will be merged together for the overlapping
         *  parts of the ranges (e.g.: setting the auto-style 'a1' to the range
         *  A1:B2, and the value 2 to the range B2:C3 will result in setting
         *  both the auto-style and the value to cell B2).
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated, or that will be rejected with an object with 'cause'
         *  property set to one of the following error codes:
         *  - 'cells:overflow': The cell ranges contain too many filled cells.
         *  - 'cols:overflow': Too many entire columns in the cell ranges.
         *  - 'rows:overflow': Too many entire rows in the cell ranges.
         */
        this.generateRangeOperations = SheetUtils.profileAsyncMethod('CellCollection.generateRangeOperations()', function (generator, ranges) {
            SheetUtils.log('ranges=' + ranges);

            // divide the original ranges into column ranges, row ranges, and regular cell ranges
            var rangeGroups = getRangeGroups(ranges);
            // add the 'rangeMode' property to the cell ranges
            var taggedRanges = getTaggedRanges(ranges);

            // transport the changed ranges through the promise chain
            var promise = this.createResolvedPromise(new RangeArray());

            // let the row collection generate the operations for entire formatted rows
            promise = generateChainedRowIntervalOperations(promise, rangeGroups, function (rowIntervals) {
                return rowCollection.generateIntervalOperations(generator, rowIntervals);
            });

            // let the column collection generate the operations for entire formatted columns
            promise = generateChainedColIntervalOperations(promise, rangeGroups, function (colIntervals) {
                return colCollection.generateIntervalOperations(generator, colIntervals);
            });

            // generate the cell operations for all ranges, merge cell content objects of overlapping ranges
            return generateChainedCellRangeOperations(promise, function () {
                return generateCellRangeOperations(generator, taggedRanges, function (fillDataArray) {
                    return fillDataArray.reduce(_.extend, {});
                });
            });
        });

        /**
         * Generates the cell operations, and the undo operations, to fill the
         * same cell contents into entire cell ranges in the sheet.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *
         * @param {Object} contents
         *  The new properties to be applied at all cells. See public method
         *  CellCollection.generateCellContentOperations() for details.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated, or that will be rejected with an object with 'cause'
         *  property set to one of the following error codes:
         *  - 'cells:overflow': The cell ranges contain too many filled cells.
         *  - 'cols:overflow': Too many entire columns in the cell ranges.
         *  - 'rows:overflow': Too many entire rows in the cell ranges.
         */
        this.generateFillOperations = SheetUtils.profileAsyncMethod('CellCollection.generateFillRangeOperations()', function (generator, ranges, contents) {
            SheetUtils.log('ranges=' + ranges + ' contents=', contents);

            // do not modify the cell contents object passed from caller
            contents = _.clone(contents);
            // add an arbitrary cache key to the contents object that causes internal auto-style caching
            // (all ranges will be formatted with the same new formatting attributes)
            contents.cacheKey = _.uniqueId('key');

            // create or remove hyperlinks for the cell ranges
            if (_.isString(contents.url)) {
                sheetModel.getHyperlinkCollection().generateHyperlinkOperations(generator, ranges, contents.url);
                delete contents.url;
            }

            // divide the original ranges into column ranges, row ranges, and regular cell ranges
            var rangeGroups = getRangeGroups(ranges);
            // add the 'rangeMode' property to the cell ranges
            var taggedRanges = getTaggedRanges(ranges);
            // the change flags specifying how the cell will be potentially changed
            var changeFlags = getChangeFlags(contents);

            // if ranges will be formatted, the cell ranges not covering entire columns or rows must not exceed a specific limit
            if (changeFlags.style && rangeGroups.cellRanges && (rangeGroups.cellRanges.merge().cells() > SheetUtils.MAX_FILL_CELL_COUNT)) {
                return SheetUtils.makeRejected('cells:overflow');
            }

            // when filling with values/formulas, the entire ranges must not exceed a specific limit (but not when clearing or undefining ranges)
            if ((changeFlags.value || changeFlags.formula) && (taggedRanges.merge().cells() > SheetUtils.MAX_FILL_CELL_COUNT)) {
                return SheetUtils.makeRejected('cells:overflow');
            }

            // transport the changed ranges through the promise chain
            var promise = this.createResolvedPromise(new RangeArray());

            // let the row collection generate the operations for entire formatted rows
            promise = generateChainedRowIntervalOperations(promise, rangeGroups, function (rowIntervals) {
                return rowCollection.generateFillOperations(generator, rowIntervals, contents);
            });

            // let the column collection generate the operations for entire formatted columns
            promise = generateChainedColIntervalOperations(promise, rangeGroups, function (colIntervals) {
                return colCollection.generateFillOperations(generator, colIntervals, contents);
            });

            // generate the cell operations for all ranges, using the same cell contents object for all ranges
            // (the same contents object can be recycled, generateCellRangeOperations() creates a clone internally)
            return generateChainedCellRangeOperations(promise, function () {
                return generateCellRangeOperations(generator, taggedRanges, _.constant(contents));
            });
        });

        /**
         * Generates the cell operations, and the undo operations, to fill the
         * cells next to the specified range with automatically generated
         * values, formulas, and formatting.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {Range} range
         *  The address of the source cell range with values, formulas, and
         *  formatting to be expanded into one direction.
         *
         * @param {String} border
         *  Specifies the direction to expand the contents of the passed range.
         *  Allowed values are 'left', 'right', 'top', or 'bottom'.
         *
         * @param {Number} count
         *  The number of columns/rows to extend the cell range into the
         *  specified direction.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated, or that will be rejected with an object with 'cause'
         *  property set to one of the following error codes:
         *  - 'cells:overflow': The cell ranges contain too many filled cells.
         *  - 'cols:overflow': Too many entire columns in the cell ranges.
         *  - 'rows:overflow': Too many entire rows in the cell ranges.
         */
        this.generateAutoFillOperations = SheetUtils.profileAsyncMethod('CellCollection.generateAutoFillOperations()', function (generator, range, border, count, options) {
            var // collection of all merged ranges of this sheet
                mergeCollection = sheetModel.getMergeCollection(),
                // use the correct collection (col or row)
                colRowCollection = ((docModel.isColRange(range)) ? colCollection : rowCollection),
                // collection of all hyperlinks of this sheet
                hyperlinkCollection = sheetModel.getHyperlinkCollection(),
                // column or row direction
                columns         = /^(top|bottom)$/.test(border),
                // reverse running or not
                reverse         = /^(left|top)$/.test(border),
                // targetRange (cells)
                targetRange     = createTargetRange(range, columns, count, reverse),
                // whether a coplete col/row is selected or not
                completeColRow  = (docModel.isColRange(range) || docModel.isRowRange(range)),
                // targetInterval (cols or rows)
                targetInterval  = (completeColRow) ? createTargetInterval(range.interval(!columns), count, reverse) : null,
                // promise
                promise         = $.when();

            // if (mergeCollection.rangesOverlapAnyMergedRange(targetRange)) {
            if (mergeCollection.getMergedRanges(targetRange).length !== mergeCollection.getMergedRanges(targetRange, 'contain').length) {
                return SheetUtils.makeRejected('autofill:merge:overlap');
            }

            // CLEAR target cols/rows
            if (completeColRow) {
                promise = promise.then(function () {
                    return colRowCollection.generateFillOperations(generator, targetInterval, { s: '', a: '' });
                });
            }

            // CLEAR target cells
            promise = promise.then(function () {
                return self.generateFillOperations(generator, targetRange, { u: true });
            });

            // CLEAR hyperlink-ranges in target-range
            promise = promise.then(function () {
                return self.generateFillOperations(generator, targetRange, { url: '' });
            });

            // CLEAR merged ranges within target range
            if (mergeCollection.rangesOverlapAnyMergedRange(targetRange)) {
                promise = promise.then(function () {
                    return mergeCollection.generateMergeCellsOperations(generator, targetRange, 'unmerge');
                });
            }

            // generate autofill for cols/rows (styles)
            if (completeColRow) {
                promise = promise.then(function () {
                    return colRowCollection.generateAutoFillOperations(generator, range.interval(!columns), targetInterval, reverse);
                });
            }

            // generate hyperlink ranges in target range
            if (hyperlinkCollection.rangeOverlapsHyperlinks(range)) {
                promise = promise.then(function () {
                    return hyperlinkCollection.generateAutoFillOperations(generator, range, targetRange, columns, reverse);
                });
            }

            // generate merged ranges in target range
            if (mergeCollection.rangesOverlapAnyMergedRange(range)) {
                promise = promise.then(function () {
                    return mergeCollection.generateAutoFillOperations(generator, range, targetRange, columns, reverse);
                });
            }

            // generate autofill for cells (styles + content)
            promise = promise.then(function () {
                return generateAutoFillOperations(generator, range, border, count, options);
            });

            promise.done(function () {
                generator.reverseOperations({ undo: true });
            });

            return promise;
        });

        /**
         * Generates the cell operations, and the undo operations, to fill the
         * cell ranges in the sheet with outer and/or inner borders.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *
         * @param {Object} borderAttributes
         *  The border attributes to be applied at all cells. May contain
         *  regular border attributes, as supported by the document operations
         *  ('borderTop', 'borderBottom', 'borderLeft', 'borderRight') which
         *  will be applied at the outer boundaries of the passed ranges; and
         *  the pseudo attributes 'borderInsideHor' and 'borderInsideVert',
         *  which will be applied to the inner cells of the cell ranges.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated, or that will be rejected with an object with 'cause'
         *  property set to one of the following error codes:
         *  - 'cells:overflow': The cell ranges contain too many filled cells.
         *  - 'cols:overflow': Too many entire columns in the cell ranges.
         *  - 'rows:overflow': Too many entire rows in the cell ranges.
         */
        this.generateBorderOperations = (function () {

            // Returns the parts of the passed ranges that will receive a new border attribute.
            // Parameters 'columns'/'leading' specify the position of the cell border to be modified.
            // The resulting ranges will contain an additional property 'fillData' with border settings
            // as expected by the helper method SheetOperationsGenerator.createBorderContents().
            function getChangedBorderRanges(ranges, borderAttributes, columns, leading) {

                // the resulting ranges
                var resultRanges = new RangeArray();

                // adds the passed style settings to the passed ranges, and appends them to the resulting range array
                function appendRanges(newRanges, fillData) {
                    return resultRanges.append(Utils.addProperty(newRanges, 'fillData', fillData));
                }

                // the keys of the borders to be processed
                var outerKey = SheetUtils.getOuterBorderKey(columns, leading);
                var innerKey = SheetUtils.getInnerBorderKey(columns);

                // get the border attributes, nothing to do without outer nor inner border attribute
                var outerBorder = borderAttributes[SheetUtils.getBorderName(outerKey)];
                var innerBorder = borderAttributes[SheetUtils.getBorderName(innerKey)];
                if (!outerBorder && !innerBorder) { return null; }

                // add the adjacent ranges whose opposite borders need to be deleted while setting a range border
                // (if the outer border will not be changed, the borders of the adjacent cells will not be modified neither)
                if (outerBorder) {
                    // the adjacent ranges at the specified range border whose existing borders will be deleted
                    var adjacentRanges = getAdjacentRanges(ranges, columns, leading);
                    // opposite borders will be deleted, e.g. the right border left of the original range (but not if
                    // they are equal to the outer border set at the cell range, to reduce the amount of changed cells)
                    var oppositeKey = SheetUtils.getOuterBorderKey(columns, !leading);
                    appendRanges(adjacentRanges, { key: oppositeKey, keepBorder: outerBorder });
                }

                // shortcut (outer and inner borders are equal): return entire ranges instead of splitting into outer/inner
                if (outerBorder && innerBorder && Border.isEqual(outerBorder, innerBorder)) {
                    return appendRanges(ranges.clone(true), { key: outerKey, border: outerBorder, cacheType: 'outer' });
                }

                // divide the ranges into ranges covering entire columns/rows and other ranges
                var rangeGroups = ranges.group(function (range) { return docModel.isFullRange(range, !columns) ? 'fullRanges' : 'cellRanges'; });

                // do not set outer borders at entire column/row ranges, use inner border for all cells
                if (innerBorder && rangeGroups.fullRanges) {
                    appendRanges(rangeGroups.fullRanges.clone(true), { key: outerKey, border: innerBorder, cacheType: 'outer' });
                }

                // outer border will be set to the outer cells only
                if (outerBorder && rangeGroups.cellRanges) {
                    var subRangeMethod = columns ? (leading ? 'leadingCol' : 'trailingCol') : (leading ? 'headerRow' : 'footerRow');
                    appendRanges(RangeArray.invoke(rangeGroups.cellRanges, subRangeMethod), { key: outerKey, border: outerBorder, cacheType: 'outer' });
                }

                // inner border will be set to the remaining cells
                if (innerBorder && rangeGroups.cellRanges) {
                    var offset = leading ? 1 : 0;
                    appendRanges(RangeArray.map(rangeGroups.cellRanges, function (range) {
                        return range.singleLine(columns) ? null : range.lineRange(columns, offset, range.size(columns) - 1);
                    }), { key: outerKey, border: innerBorder, cacheType: 'inner' });
                }

                return resultRanges;
            }

            // the actual implementation of the public method returned from local scope
            function generateBorderOperations(generator, ranges, borderAttributes) {
                SheetUtils.log('ranges=' + ranges + ' attributes=', borderAttributes);

                // ensure an array, remove duplicates (but do not merge the ranges)
                ranges = RangeArray.get(ranges).unify();
                // divide the original ranges into column ranges, row ranges, and regular cell ranges
                var rangeGroups = getRangeGroups(ranges);

                // prepare the range arrays with the parts of the passed ranges that will get a new single border
                var rangesT = getChangedBorderRanges(ranges, borderAttributes, false, true);
                var rangesB = getChangedBorderRanges(ranges, borderAttributes, false, false);
                var rangesL = getChangedBorderRanges(ranges, borderAttributes, true,  true);
                var rangesR = getChangedBorderRanges(ranges, borderAttributes, true,  false);
                // add the 'rangeMode' property to the cell ranges
                var taggedRanges = getTaggedRanges(new RangeArray(rangesT, rangesB, rangesL, rangesR));

                // root cache key for using the same auto-style cache keys across all generators
                var cacheKey = _.uniqueId('key');
                // transport the changed ranges through the promise chain
                var promise = this.createResolvedPromise(new RangeArray());

                // let the row collection generate the operations for entire formatted rows
                promise = generateChainedRowIntervalOperations(promise, rangeGroups, function (rowIntervals) {
                    return rowCollection.generateBorderOperations(generator, rowIntervals, borderAttributes, cacheKey);
                });

                // let the column collection generate the operations for entire formatted columns
                promise = generateChainedColIntervalOperations(promise, rangeGroups, function (colIntervals) {
                    return colCollection.generateBorderOperations(generator, colIntervals, borderAttributes, cacheKey);
                });

                // generate the cell operations for all ranges, use the border attributes resolved above
                return generateChainedCellRangeOperations(promise, function () {
                    return generateCellRangeOperations(generator, taggedRanges, function (fillDataArray) {
                        return generator.createBorderContents(fillDataArray, cacheKey);
                    });
                });
            }

            return SheetUtils.profileAsyncMethod('CellCollection.generateBorderOperations()', generateBorderOperations);
        }());

        /**
         * Generates the cell operations, and the undo operations, to change
         * the formatting of existing border lines of the passed cell ranges in
         * the sheet.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *
         * @param {Border} border
         *  A border attribute value, which may be incomplete. All properties
         *  contained in this object (color, line style, line width) will be
         *  set for all visible borders in the cell range. Omitted border
         *  properties will not be changed.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated, or that will be rejected with an object with 'cause'
         *  property set to one of the following error codes:
         *  - 'cells:overflow': The cell ranges contain too many filled cells.
         *  - 'cols:overflow': Too many entire columns in the cell ranges.
         *  - 'rows:overflow': Too many entire rows in the cell ranges.
         */
        this.generateVisibleBorderOperations = (function () {

            // Returns the adjacent ranges of the passed ranges whose border will be changed additionally to the
            // covered ranges. The resulting ranges will contain an additional property 'fillData' with border
            // settings as expected by the helper method SheetOperationsGenerator.createVisibleBorderContents().
            function getAdjacentBorderRanges(ranges, columns, leading) {
                var adjacentRanges = getAdjacentRanges(ranges, columns, leading);
                return Utils.addProperty(adjacentRanges, 'fillData', { keys: SheetUtils.getOuterBorderKey(columns, !leading) });
            }

            // the actual implementation of the public method returned from local scope
            function generateVisibleBorderOperations(generator, ranges, border) {
                SheetUtils.log('ranges=' + ranges + ' border=' + JSON.stringify(border));

                // ensure an array, remove duplicates (but do not merge the ranges)
                ranges = RangeArray.get(ranges).unify();
                // divide the original ranges into column ranges, row ranges, and regular cell ranges
                var rangeGroups = getRangeGroups(ranges);

                // all borders need to be changed inside the passed ranges
                var fillRanges = Utils.addProperty(ranges.clone(true), 'fillData', { keys: 'tblrdu' });
                // the adjacent cells outside the ranges (the adjoining borders will be changed too)
                var rangesT = getAdjacentBorderRanges(ranges, false, true);
                var rangesB = getAdjacentBorderRanges(ranges, false, false);
                var rangesL = getAdjacentBorderRanges(ranges, true,  true);
                var rangesR = getAdjacentBorderRanges(ranges, true,  false);
                // add the 'rangeMode' property to the cell ranges
                var taggedRanges = getTaggedRanges(new RangeArray(fillRanges, rangesT, rangesB, rangesL, rangesR));

                // root cache key for using the same auto-style cache keys across all generators
                var cacheKey = _.uniqueId('key');
                // transport the changed ranges through the promise chain
                var promise = this.createResolvedPromise(new RangeArray());

                // let the row collection generate the operations for entire formatted rows
                promise = generateChainedRowIntervalOperations(promise, rangeGroups, function (rowIntervals) {
                    return rowCollection.generateVisibleBorderOperations(generator, rowIntervals, border, cacheKey);
                });

                // let the column collection generate the operations for entire formatted columns
                promise = generateChainedColIntervalOperations(promise, rangeGroups, function (colIntervals) {
                    return colCollection.generateVisibleBorderOperations(generator, colIntervals, border, cacheKey);
                });

                // generate the cell operations for all ranges, use the border attributes resolved above
                return generateChainedCellRangeOperations(promise, function () {
                    return generateCellRangeOperations(generator, taggedRanges, function (fillDataArray) {
                        return generator.createVisibleBorderContents(fillDataArray, border, cacheKey);
                    });
                });
            }

            return SheetUtils.profileAsyncMethod('CellCollection.generateVisibleBorderOperations()', generateVisibleBorderOperations);
        }());

        /**
         * Generates the cell operations, and the undo operations, to merge the
         * specified cell ranges. The first value cell of each cell range will
         * be moved to the top-left corner of the cell range, and all other
         * cells covered by the ranges will be deleted. The creation of column
         * or row formatting for merged ranges covering entire columns or rows
         * will be handled automatically.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *  Multiple ranges MUST NOT overlap each other!
         *
         * @param {String} type
         *  The type of the merge operation. Must be one of the value supported
         *  by the document operation 'mergeCells'.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated.
         */
        this.generateMergeCellsOperations = SheetUtils.profileAsyncMethod('CellCollection.generateMergeCellsOperations()', function (generator, ranges, type) {
            SheetUtils.log('ranges=' + ranges + ' type=' + type);

            // nothing to do when unmerging cells
            if (type === 'unmerge') {
                return this.createResolvedPromise(new RangeArray());
            }

            // a private generator for restoring all cell-styles
            var cellOpGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });
            // a private generator for restoring all range borders
            var borderOpGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });

            // whether to merge single columns in the passed ranges
            var vertical = type === 'vertical';
            // whether to merge single rows in the passed ranges
            var horizontal = type === 'horizontal';
            // the ranges to be processed, with additional information in 'fillData' properties
            var fillRanges = new RangeArray();

            // horizontal/vertical merging does not have any effect for single-column/row ranges
            ranges = RangeArray.reject(ranges, function (range) {
                return horizontal ? range.singleCol() : vertical ? range.singleRow() : range.single();
            });

            // process each range for its own (ranges do not overlap)
            var promise = this.iterateArraySliced(ranges, function (range) {

                // number of columns/rows to be processed according to merge type
                var count = vertical ? range.cols() : horizontal ? range.rows() : 1;

                // process all columns/rows separately, or the entire range, according to merge type
                return this.repeatSliced(function (index) {

                    // homogeneous border for each single edge
                    var homogeneousBorders = getHomogeneousOutlineBorders(range);

                    // the cell range to be searched for a value cell
                    var searchRange = vertical ? range.colRange(index) : horizontal ? range.rowRange(index) : range;

                    // find the first cell model with a value in the current range
                    var result = createAddressIterator(searchRange, { type: 'value' }).next();
                    var cellModel = result.done ? null : result.model;

                    // fill the entire range with the same formatting (either from first value cell, or from origin)
                    var styleRange = range.clone();
                    styleRange.fillData = {
                        s: cellModel ? cellModel.s : this.getCellStyleId(range.start),
                        a: {
                            cell: {
                                borderTop: Border.NONE,
                                borderLeft: Border.NONE,
                                borderBottom: Border.NONE,
                                borderRight: Border.NONE
                            }
                        }
                    };
                    styleRange.homogeneousBorders = homogeneousBorders;
                    fillRanges.push(styleRange);

                    // move contents of the first value cell to the origin of the merged range
                    if (cellModel && !range.startsAt(cellModel.a)) {

                        // move value and formula (but not the formatting, see below) to top-left corner
                        var startRange = new Range(range.start);
                        startRange.rangeMode = 'skip';
                        startRange.fillData = {
                            v: cellModel.v,
                            f: cellModel.f
                        };
                        fillRanges.push(startRange);

                        // delete values and formulas from all other cells of the range
                        new RangeArray(range).difference(new Range(range.start)).forEach(function (partRange) {
                            partRange.rangeMode = 'skip';
                            partRange.fillData = { v: null, f: null };
                            fillRanges.push(partRange);
                        });
                    }

                }, { delay: 'immediate', cycles: count });
            }, { delay: 'immediate' });

            // generate the resulting fill operations
            promise = promise.then(function () {
                return self.generateRangeOperations(cellOpGenerator, fillRanges);
            });

            promise = promise.then(function () {
                var promise2 = $.when();

                fillRanges.forEach(function (r) {
                    if (r.homogeneousBorders) {
                        promise2 = promise2.then(function () {
                            return self.generateBorderOperations(borderOpGenerator, r, r.homogeneousBorders);
                        });
                    }
                });

                return promise2;
            });

            return promise.done(function () {
                generator.appendOperations(cellOpGenerator);
                generator.appendOperations(borderOpGenerator);

                generator.appendOperations(borderOpGenerator, { undo: true });
                generator.appendOperations(cellOpGenerator, { undo: true });
            });
        });

        /**
         * Generates the operations, and the undo operations, to insert blank
         * cells at the specified cell ranges, or to delete the specified cell
         * ranges, by moving the remaining existing cells to the left, right,
         * top, or bottom.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {RangeArray|Range} ranges
         *  An array of cell range addresses, or a single cell range address.
         *
         * @param {String} direction
         *  Where to move the existing cells to. MUST be one of the following
         *  values:
         *  - 'left':
         *      The specified cells will be deleted, the following cells right
         *      of the ranges will be moved to the left.
         *  - 'right':
         *      The specified cells will be made blank by moving the existing
         *      cells to the right. Cells moved outside the sheet will be
         *      deleted.
         *  - 'up':
         *      The specified cells will be deleted, the following cells below
         *      the ranges will be moved up.
         *  - 'down':
         *      The specified cells will be made blank by moving the existing
         *      cells down. Cells moved outside the sheet will be deleted.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the resulting merged index
         *  intervals) when all operations have been generated.
         */
        this.generateMoveCellsOperations = (function () {

            // generates the undo operations to restore all existing cells in the specified cell ranges
            function generateRestoreCellOperations(generator, ranges) {
                return self.generateFillOperations(generator, ranges, { u: true });
            }

            // generates the undo operations to restore all existing cells in the passed column/row intervals
            function generateRestoreIntervalOperations(generator, moveDesc) {
                var undoRanges = docModel.makeFullRanges(moveDesc.deleteIntervals, moveDesc.columns);
                return generateRestoreCellOperations(generator, undoRanges);
            }

            // generates the operations and undo operations for other sheet collections in the document, before the cells will be moved
            function generateCollectionOperationsBeforeMove(generator, moveDescs, sharedFormulas) {

                // the change descriptor to be passed to the generator methods of the collections
                var changeDesc = { type: 'moveCells', sheet: sheetModel.getIndex(), moveDescs: moveDescs };

                generator.setSheetIndex(changeDesc.sheet);
                var promise = Utils.invokeChainedAsync(
                    // create the undo operations for merged ranges (only in the own sheet)
                    function () { return sheetModel.getMergeCollection().generateMoveCellsOperations(generator, moveDescs); },
                    // create the undo operations for hyperlink ranges
                    function () { return sheetModel.getHyperlinkCollection().generateMoveCellsOperations(generator, moveDescs); },
                    // create the operations for table ranges (only in the own sheet) before updating formulas
                    function () { return sheetModel.getTableCollection().generateMoveCellsOperations(generator, moveDescs); },
                    // create the operations and undo operations for the globally defined names
                    function () { return docModel.getNameCollection().generateFormulaOperations(generator, changeDesc); }
                );

                // create the operations and undo operations for all sheet collections with formulas
                promise = promise.then(function () {
                    return docModel.generateOperationsForAllSheets(generator, function (currSheetModel) {
                        return Utils.invokeChainedAsync(
                            function () { return currSheetModel.getNameCollection().generateFormulaOperations(generator, changeDesc); },
                            function () { return currSheetModel.getValidationCollection().generateFormulaOperations(generator, changeDesc); },
                            function () { return currSheetModel.getCondFormatCollection().generateFormulaOperations(generator, changeDesc); }
                        );
                    });
                });

                // generate opertions for shared formulas
                promise = promise.then(function () {
                    return self.iterateSliced(sharedFormulaMap.iterator(), function (sharedFormula) {

                        if (!sharedFormula.range || !sharedFormula.anchor) { return; }

                        var deleteSharedFormula = false,
                            addressArray = [],
                            currentSharedFormula = sharedFormula.clone(),
                            promise2 = null;

                        // TODO YR use iterateSliced
                        moveDescs.forEach(function (moveDesc) {
                            // move the shared formula anchor
                            sharedFormula.anchor = moveDesc.transformAddress(sharedFormula.anchor);
                            // move the shared formula range
                            sharedFormula.range = moveDesc.transformRange(sharedFormula.range);
                        });

                        //TODO YR check if the range is affected

                        // iterate throw the cells and set the shared formula range and check if the formula gets errors
                        if (sharedFormula.range && sharedFormula.anchor) {
                            // TODO YR move sharedFormulaTokenArray up to use it in the next promise
                            var sharedFormulaTokenArray = new TokenArray(sheetModel, 'cell');
                            sharedFormulaTokenArray.parseFormula('op', sharedFormula.formula);

                            // Check the shared formula cells if this token array contains any reference tokens,
                            // or any tokens referring to defined names.
                            if (sharedFormulaTokenArray.hasRefTokens()) {
                                // check if the shared formula anchor is valid and the cells count has not changed, otherwise delete the shared formula
                                if (sharedFormula.anchor && currentSharedFormula.range.cells() === sharedFormula.range.cells()) {

                                    var sharedForumlaRangeIterator = createAddressIterator(currentSharedFormula.range, { type: 'sharedformula' });

                                    promise2 = self.iterateSliced(sharedForumlaRangeIterator, function (address, result) {

                                        // Check the address is filtered by shared fromula id or if the shared formula cell is a cell of the current sharedFormula and add the the address to the addressArray
                                        if (sharedFormula.id === result.model.si) {
                                            addressArray.push(address);
                                            if (!deleteSharedFormula) {
                                                var sharedFormulaTokenArray = self.getCellTokenArray(address);
                                                var value = sharedFormulaTokenArray.resolveOperation('op', changeDesc);

                                                if (value && value.error) {
                                                    deleteSharedFormula = true;
                                                }
                                            }
                                        }
                                    }, { delay: 'immediate' });
                                } else {
                                    deleteSharedFormula = true;
                                }
                            }
                        } else {
                            deleteSharedFormula = true;
                        }
                        promise2 = promise2 || $.when();

                        // update the shared formula or the anchor cell
                        promise2 = promise2.then(function () {

                            if (deleteSharedFormula) {
                                sharedFormula.anchor = currentSharedFormula.anchor;
                                sharedFormula.range = currentSharedFormula.range;
                            } else {
                                var newTokenArray = new TokenArray(sheetModel, 'cell');
                                newTokenArray.parseFormula('op', sharedFormula.formula);
                                var value = newTokenArray.resolveOperation('op', changeDesc);
                                // if the range, anchor or the formula of the shared formula has changed, add the current shared formula and new shared formula to the sharedFormulas array
                                // to create a operation after the move operations
                                if (!currentSharedFormula.range.equals(sharedFormula.range) || !currentSharedFormula.anchor.equals(sharedFormula.anchor) || (value && value.new !== sharedFormula.formula)) {
                                    sharedFormula.formula = value && value.new ? value.new : sharedFormula.formula;
                                    sharedFormulas.newAnchorCells.push({
                                        sharedFormula: sharedFormula.clone(),
                                        currentSharedFormula: currentSharedFormula.clone()
                                    });

                                }
                            }

                        });

                        // delete the shared formula and create simple formula cells
                        promise2 = promise2.then(function () {
                            if (!deleteSharedFormula || !sharedFormula.range) { return; }

                            var deleteSharedFormulaBuilder = new CellOperationsBuilder(sheetModel, generator, self, { prependUndo: true });
                            var filteredAddresses = addressArray.length;

                            var sharedForumlaRangeIterator = filteredAddresses ?
                                IteratorUtils.createArrayIterator(addressArray) :
                                createAddressIterator(sharedFormula.range, { type: 'sharedformula' });

                            var promiseDelete = self.iterateSliced(sharedForumlaRangeIterator, function (address) {
                                var sharedFormulaCellModel = self.getCellModel(address);

                                if (filteredAddresses || sharedFormula.id === sharedFormulaCellModel.si) {
                                    var sharedFormulaTokenArray = self.getCellTokenArray(address);
                                    var value = sharedFormulaTokenArray.resolveOperation('op', changeDesc);

                                    var formula;
                                    if (value && value.new) {
                                        formula = value.new;
                                    } else if (sharedFormulaCellModel.f) {
                                        formula = sharedFormulaCellModel.f;
                                    } else {
                                        formula = sharedFormulaTokenArray.getFormula('op');
                                    }
                                    // Set the formula for the former shared formula cell and remove the shared formula ID and range
                                    deleteSharedFormulaBuilder.createCells(address, { f: formula, si: null, sr: null }, 1, { collectsharedformulas: false });
                                }
                            }, { delay: 'immediate' });

                            return promiseDelete.then(function () {
                                return deleteSharedFormulaBuilder.finalizeOperations();
                            });
                        });

                        return promise2;
                    }, { delay: 'immediate' });
                });

                return promise;
            }

            // generates the operations and undo operations for other sheet collections in the document, after the cells have been moved
            function generateCollectionOperationsAfterMove(generator, moveDescs, columns, insert) {

                // the change descriptor to be passed to the generator methods of the collections
                var changeDesc = { type: 'moveCells', sheet: sheetModel.getIndex(), moveDescs: moveDescs };

                // create the operations and undo operations for all drawing objects
                var promise = docModel.generateOperationsForAllSheets(generator, function (currSheetModel) {
                    return Utils.invokeChainedAsync(
                        function () { return currSheetModel.getCellCollection().generateFormulaOperations(generator, changeDesc); },
                        function () { return currSheetModel.getDrawingCollection().generateFormulaOperations(generator, changeDesc); }
                    );
                });

                // insertion mode: create the operations to copy the cell auto-styles into the new blank ranges
                if (insert) {
                    promise = promise.then(function () {
                        return self.iterateArraySliced(moveDescs, function (moveDesc) {
                            var columns = moveDesc.columns;
                            return self.iterateArraySliced(moveDesc.targetIntervals, function (blankInterval) {

                                // nothing to do when inserting at the beginning of the sheet
                                if (blankInterval.first === 0) { return; }

                                // Create a parallel iterator that visits all columns/rows with at least one defined cell before or after the insertion line.
                                // Example: When inserting columns C:D, these columns are now empty, and the old existing cells in column B need to be extended
                                // into the new blank columns, with special border processing of the (formerly adjacent) cells in column B and column E.
                                var prevLine = moveDesc.createLineRange(blankInterval.first - 1);
                                var nextLine = moveDesc.createLineRange(blankInterval.last + 1);
                                var iterator = createParallelAddressIterator([prevLine, nextLine], { type: 'defined' });

                                // collect the range addresses for inserting defined cells into the new blank ranges
                                var fillRanges = new RangeArray();
                                var promise2 = self.iterateSliced(iterator, function (addresses) {

                                    // get the identifiers of the auto-styles before and after the new blank range
                                    var styleId1 = self.getCellStyleId(addresses[0]);
                                    var styleId2 = self.getCellStyleId(addresses[1]);

                                    // calculate the new auto-style to be inserted into the blank cells
                                    var newStyleId = generator.generateInsertedAutoStyle(styleId1, styleId2, columns);

                                    // try to extend an existing fill range, otherwise create a new range
                                    var lastFillRange = fillRanges.last();
                                    if (lastFillRange && (lastFillRange.getEnd(!columns) + 1 === addresses[0].get(!columns)) && autoStyles.areEqualStyleIds(lastFillRange.fillData.s, newStyleId)) {
                                        lastFillRange.end.move(1, !columns);
                                    } else {
                                        var fillRange = new Range(addresses[0].clone().move(1, columns), addresses[1].clone().move(-1, columns));
                                        fillRange.fillData = { s: newStyleId };
                                        fillRanges.push(fillRange);
                                    }
                                }, { delay: 'immediate' });

                                // create the fill operations, but ignore the undo operations
                                return promise2.then(function () {
                                    return generateCellRangeOperations(generator, fillRanges, function (fillDataArray) {
                                        return fillDataArray[0];
                                    }, { createUndo: false });
                                });
                            }, { delay: 'immediate' });
                        }, { delay: 'immediate' });
                    });
                }

                // bug 39869: create the missing header labels for table ranges, after inserting columns (only in the own sheet)
                if (columns && insert) {
                    promise = promise.then(function () {
                        return sheetModel.getTableCollection().generateMissingHeaderLabelOperations(generator);
                    });
                }

                return promise;
            }

            // generates the operations to create the new anchor cell
            function generateSharedFormulaOperations(generator, createSharedFormulaCellGenerator, sharedFormulas) {
                return self.iterateArraySliced(sharedFormulas.newAnchorCells, function (obj) {
                    var changeanchorBuilder = new CellOperationsBuilder(sheetModel, createSharedFormulaCellGenerator, self);
                    changeanchorBuilder.createCells(obj.sharedFormula.anchor, { f: obj.sharedFormula.formula, si: obj.sharedFormula.id, sr: obj.sharedFormula.range, sa: obj.currentSharedFormula.anchor }, 1, { collectsharedformulas: false });
                    return changeanchorBuilder.finalizeOperations();
                }).done(function () {
                    generator.appendOperations(createSharedFormulaCellGenerator);
                });
            }

            // generates the operations to insert/delete entire columns or rows; the returned
            // promise will be resolved with an array containing the move descriptor as element
            function generateMoveIntervalOperations(generator, ranges, columns, insert) {

                // the column/row collection to be modified
                var collection = columns ? colCollection : rowCollection;
                // the band interval (full interval in opposite direction)
                var bandInterval = new Interval(0, docModel.getMaxIndex(!columns));
                // create the move descriptor containing all needed interval information
                var moveDesc = new MoveDescriptor(docModel, bandInterval, ranges.intervals(columns), columns, insert);

                // a private undo generator for restoring existing cells
                var undoGenerator = sheetModel.createOperationsGenerator();
                // create undo operations to restore the existing cells in the deleted intervals
                // (must be done before generating and immediately applying the insert/delete operations)
                var promise = generateRestoreIntervalOperations(undoGenerator, moveDesc);

                // a private operations generator for additional operations from other sheet collections
                var beforeGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });

                // An shared formula helper object which includes
                // - newAnchorCells : Array with Objects which includes the current sharedFormula and the new sharedFormula object,
                //                    to create the create cells operations.
                var sharedFormulas = {
                    newAnchorCells: []
                };

                // create the operations and undo operations for other sheet contents (must be done before
                // generating and immediately applying the column/row interval operations)
                promise = promise.then(function () {
                    return generateCollectionOperationsBeforeMove(beforeGenerator, [moveDesc], sharedFormulas).done(function () {
                        generator.appendOperations(beforeGenerator);
                    });
                });

                // create the insert/delete operations for all intervals
                promise = promise.then(function () {
                    return collection.generateMoveOperations(generator, moveDesc);
                });

                // a private operations generator for additional operations from other sheet collections
                var afterGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });

                // create the operations and undo operations for other sheet contents after moving the cells
                promise = promise.then(function () {
                    return generateCollectionOperationsAfterMove(afterGenerator, [moveDesc], columns, insert).done(function () {
                        generator.appendOperations(afterGenerator);
                    });
                });

                var createSharedFormulaCellGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });

                promise = promise.then(function () {
                    return generateSharedFormulaOperations(generator, createSharedFormulaCellGenerator, sharedFormulas);
                });

                // append the undo operations generated before, resolve the promise with the move descriptor
                return promise.done(function () {
                    generator.prependOperations(afterGenerator, { undo: true });
                    generator.prependOperations(createSharedFormulaCellGenerator, { undo: true });
                    generator.appendOperations(undoGenerator, { undo: true });
                    generator.appendOperations(beforeGenerator, { undo: true });
                });
            }

            // generates the 'moveCells' operations to move the specified cell ranges;
            // the returned promise will be resolved with an array of move descriptors
            function generateMoveOperations(generator, ranges, columns, insert) {

                // the move direction for document operations, and for undo operations
                var dir = columns ? (insert ? 'right' : 'left') : (insert ? 'down' : 'up');
                var undoDir = columns ? (insert ? 'left' : 'right') : (insert ? 'up' : 'down');
                // create the column/row bands of the ranges, containing the merged interval arrays per band
                var bandIntervals = ranges[columns ? 'getRowBands' : 'getColBands']({ intervals: true });
                // the move descriptors for all column/row bands to resolve the final promise with
                var moveDescs = bandIntervals.map(function (bandInterval) {
                    return new MoveDescriptor(docModel, bandInterval, bandInterval.intervals, columns, insert);
                });

                // a private undo generator for restoring existing cells
                var undoGenerator = sheetModel.createOperationsGenerator();
                // the cell ranges containing the cells that need to be restored on undo
                var undoRanges = RangeArray.map(moveDescs, function (moveDesc) {
                    return moveDesc.createRanges(moveDesc.deleteIntervals);
                });
                // create undo operations to restore the existing cells in the deleted ranges
                // (must be done before generating and immediately applying the 'moveCells' operations)
                var promise = generateRestoreCellOperations(undoGenerator, undoRanges);

                // a private operations generator for additional operations from other sheet collections
                var beforeGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });

                // An shared formula helper object which includes
                // - newAnchorCells : Array with Objects which includes the current sharedFormula and the new sharedFormula object,
                //                    to create the create cells operations.
                var sharedFormulas = {
                    newAnchorCells: []
                };

                // create the operations and undo operations for other sheet contents (must be done before
                // generating and immediately applying the column/row interval operations)
                promise = promise.then(function () {
                    return generateCollectionOperationsBeforeMove(beforeGenerator, moveDescs, sharedFormulas).done(function () {
                        generator.appendOperations(beforeGenerator);
                    });
                });

                // generate and apply the 'moveCells' operations for all intervals in all bands
                promise = promise.then(function () {
                    return self.iterateArraySliced(moveDescs, function (moveDesc) {
                        return self.iterateArraySliced(moveDesc.targetIntervals, function (interval) {
                            var targetRange = moveDesc.createRange(interval);
                            generator.generateRangeOperation(Operations.MOVE_CELLS, targetRange, { dir: dir });
                            generator.generateRangeOperation(Operations.MOVE_CELLS, targetRange, { dir: undoDir }, { undo: true, prepend: true });
                        }, { delay: 'immediate', reverse: !insert }); // delete mode: in reversed order
                    }, { delay: 'immediate' });
                });

                // a private operations generator for additional operations from other sheet collections
                var afterGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });
                // create the operations and undo operations for other sheet contents after moving the cells
                promise = promise.then(function () {
                    return generateCollectionOperationsAfterMove(afterGenerator, moveDescs, columns, insert).done(function () {
                        generator.appendOperations(afterGenerator);
                    });
                });

                var createSharedFormulaCellGenerator = sheetModel.createOperationsGenerator({ applyImmediately: true });

                promise = promise.then(function () {
                    return generateSharedFormulaOperations(generator, createSharedFormulaCellGenerator, sharedFormulas);
                });

                // append the undo operations for cells, resolve the promise with the array of move descriptors
                return promise.done(function () {
                    generator.prependOperations(afterGenerator, { undo: true });
                    generator.prependOperations(createSharedFormulaCellGenerator, { undo: true });
                    generator.appendOperations(undoGenerator, { undo: true });
                    generator.appendOperations(beforeGenerator, { undo: true });
                });
            }

            // the actual implementation of the public method returned from local scope
            function generateMoveCellsOperations(generator, ranges, direction) {
                SheetUtils.log('ranges=' + ranges + ' direction=' + direction);

                // divide the ranges into column, row, and cell ranges (mixed range types not supported)
                var rangeGroups = getRangeGroups(ranges);
                if (_.isEmpty(rangeGroups)) { return $.when(new IntervalArray()); }
                if (_.size(rangeGroups) > 1) { return $.Deferred().reject(); }

                // whether to change the column index of the cells (exit on wrong type of ranges)
                var columns = (direction === 'left') || (direction === 'right');
                // whether to insert or delete the columns/rows
                var insert = direction === (columns ? 'right' : 'down');

                // fail if inserting/deleting vertically in a column range, or horizontally in a row range
                if (columns ? rangeGroups.rowRanges : rangeGroups.colRanges) { return $.Deferred().reject(); }

                // create insert/delete operations for entire columns or rows, or 'moveCells' operations for simple cell ranges,
                // together with all necessary undo operations to restore all deleted cells, and to move back all shifted cells
                if (rangeGroups.colRanges) { return generateMoveIntervalOperations(generator, rangeGroups.colRanges, true, insert); }
                if (rangeGroups.rowRanges) { return generateMoveIntervalOperations(generator, rangeGroups.rowRanges, false, insert); }
                return generateMoveOperations(generator, rangeGroups.cellRanges, columns, insert);
            }

            return SheetUtils.profileAsyncMethod('CellCollection.generateMoveCellsOperations()', generateMoveCellsOperations);
        }());

        /**
         * Generates the operations and undo operations to update or restore
         * the cell formulas in this collection.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {Object} changeDesc
         *  The properties describing the document change. The properties that
         *  are expected in this descriptor depend on the change type in its
         *  'type' property. See method TokenArray.resolveOperation() for more
         *  details.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when all operations have been
         *  generated.
         */
        this.generateFormulaOperations = function (generator, changeDesc) {

            // nothing to do without any cells
            if (!usedRange) { return $.when(); }

            // create a cell operations builder for the formula cells
            var builder = new CellOperationsBuilder(sheetModel, generator);

            // visit all formula cells in the correct order
            var iterator = createAddressIterator(usedRange, { type: 'formula' });
            var promise = self.iterateSliced(iterator, function (address, result) {
                var formulaResult = result.model.t ? result.model.t.resolveOperation('op', changeDesc) : null;
                if (formulaResult) { builder.createCells(address, { f: formulaResult.new }, 1); }
            }, { delay: 'immediate' });

            // finalize the cell operations
            return promise.then(function () {
                return builder.finalizeOperations();
            });
        };

        /**
         * Generates the operations to set the results of formulas that have
         * been recalculated by the dependency manager of the document.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {SimpleMap<Object>} resultMap
         *  A map with result descriptors. Each map element is an object with
         *  the following properties:
         *  - {Address} address
         *      The address of the formula cell to be updated.
         *  - {Object} result
         *      The result descriptor returned from the formula interpreter.
         *
         * @returns {CellCollection}
         *  A reference to this instance.
         */
        this.generateFormulaResultOperations = function (generator, resultMap) {

            // the cell operation builder
            var builder = new CellOperationsBuilder(sheetModel, generator);

            // sort the results by address (needed by cell operation builder)
            var resultDescs = resultMap.values().sort(function (desc1, desc2) {
                return Address.compare(desc1.address, desc2.address);
            });

            // process all formula results
            resultDescs.forEach(function (resultDesc) {

                // add the new result value into the 'changeCells' operation
                builder.createCells(resultDesc.address, { v: resultDesc.result.value }, 1, { collectsharedformulas: false });

                // write the dynamic URL created by HYPERLINK functions directly into the cell model
                if ('url' in resultDesc.result) {
                    updateCellModel(resultDesc.address, { url: resultDesc.result.url });
                }
            });

            // generate the cell operations
            builder.finalizeOperations({ createUndo: false, filter: true });
            return this;
        };

        /**
         * Generates the operations to update the changed shared formula cells.
         * It remove the changed cells from the sharedFormula and update
         * the anchor, range and formula.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {AddressArray} sharedFormulasForUpdate the changed shared formula cell array,
         *  with the sharedFormula ID as Key.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when all operations have been
         *  generated.
         */
        this.generateSharedFormulaUpdateOperations = function (generator, sharedFormulasForUpdate) {

            return self.iterateObjectSliced(sharedFormulasForUpdate, function (addresses, key) {

                var sharedFormula = sharedFormulaMap.get(key);

                if (sharedFormula) {
                    var updateSharedFormula = false;

                    var cellOperationsBuilder = new CellOperationsBuilder(sheetModel, generator, self, { prependUndo: true });

                    var sharedFormulaPromise = self.iterateArraySliced(addresses, function (sharedFormulaAddress) {

                        var promise = getSharedFormulaChanges(sharedFormula, { getrange: true, rangeexcludeaddress: addresses });

                        return promise.done(function (result) {

                            if (result.range && sharedFormula.anchor.equals(sharedFormulaAddress) && result.newanchor) {
                                // Anchor Address is removed from shared formula, change the anchor address
                                sharedFormula.formula = self.getCellFormula(result.newanchor, 'op');
                                sharedFormula.setAnchor(result.newanchor);
                                sharedFormula.setRange(result.range);
                                updateSharedFormula = true;
                            } else if (result.range && !sharedFormula.range.equals(result.range)) {
                                sharedFormula.setRange(result.range);
                                updateSharedFormula = true;
                            }

                            // Remove the shared formula id and range from the cell
                            cellOperationsBuilder.createCells(sharedFormulaAddress, { si: null, sr: null }, 1, { collectsharedformulas: false });
                        });
                    });

                    sharedFormulaPromise = sharedFormulaPromise.then(function () {
                        return cellOperationsBuilder.finalizeOperations();
                    });

                    return sharedFormulaPromise.then(function () {
                        if (updateSharedFormula) {
                            // Create new CellOperationsBuilder, because the new Cell can be behind the last created Cell
                            cellOperationsBuilder = new CellOperationsBuilder(sheetModel, generator, self, { prependUndo: true });
                            cellOperationsBuilder.createCells(sharedFormula.anchor, { f: sharedFormula.formula, si: sharedFormula.id, sr: sharedFormula.range }, 1, { collectsharedformulas: false });
                            return cellOperationsBuilder.finalizeOperations();
                        }
                    });
                }
                return self.createResolvedPromise();
            });
        };

        /**
         * Inserts one or more formulas calculating subtotal results into or
         * next to the specified cell range.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {String} funcKey
         *  The resource key of the subtotal function to be inserted into the
         *  formulas.
         *
         * @param {Range} range
         *  The source range used to generate the formulas.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved with the address of the target
         *  range inserted into the generated formulas when all operations have
         *  been generated.
         */
        this.generateAutoFormulaOperations = function (generator, funcKey, range) {

            function createCellContents(sourceRange, refAddress) {

                // generate the formula expression, and create a token array
                var formula = formulaGrammar.generateAutoFormula(docModel, funcKey, sourceRange);
                var tokenArray = new TokenArray(sheetModel, 'cell');
                tokenArray.parseFormula('op', formula);

                // calculate the formula result, this will also return the resulting number format
                var result = tokenArray.interpretFormula('val', { refAddress: refAddress, targetAddress: refAddress });

                // create the cell contents object, and add the number format (except standard formats)
                var contents = { f: formula, v: result.value };
                var formatCode = result.format ? getNewCellFormatCode(refAddress, result.format) : null;
                if (formatCode) { contents.format = formatCode; }

                return contents;
            }

            function findNextEmptyAddress(address, direction) {
                var cellIt = self.createLinearAddressIterator(address, direction);
                var emptyAddress = null;
                IteratorUtils.forEach(cellIt, function (address) {
                    if (self.isBlankCell(address)) {
                        emptyAddress = address;
                        return Utils.BREAK;
                    }
                });
                return emptyAddress;
            }

            var singleRow = range.singleRow();
            var contents = [];
            var emptyRow = -1;
            var x = -1;
            var targetRange = null;
            var startAddress = null;

            if (singleRow) {
                var firstAddress = self.findFirstCell(range, { type: 'value' });
                if (!firstAddress) {
                    targetRange = range;
                    singleRow = false;
                    var firstCol = getAutoFormulaRange(range.start.clone(), 'up', true).range;
                    if (!firstCol) { return $.when(range); }
                    range = Range.create(firstCol.start[0], firstCol.start[1], range.end[0], range.end[1]);
                }
            }
            if (singleRow) {
                startAddress = findNextEmptyAddress(new Address(range.end[0], Math.max(range.start[1], range.end[1])), 'right');

                var contentRange = Range.create(range.start[0], range.start[1], Math.min(range.end[0], startAddress[0] - 1), range.end[1]);

                contents.push({ c: [createCellContents(contentRange, startAddress)] });
                var targetX = startAddress[0] <= range.end[0] ? range.end[0] : range.end[0] + 1;
                targetRange = Range.create(range.start[0], range.start[1], targetX, range.end[1]);

            } else {

                for (x = range.start[0]; x <= range.end[0]; x++) {
                    var next = findNextEmptyAddress(new Address(x, range.start[1]), 'down');
                    emptyRow = Math.max(emptyRow, next[1]);
                }

                var targetY = emptyRow <= range.end[1] ? range.end[1] : range.end[1] + 1;
                startAddress = new Address(range.start[0], Math.max(emptyRow, range.end[1]));
                var sourceRange = Range.create(range.start[0], range.start[1], range.start[0], targetY - 1);
                var contentLine = [];
                contents.push({ c: contentLine });

                for (x = range.start[0]; x <= range.end[0]; x++) {
                    sourceRange.setBoth(x, true);
                    contentLine.push(createCellContents(sourceRange, new Address(x, Math.max(emptyRow, range.end[1]))));
                }
                if (!targetRange) {
                    targetRange = Range.create(range.start[0], range.start[1], range.end[0], targetY);
                }
            }
            // generate and apply the operations
            return self.generateCellContentOperations(generator, startAddress, contents).then(_.constant(targetRange));
        };

        /**
         * Generates the cell operations, and the undo operations, to sort the
         * given range with repect to a possible header-line
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator to be filled with the operations.
         *
         * @param {Range} range
         *  The address of the source cell range which should be sorted.
         *
         * @param {Address} activeCell
         *  The current active cell.
         *
         * @param {String} orderBy
         *  'ascending' / 'descending'
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {String} [options.direction='vertical horizontal']
         *      The direction in which we would sort.
         *  @param {Boolean} [options.headers]
         *      Whether a header exists, or not.
         *  @param {Number} [options.colrow]
         *      The col/row index which should be sorted
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved (with the addresses of the ranges
         *  that have really been changed) when all operations have been
         *  generated, or that will be rejected with an object with 'cause'
         *  property set to one of the following error codes:
         */
        this.generateSortOperations = function (generator, range, activeCell, orderBy, options) {

            var direction       = Utils.getStringOption(options, 'direction', 'vertical'),
                startAddress    = range.start;

            if (sheetModel.getMergeCollection().rangesOverlapAnyMergedRange(range)) {
                self.trigger('sort:merge:overlap');
                return $.when();
            }

            // transforms a 1D-Array to 2D-Array
            function transformRangeArray(rangeContents, cols) {
                var returnArray = [];

                _.each(rangeContents, function (obj, index) {
                    var line    = (Math.ceil((index + 1) / cols) - 1),
                        nr      = ((index + 1) - ((line) * cols) - 1);

                    if (!returnArray[line]) { returnArray[line] = []; }

                    returnArray[line][nr] = obj;
                });

                return returnArray;
            }

            // get the hidden col/row entries
            function getHidden() {
                var hidden      = [],
                    columns     = (direction === 'horizontal'),
                    collection  = (!columns) ? rowCollection : colCollection,
                    startIndex  = (!columns) ? startAddress[1] : startAddress[0],
                    interval    = range.interval(columns),
                    iterator    = IteratorUtils.createIntervalIterator(interval.first, interval.last);

                IteratorUtils.forEach(iterator, function (index) {
                    if (!collection.isEntryVisible(index)) {
                        hidden.push((index - startIndex));
                    }
                });

                return (hidden.length === 0) ? null : hidden.reverse();
            }
            // Iterate all shared formulas and test if the shared formula range overlaps the sort range.
            // If it overlaps iterate all cells from the shared formula range and remove the link to the
            // shared formula from the cell and set the cell formula.
            function removeCellsFromSharedFormulas(generator, range) {
                return self.iterateSliced(sharedFormulaMap.iterator(), function (sharedFormula) {
                    if (!sharedFormula.range || !sharedFormula.anchor) { return; }

                    // test if the shared formula range overlaps the sort range
                    if (sharedFormula.range.overlaps(range)) {
                        var deleteSharedFormulaBuilder = new CellOperationsBuilder(sheetModel, generator, self, { prependUndo: true });

                        var sharedForumlaRangeIterator = createAddressIterator(sharedFormula.range, { type: 'sharedformula' });

                        // remove the shared formula link from the cells
                        var promiseDelete = self.iterateSliced(sharedForumlaRangeIterator, function (address) {
                            var sharedFormulaCellModel = self.getCellModel(address);

                            if (sharedFormula.id === sharedFormulaCellModel.si) {
                                var sharedFormulaTokenArray = self.getCellTokenArray(address);

                                var formula;
                                if (sharedFormulaCellModel.f) {
                                    formula = sharedFormulaCellModel.f;
                                } else {
                                    formula = sharedFormulaTokenArray.getFormula('op');
                                }
                                // Set the formula for the former shared formula cell and remove the shared formula ID and range
                                deleteSharedFormulaBuilder.createCells(address, { f: formula, si: null, sr: null }, 1, { collectsharedformulas: false });
                            }
                        }, { delay: 'immediate' });

                        return promiseDelete.then(function () {
                            return deleteSharedFormulaBuilder.finalizeOperations();
                        });
                    }
                });
            }

            var promise = removeCellsFromSharedFormulas(generator, range);

            // create and apply the operation
            return promise.then(function () {

                var endAddress      = range.end,
                    hasheadline     = Utils.getBooleanOption(options, 'headers'),
                    colrow          = Utils.getNumberOption(options, 'colrow', (direction === 'vertical') ? activeCell[0] : activeCell[1]),
                    hiddenIndexes   = getHidden();

                // reset the index realtive to the selected range
                colrow = (direction === 'vertical') ? (colrow - range.start[0]) : (colrow - range.start[1]);

                // if no headline-handling is set, try to detect automatically
                if (_.isUndefined(hasheadline)) {
                    hasheadline = self.hasRangeHeadline(range, direction);
                }

                // shrink the range, when there were headlines inside the selected range
                if (hasheadline) {
                    startAddress = (direction === 'vertical') ? new Address((startAddress[0]), (startAddress[1] + 1)) : new Address((startAddress[0] + 1), (startAddress[1]));
                    range = new Range(startAddress, endAddress);
                }

                // get all values of given range
                var arrContents = transformRangeArray(self.getRangeCellModels(range), range.cols());

                // swap the 2D array in case of 'horizontal' sorting
                if (direction !== 'vertical') { arrContents = _.zip.apply(_, arrContents); }

                // remove hidden col/row entries
                if (hiddenIndexes) {
                    hiddenIndexes.forEach(function (i) { arrContents.splice(i, 1); });
                }

                // first, the proper sorting (by 'FormulaUtils.compareScalars') of styles
                arrContents.sort(function (obj1, obj2) {
                    var res     = FormulaUtils.compareScalars(obj1[colrow].s, obj2[colrow].s, { nullMode: (orderBy === 'ascending') ? 'greater' : 'less' });

                    // reverse the sorted array if the sort direction was 'desc'
                    return (orderBy === 'descending') ? (res * -1) : res;
                });

                // then, the proper sorting (by 'FormulaUtils.compareScalars') of values
                arrContents.sort(function (obj1, obj2) {
                    var res     = FormulaUtils.compareScalars(obj1[colrow].v, obj2[colrow].v, { nullMode: (orderBy === 'ascending') ? 'greater' : 'less' });

                    // reverse the sorted array if the sort direction was 'desc'
                    return (orderBy === 'descending') ? (res * -1) : res;
                });

                // add blank entries for hidden cols/rows
                if (hiddenIndexes) {
                    hiddenIndexes.sort();
                    hiddenIndexes.forEach(function (i) { arrContents.splice(i, 0, {}); });
                }

                // build the 'contents' object which should be used in the operations
                var matrix = arrContents.map(function (arrObj, colIndex) {
                    return (_.includes(hiddenIndexes, colIndex)) ? {} : { c: arrObj };
                });

                // swap the matrix in case of 'horizontal' sorting
                if (direction !== 'vertical' && matrix.length > 0) {
                    matrix = _.times(matrix[0].c.length, function (col) {
                        // transform a column from the old matrix to a row in the new matrix
                        var cells = _.times(matrix.length, function (row) {
                            return ((matrix[row].c)) ? matrix[row].c[col] : {};
                        });
                        return { c: cells };
                    });
                }

                return self.generateCellContentOperations(generator, range.start, matrix);
            });
        };

        // initialization -----------------------------------------------------

        // additional processing and event handling after the document has been imported
        sheetModel.waitForImportSuccess(function () {

            // update the sheet indexes after the sheet collection has been manipulated
            this.listenTo(docModel, 'transform:sheet', function (event, toSheet, fromSheet) {
                modelMap.forEach(function (cellModel) {
                    if (cellModel.t) { cellModel.t.transformSheet(toSheet, fromSheet); }
                });
            });

        }, this);

        // destroy all class members on destruction
        this.registerDestructor(function () {
            colMatrix.destroy();
            rowMatrix.destroy();
            self = docModel = numberFormatter = formulaGrammar = autoStyles = null;
            sheetModel = colCollection = rowCollection = listCollection = null;
            modelMap = colMatrix = rowMatrix = sharedFormulaMap = null;
        });

    } // class CellCollection

    // exports ================================================================

    return TriggerObject.extend({ constructor: CellCollection });

});
